// ############################################################################################
//  vdspsb.cpp
//  Implementation of _CVortexSoundBuffer class using direct access to /dev/dsp
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
// ***** VortexGE headers
#include <vcwave.h>
#include <vmdisp.h>
#include <auddsp/vdspsb.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** VIF header
#define _VSW_ID1_ 'V'
#define _VSW_ID2_ 'S'
#define _VSW_ID3_ 'W'
#define _VSW_VER_ 100

// ***** Compression type
#define _VSW_NONE_  0
#define _VSW_DUMMY_ 1

// ============================================================================================
// Local functions prototype
// ============================================================================================
// ***** Image functions
_vuint32 _CompressSoundBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff, _vuint32 un32SrcBuffSize);
_vuint32 _UncompressSoundBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff);

// ============================================================================================
// Structures
// ============================================================================================
// ***** VSW header
struct _TVSWHeader
{ 
  _vchar              acID[3];       // File ID, must be "VSW"
  _vuint8             un8Version;    // File version
  _TVortexAudioFormat tFormat;       // The audio format
  _vuint32            un32Size;      // Size of the uncompressed audio
  _vuint8             un8CompL;      // Compression type (left/mono channel)
  _vuint8             un8CompR;      // Compression type (right channel)
  _vuint32            un32SizeCompL; // Compressed audio size in bytes (left/mono channel)
  _vuint32            un32SizeCompR; // Compressed audio size in bytes (right channel)
  _vuint16            un16Reserved1; // Reserved, must be zero
  _vuint32            un32Reserved2; // Reserved, must be zero
} __attribute__((packed));

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The audio sound-buffer class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexSoundBuffer::_CVortexSoundBuffer()
{
  memset(&mt_tD, 0, sizeof(mt_tD));
  _pfnGD_ = NULL;
}
_CVortexSoundBuffer::~_CVortexSoundBuffer()
{
  _FreeBuffer_();
}

// ***** Sound buffer functions
_vbool _CVortexSoundBuffer::AdjustAmplitude(_vfloat fLeftGain, _vfloat fRightGain)
{
  // ..... Temporary variables
  _vuint32 un32Cnt, un32Size;
 
  // ..... Check 
  if(mt_tD.pucLeft == NULL)
  {
    VError(_F_, _L_, "No sound data to amplify (this _CVortexSoundBuffer is uninitialized)");
    return(false);
  }

  // ..... Adjust amplitude
  if(mt_tF.un8Bits == 8)
  {
    un32Size = _vuint32(mt_tD.pucLeftMax - mt_tD.pucLeft);
    _vint8  *n8L = (_vint8*) mt_tD.pucLeft;
    _vint8  *n8R = (_vint8*) mt_tD.pucRight;
    _vint32 n32Tmp;
    for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    {
      n32Tmp = _vint32(_vfloat(*n8L) * fLeftGain);
      if(n32Tmp > 127) n32Tmp = 127;
      if(n32Tmp < -128) n32Tmp = -128;
      *n8L++ = _vint8(n32Tmp);
      if(mt_tF.un8Channels == 2)
      {
        n32Tmp = _vint32(_vfloat(*n8R) * fRightGain);
        if(n32Tmp > 127) n32Tmp = 127;
        if(n32Tmp < -128) n32Tmp = -128;
        *n8R++ = _vint8(n32Tmp);
      }
    }
  }
  else
  {
    un32Size = _vuint32(mt_tD.pucLeftMax - mt_tD.pucLeft) / 2;
    _vint16 *n16L = (_vint16*) mt_tD.pucLeft;
    _vint16 *n16R = (_vint16*) mt_tD.pucRight;
    _vint32 n32Tmp;
    for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    {
      n32Tmp = _vint32(_vfloat(*n16L) * fLeftGain);
      if(n32Tmp > 32767) n32Tmp = 32767;
      if(n32Tmp < -32768) n32Tmp = -32768;
      *n16L++ = _vint16(n32Tmp);
      if(mt_tF.un8Channels == 2)
      {
        n32Tmp = _vint32(_vfloat(*n16R) * fRightGain);
        if(n32Tmp > 32767) n32Tmp = 32767;
        if(n32Tmp < -32768) n32Tmp = -32768;
        *n16R++ = _vint16(n32Tmp);
      }
    }
  }

  // ..... OK !
  return(true);
}
_vbool _CVortexSoundBuffer::ConvertStereoToMono(_vvoid)
{
  // ..... Check 
  if(mt_tD.pucLeft == NULL)
  {
    VError(_F_, _L_, "No sound data to convert (this _CVortexSoundBuffer is uninitialized)");
    return(false);
  }
  if(mt_tF.un8Channels == 1) return(true);

  // ..... Do conversion
  _Stereo2Mono_();
  _AssignCnvFunc_(mt_tD.un32PosStep != 0);

  // ..... OK !
  return(true);
}
_vbool _CVortexSoundBuffer::Convert16BitTo8Bit(_vvoid)
{
  // ..... Check 
  if(mt_tD.pucLeft == NULL)
  {
    VError(_F_, _L_, "No sound data to convert (this _CVortexSoundBuffer is uninitialized)");
    return(false);
  }
  if(mt_tF.un8Bits == 8) return(true);

  // ..... Do conversion
  if(!_1628_()) return(false);
  _AssignCnvFunc_(mt_tD.un32PosStep != 0);

  // ..... OK !
  return(true);
}
_vvoid _CVortexSoundBuffer::_ExtractFormat_(_TVortexAudioFormat *ptFormat, _EVortexRawFormat eFormat)
{
  if(eFormat == ERF_U8Mono)
  {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_U8Stereo)
  {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_S8Mono)
  {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_S8Stereo)
  {
    ptFormat->un8Bits     = 8;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_U16LEMono)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_U16LEStereo)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_S16LEMono)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_S16LEStereo)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 0;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_U16BEMono)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_U16BEStereo)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 0;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 2;
  }
  else if(eFormat == ERF_S16BEMono)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 1;
  }
  else if(eFormat == ERF_S16BEStereo)
  {
    ptFormat->un8Bits     = 16;
    ptFormat->un8Sign     = 1;
    ptFormat->un8Endian   = 1;
    ptFormat->un8Channels = 2;
  }
}
_vvoid _CVortexSoundBuffer::_SwapEndian_(_vvoid)
{
  // ..... Temporary variables
  _vuint32 un32Cnt, un32Size;
  _vuchar  ucTmp;
 
  // ..... Swap the endians
  un32Size = _vuint32(mt_tD.pucLeftMax - mt_tD.pucLeft);
  for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt += 2)
  {
    ucTmp = mt_tD.pucLeft[un32Cnt];
    mt_tD.pucLeft[un32Cnt] = mt_tD.pucLeft[un32Cnt + 1];
    mt_tD.pucLeft[un32Cnt + 1] = ucTmp;
    if(mt_tF.un8Channels == 2)
    {
      ucTmp = mt_tD.pucRight[un32Cnt];
      mt_tD.pucRight[un32Cnt] = mt_tD.pucRight[un32Cnt + 1];
      mt_tD.pucRight[un32Cnt + 1] = ucTmp;
    }
  }

  // ..... Adjust the format
  if(mt_tF.un8Endian == 0)
    mt_tF.un8Endian = 1;
  else
    mt_tF.un8Endian = 0;
}
_vvoid _CVortexSoundBuffer::_InvertSign_(_vvoid)
{
  // ..... Temporary variables
  _vuint32 un32Cnt, un32Size;
 
  // ..... Invert sign
  un32Size = _vuint32(mt_tD.pucLeftMax - mt_tD.pucLeft);
  if(mt_tF.un8Bits == 8)
  {
    for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    {
      mt_tD.pucLeft[un32Cnt] ^= 0x80;
      if(mt_tF.un8Channels == 2) mt_tD.pucRight[un32Cnt] ^= 0x80;
    }
  }
  else
  {
    if(mt_tF.un8Endian == 0)
    {
      for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt += 2)
      {
        mt_tD.pucLeft[un32Cnt + 1] ^= 0x80;
        if(mt_tF.un8Channels == 2) mt_tD.pucRight[un32Cnt + 1] ^= 0x80;
      }
    }
    else
    {
      for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt += 2)
      {
        mt_tD.pucLeft[un32Cnt] ^= 0x80;
        if(mt_tF.un8Channels == 2) mt_tD.pucRight[un32Cnt] ^= 0x80;
      }
    }
  }

  // ..... Adjust the format
  if(mt_tF.un8Sign == 0)
    mt_tF.un8Sign = 1;
  else
    mt_tF.un8Sign = 0;
}
_vvoid _CVortexSoundBuffer::_Stereo2Mono_(_vvoid)
{
  // ..... Temporary variables
  _vuint32 un32Cnt, un32Size;
  _vbool   bAllMuted = true;
 
  // ..... Convert stereo to mono (for signed sound buffer only)
  if(mt_tF.un8Bits == 8)
  {
    un32Size = _vuint32(mt_tD.pucLeftMax - mt_tD.pucLeft);
    _vint8  *n8L = (_vint8*) mt_tD.pucLeft;
    _vint8  *n8R = (_vint8*) mt_tD.pucRight;
    _vint32 n32Tmp;
    for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    {
      n32Tmp = _vint32(*n8L) + _vint32(*n8R++);
      *n8L   = _vint8(n32Tmp / 2);
      if(*n8L != 0) bAllMuted = false;
      n8L++;
    }
  }
  else
  {
    un32Size = _vuint32(mt_tD.pucLeftMax - mt_tD.pucLeft) / 2;
    _vint16 *n16L = (_vint16*) mt_tD.pucLeft;
    _vint16 *n16R = (_vint16*) mt_tD.pucRight;
    _vint32 n32Tmp;
    for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    {
      n32Tmp = _vint32(*n16L) + _vint32(*n16R++);
      *n16L  = _vint16(n32Tmp / 2);
      if(*n16L != 0) bAllMuted = false;
      n16L++;
    }
  }
  if(bAllMuted) memcpy(mt_tD.pucLeft, mt_tD.pucRight, un32Size);

  // ..... Adjust the format
  free(mt_tD.pucRight);
  mt_tD.pucRight    = NULL;
  mt_tD.pucRightMax = NULL;
  mt_tF.un8Channels = 1;
}
_vbool _CVortexSoundBuffer::_1628_(_vvoid)
{
  // ..... Temporary variables
  _vuint32 un32Cnt, un32Size;
  _vuchar  *pucL, *pucR;
 
  // ..... Allocate memory
  un32Size = _vuint32(mt_tD.pucLeftMax - mt_tD.pucLeft) / 2;
  pucL = (_vuchar *) malloc(un32Size);
  if(pucL == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to get allocate memory for temporary sound buffer");
    return(false);
  }
  pucR = NULL;
  if(mt_tF.un8Channels == 2)
  {
    pucR = (_vuchar *) malloc(un32Size);
    if(pucR == NULL)
    {
      free(pucL);
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to get allocate memory for temporary sound buffer");
      return(false);
    }
  }
  
  // ..... Convert 16-bit to 8-bit (can be used for signed and unsigned sound buffer)
  if(mt_tF.un8Sign == 0)
  {
    _vuint8  *un8L  = (_vuint8 *) pucL;
    _vuint8  *un8R  = (_vuint8 *) pucR;
    _vuint16 *un16L = (_vuint16 *) mt_tD.pucLeft;
    _vuint16 *un16R = (_vuint16 *) mt_tD.pucRight;
    for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    {
      *un8L++ = _vuint8(*un16L++ / 256);
      if(mt_tF.un8Channels == 2) *un8R++ = _vuint8(*un16R++ / 256);
    }
  }
  else
  {
    _vint8  *n8L  = (_vint8 *) pucL;
    _vint8  *n8R  = (_vint8 *) pucR;
    _vint16 *n16L = (_vint16 *) mt_tD.pucLeft;
    _vint16 *n16R = (_vint16 *) mt_tD.pucRight;
    for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    {
      *n8L++ = _vuint8(*n16L++ / 256);
      if(mt_tF.un8Channels == 2) *n8R++ = _vuint8(*n16R++ / 256);
    }
  }

  // ..... Free old buffer
  free(mt_tD.pucLeft);
  if(mt_tD.pucRight) free(mt_tD.pucRight);
  
  // ..... Adjust the buffer
  mt_tD.pucLeft     = pucL;
  mt_tD.pucLeftMax  = mt_tD.pucLeft + un32Size - 1;
  if(mt_tF.un8Channels == 2)
  {
    mt_tD.pucRight    = pucR;
    mt_tD.pucRightMax = mt_tD.pucRight + un32Size - 1;
  }
  if(mt_tD.un32PosMax) mt_tD.un32PosMax  = mt_tD.un32PosMax >> 1;

  // ..... Adjust the format
  mt_tF.un8Bits = 8;
  
  // ..... OK !
  return(true);
}
_vvoid _CVortexSoundBuffer::_AssignCnvFunc_(_vbool bDifFreq)
{
  if(bDifFreq)
  {
    if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1)  &&
       (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 1))
      _pfnGD_ = _X8Mn8Mn_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _X8Mn8St_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 1))
      _pfnGD_ = _X8Mn16Mn_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _X8Mn16St_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 2) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _X8St8St_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 2) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _X8St16St_;
    else if((mt_tF.un8Bits == 16) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 1))
      _pfnGD_ = _X16Mn16Mn_;
    else if((mt_tF.un8Bits == 16) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _X16Mn16St_;
    else if((mt_tF.un8Bits == 16) && (mt_tF.un8Channels == 2) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _X16St16St_;
  }
  else
  {
    if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1)  &&
       (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 1))
      _pfnGD_ = _8Mn8Mn_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _8Mn8St_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 1))
      _pfnGD_ = _8Mn16Mn_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _8Mn16St_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 2) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _8St8St_;
    else if((mt_tF.un8Bits == 8) && (mt_tF.un8Channels == 2) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _8St16St_;
    else if((mt_tF.un8Bits == 16) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 1))
      _pfnGD_ = _16Mn16Mn_;
    else if((mt_tF.un8Bits == 16) && (mt_tF.un8Channels == 1) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _16Mn16St_;
    else if((mt_tF.un8Bits == 16) && (mt_tF.un8Channels == 2) &&
            (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 2))
      _pfnGD_ = _16St16St_;
  }
}
_vvoid _CVortexSoundBuffer::_FreeBuffer_(_vvoid)
{
  // ..... If this sound buffer is playing, stop it first
  for(_vuint8 un8Cnt = 0; un8Cnt < _MAX_SB_; un8Cnt++)
  {
    if(_pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB == this)
    {
#ifdef _IA64_  
      VInfo(_F_, _L_, "Removing playing sound buffer %X from list #%d", _vuint64(this), un8Cnt);
#else
      VInfo(_F_, _L_, "Removing playing sound buffer %X from list #%d", _vuint32(this), un8Cnt);
#endif
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bPlaying = false;
      _pcsAudioEnv_->mt_atPSB[un8Cnt].tData.bLooping = false;
      if(_pcsAudioEnv_->mt_csTWriter.IsRunning()) usleep(100000);
      _pcsAudioEnv_->mt_atPSB[un8Cnt].pcsSB = NULL;
      break;
    }
  }
  if(_pcsAudioEnv_->mt_csTWriter.IsRunning()) usleep(100000);
  
  // ..... Free memory
  if(mt_tD.pucLeft)  free(mt_tD.pucLeft);
  if(mt_tD.pucRight) free(mt_tD.pucRight);
}

// ***** File I/O functions
_vbool _CVortexSoundBuffer::LoadRAW(const _vchar *pcFileName, _EVortexRawFormat eFormat, _vint32 n32Frequency)
{
  // ..... Temporary variables
  struct stat         tFS;
  _TVortexAudioFormat tF;
  _vint32             n32Handle;
  _vuint32            un32Size, un32Cnt;
  _vchar              acFile[256];
  _vuchar              *pucL, *pucR, *pucTmpL, *pucTmpR;

  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".raw") == NULL) strcat(acFile, ".raw");
  n32Handle = open(acFile, O_RDONLY);
  if(n32Handle< 0)
  {
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    
  if(fstat(n32Handle, &tFS) < 0)
  {
    close(n32Handle);
    VSetErrorCode(VERR_GetFileStatus);
    VError(_F_, _L_, "Failed to get file status from \"%s\"", acFile);
    return(false);
  }
  
  // ..... Calculate some informations
  _ExtractFormat_(&tF, eFormat);
  tF.un32Frequency = n32Frequency;
  un32Size = (tFS.st_size + 3) & (~3);
  if(tF.un8Channels == 2) un32Size >>= 1;
  if(un32Size > _MAX_WBUFF_SIZE_)
  {
    close(n32Handle);
    VSetErrorCode(VERR_NotSupported);
    VError(_F_, _L_, "Cannot load file of which channel size > %u bytes", _MAX_WBUFF_SIZE_);
    return(false);
  }    
  
  // ..... Allocate memory for buffer
  pucL = (_vuchar *) malloc(un32Size);
  if(pucL == NULL)
  {
    close(n32Handle);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to get allocate memory for sound buffer");
    return(false);
  }
  memset(pucL, 0, un32Size);
  pucR = NULL;
  if(tF.un8Channels == 2)
  {
    pucR = (_vuchar *) malloc(un32Size);
    if(pucR == NULL)
    {
      free(pucL);
      close(n32Handle);
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to get allocate memory for sound buffer");
      return(false);
    }
    memset(pucR, 0, un32Size);
  }
  
  // ..... Load file
  pucTmpL = pucL;
  pucTmpR = pucR;
  un32Cnt = 0;
  while(un32Cnt < _vuint32(tFS.st_size))
  {
    if(tF.un8Bits == 8)
    {
      if(read(n32Handle, pucTmpL, 1) < 0) goto ERRRead;
      pucTmpL++;
      un32Cnt++;
    }
    else
    {
      if(read(n32Handle, pucTmpL, 2) < 0) goto ERRRead;
      pucTmpL += 2;
      un32Cnt += 2;
    }
    if(tF.un8Channels == 2)
    {
      if(tF.un8Bits == 8)
      {
        if(read(n32Handle, pucTmpR, 1) < 0) goto ERRRead;
        pucTmpR++;
        un32Cnt++;
      }
      else
      {
        if(read(n32Handle, pucTmpR, 2) < 0) goto ERRRead;
        pucTmpR += 2;
        un32Cnt += 2;
      }
    }
  }

  // ..... Close file
  close(n32Handle);

  // ..... Free previous sound (if any)
  _FreeBuffer_();
  memset(&mt_tD, 0, sizeof(mt_tD));
  _pfnGD_ = NULL;

  // ..... Copy data
  memcpy(&mt_tF, &tF, sizeof(mt_tF));

  // ..... Calculate some parameters
  if(mt_tF.un8Channels == 2)
  {
    mt_tD.pucLeft     = pucL;
    mt_tD.pucLeftMax  = mt_tD.pucLeft + (tFS.st_size >> 1) - 1;
    mt_tD.pucRight    = pucR;
    mt_tD.pucRightMax = mt_tD.pucRight + (tFS.st_size >> 1) - 1;
  }
  else
  {
    mt_tD.pucLeft     = pucL;
    mt_tD.pucLeftMax  = mt_tD.pucLeft + tFS.st_size - 1;
    mt_tD.pucRight    = NULL;
    mt_tD.pucRightMax = NULL;
  }
  if(mt_tF.un32Frequency != _pcsAudioEnv_->mt_tD.tFormat.un32Frequency)
  {
    if(mt_tF.un8Channels == 2)
      mt_tD.un32PosMax  = (tFS.st_size >> 1) << _WSTEP_PREC_;
    else
      mt_tD.un32PosMax  = tFS.st_size << _WSTEP_PREC_;
    mt_tD.un32PosStep = _vuint32(_vfloat(mt_tF.un32Frequency) / _vfloat(_pcsAudioEnv_->mt_tD.tFormat.un32Frequency) * _vfloat(1 << _WSTEP_PREC_));
    if(mt_tF.un8Bits == 16) mt_tD.un32PosStep <<= 1;
  }
  else
  {
    mt_tD.un32PosMax  = 0;
    mt_tD.un32PosStep = 0;
  }
  
  // ..... Assuming that this system is a little-endian (x86),
  //       swap the endian if the sound buffer's is 16-bit and not little-endian
  if((mt_tF.un8Bits == 16) && (mt_tF.un8Endian != 0)) _SwapEndian_();
    
  // ..... The VortexGE sound mixer always use signed numbers,
  //       make sure that the sound buffer is also signed
  if(mt_tF.un8Sign == 0) _InvertSign_();
  
  // ..... If the primary sound buffer is mono, there is no need for us to have
  //       a stereo sound buffer, so convert it to mono
  if((mt_tF.un8Channels == 1) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 0)) _Stereo2Mono_();
    
  // ..... If the primary sound buffer is 8-bit, there is no need for us to have
  //       an 16-bit sound buffer, so convert it to 8-bit
  if((mt_tF.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8))
  {  
    if(!_1628_())
    {
      free(mt_tD.pucLeft);
      if(mt_tD.pucRight) free(mt_tD.pucRight);
      memset(&mt_tD, 0, sizeof(mt_tD));
      memset(&mt_tF, 0, sizeof(mt_tF));
      VError(_F_, _L_, "Failed to convert 16-bit to 8-bit");
      return(false);
    }
  }

  // ..... Assign real-time conversion function
  _AssignCnvFunc_(mt_tD.un32PosStep != 0);
 
  // ..... OK !
  return(true);
  
  // ..... Error reading file
ERRRead:;
  free(pucL);
  if(mt_tF.un8Channels == 2) free(pucR);
  close(n32Handle);
  VSetErrorCode(VERR_ReadFile);
  VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
  return(false);  
}
_vbool _CVortexSoundBuffer::LoadWAV(const _vchar *pcFileName)
{
  // ..... Temporary variables
  _TVortexFILE        *ptFile;
  _TCWFPCMHEADER      tHeader;
  _TVortexAudioFormat tF;
  _vuchar             *pucL, *pucR, *pucLt, *pucRt;
  _vchar              acFile[256];
  _vuint32            un32Size, un32Get, un32Cnt;

  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".wav") == NULL) strcat(acFile, ".wav");
  ptFile = fopen(acFile, "rb");
  if(ptFile == NULL)
  {
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    

  // ..... Read header
  if(fread(&tHeader, sizeof(tHeader), 1, ptFile) < 1)
  {
    VSetErrorCode(VERR_ReadFile);
    VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
    return(false);
  }
    
  // ..... Check header
  if(memcmp(tHeader.tChunk.aucChunkID, "RIFF", 4))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" is an invalid RIFF file", acFile);
    return(false);
  }
  if(memcmp(tHeader.tChunk.aucFormat, "WAVE", 4))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" is an invalid WAVE file", acFile);
    return(false);
  }
  if(memcmp(tHeader.tFormat.aucSubChunk1ID, "fmt ", 4))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has an invalid \"fmt \" sub-chunk ID", acFile);
    return(false);
  }
  if(memcmp(tHeader.tData.aucSubChunk2ID, "data", 4))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has an invalid \"data\" sub-chunk ID", acFile);
    return(false);
  }
  if(tHeader.tChunk.un32ChunkSize != (20 + tHeader.tFormat.un32SubChunk1Size + tHeader.tData.un32SubChunk2Size))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has an invalid chunk / sub-chunk size", acFile);
    return(false);
  }
  if(tHeader.tFormat.un32SubChunk1Size != 16)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has an invalid \"fmt \" sub-chunk size, only PCM format supported", acFile);
    return(false);
  }
  if(tHeader.tFormat.un16AudioFormat != 1)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has an invalid audio format, only PCM format supported", acFile);
    return(false);
  }
  if(tHeader.tFormat.un16Channels > 2)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has more than 2 audio channels, only maximum 2 channels supported", acFile);
    return(false);
  }
  if((tHeader.tFormat.un16BitsPerSample != 8) && (tHeader.tFormat.un16BitsPerSample != 16))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has %d bps, only 8 and 16 bps supported", acFile, tHeader.tFormat.un16BitsPerSample);
    return(false);
  }
  if(tHeader.tFormat.un32ByteRate != (tHeader.tFormat.un32SampleRate * tHeader.tFormat.un16Channels * tHeader.tFormat.un16BitsPerSample / 8))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has and invalid data", acFile);
    return(false);
  }
  if(tHeader.tFormat.un16BlockAlign != (tHeader.tFormat.un16Channels * tHeader.tFormat.un16BitsPerSample / 8))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" has and invalid data", acFile);
    return(false);
  }
  
  // ..... Calculate some information
  tF.un8Bits       = tHeader.tFormat.un16BitsPerSample;
  tF.un8Sign       = (tF.un8Bits == 16) ? 1 : 0;
  tF.un8Endian     = 0;
  tF.un8Channels   = tHeader.tFormat.un16Channels;
  tF.un32Frequency = tHeader.tFormat.un32SampleRate;
  if(tF.un8Channels == 1)
    un32Size = tHeader.tData.un32SubChunk2Size;
  else 
    un32Size = tHeader.tData.un32SubChunk2Size / 2;
  if(tF.un8Bits == 8)
    un32Get = un32Size;
  else
    un32Get = un32Size / 2;
    
  // ..... Allocate memory for temporary buffer
  pucL = (_vuchar *) malloc(un32Size);
  if(pucL == NULL)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary sound buffer");
    return(false);
  }
  pucR = NULL;
  if(tF.un8Channels == 2)
  {
    pucR = (_vuchar *) malloc(un32Size);
    if(pucR == NULL)
    {
      free(pucL);
      fclose(ptFile);
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for temporary sound buffer");
      return(false);
    }
  }
  
  // ..... Read sound data
  pucLt = pucL;
  pucRt = pucR;
  for(un32Cnt = 0; un32Cnt < un32Get; un32Cnt++)
  {
    // Check
    if(feof(ptFile))
    {
      free(pucL);
      if(tHeader.tFormat.un16Channels == 2) free(pucR);
      fclose(ptFile);
      VSetErrorCode(VERR_ReadFile);
      VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
      return(false);
    }
    
    // Left / mono channel
    *pucLt++ = fgetc(ptFile);
    if(tF.un8Bits == 16) *pucLt++ = fgetc(ptFile);

    // Right channel
    if(tF.un8Channels == 2)
    {
      *pucRt++ = fgetc(ptFile);
      if(tF.un8Bits == 16) *pucRt++ = fgetc(ptFile);
    }
  }
  
  // ..... Close file
  fclose(ptFile);

  // ..... Free previous sound (if any)
  _FreeBuffer_();
  memset(&mt_tD, 0, sizeof(mt_tD));
  _pfnGD_ = NULL;

  // ..... Copy data
  memcpy(&mt_tF, &tF, sizeof(mt_tF));

  // ..... Calculate some parameters
  if(mt_tF.un8Channels == 2)
  {
    mt_tD.pucLeft     = pucL;
    mt_tD.pucLeftMax  = mt_tD.pucLeft + un32Size - 1;
    mt_tD.pucRight    = pucR;
    mt_tD.pucRightMax = mt_tD.pucRight + un32Size - 1;
  }
  else
  {
    mt_tD.pucLeft     = pucL;
    mt_tD.pucLeftMax  = mt_tD.pucLeft + un32Size - 1;
    mt_tD.pucRight    = NULL;
    mt_tD.pucRightMax = NULL;
  }
  if(mt_tF.un32Frequency != _pcsAudioEnv_->mt_tD.tFormat.un32Frequency)
  {
    mt_tD.un32PosMax  = un32Size << _WSTEP_PREC_;
    mt_tD.un32PosStep = _vuint32(_vfloat(mt_tF.un32Frequency) / _vfloat(_pcsAudioEnv_->mt_tD.tFormat.un32Frequency) * _vfloat(1 << _WSTEP_PREC_));
    if(mt_tF.un8Bits == 16) mt_tD.un32PosStep <<= 1;
  }
  else
  {
    mt_tD.un32PosMax  = 0;
    mt_tD.un32PosStep = 0;
  }
  
  // ..... Assuming that this system is a little-endian (x86) and
  //       since RIFF WAVE is always little-endian, no need to swap the endian
  //if((mt_tF.un8Bits == 16) && (mt_tF.un8Endian != 0)) _SwapEndian_();
    
  // ..... The VortexGE sound mixer always use signed numbers,
  //       make sure that the sound buffer is also signed
  if(mt_tF.un8Sign == 0) _InvertSign_();
  
  // ..... If the primary sound buffer is mono, there is no need for us to have
  //       a stereo sound buffer, so convert it to mono
  if((mt_tF.un8Channels == 1) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 0)) _Stereo2Mono_();

  // ..... If the primary sound buffer is 8-bit, there is no need for us to have
  //       an 16-bit sound buffer, so convert it to 8-bit
  if((mt_tF.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8))
  {  
    if(!_1628_())
    {
      free(mt_tD.pucLeft);
      if(mt_tD.pucRight) free(mt_tD.pucRight);
      memset(&mt_tD, 0, sizeof(mt_tD));
      memset(&mt_tF, 0, sizeof(mt_tF));
      VError(_F_, _L_, "Failed to convert 16-bit to 8-bit");
      return(false);
    }
  }

  // ..... Assign real-time conversion function
  _AssignCnvFunc_(mt_tD.un32PosStep != 0);
 
  // ..... OK !
  return(true);
}
_vbool _CVortexSoundBuffer::LoadVSW(const _vchar *pcFileName)
{
  // ..... Temporary variables
  _TVortexFILE *ptFile;
  _TVSWHeader  tHeader;
  _vuchar      *pucLoad, *pucL, *pucR;
  _vchar       acFile[256];

  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".vsw") == NULL) strcat(acFile, ".vsw");
  ptFile = fopen(acFile, "rb");
  if(ptFile == NULL)
  {
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    

  // ..... Read header
  if(fread(&tHeader, sizeof(tHeader), 1, ptFile) < 1)
  {
    VSetErrorCode(VERR_ReadFile);
    VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
    return(false);
  }
    
  // ..... Check header
  if((tHeader.acID[0] != _VSW_ID1_) || (tHeader.acID[1] != _VSW_ID2_) || (tHeader.acID[2] != _VSW_ID3_))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" is an invalid VIF file", acFile);
    return(false);
  }
  if(tHeader.un8Version != _VSW_VER_)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_IncorrectVersion);
    VError(_F_, _L_, "\"%s\" has an incorrect version", acFile);
    return(false);
  }
  if((tHeader.un8CompL != _VSW_NONE_) && (tHeader.un8CompL != _VSW_DUMMY_))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFileHeader);
    VError(_F_, _L_, "Invalid compression type found in file \"%s\"", acFile);
    return(false);
  }
  if((tHeader.un8CompR != _VSW_NONE_) && (tHeader.un8CompR != _VSW_DUMMY_))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFileHeader);
    VError(_F_, _L_, "Invalid compression type found in file \"%s\"", acFile);
    return(false);
  }
  if(tHeader.un16Reserved1 || tHeader.un32Reserved2)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFileHeader);
    VError(_F_, _L_, "Invalid header data found in file \"%s\"", acFile);
    return(false);
  }

  // ..... Allocate memory for temporary buffer
  if(tHeader.un32SizeCompL > tHeader.un32SizeCompR)
    pucLoad = (_vuchar *) malloc(tHeader.un32SizeCompL);
  else
    pucLoad = (_vuchar *) malloc(tHeader.un32SizeCompR);
  if(pucLoad == NULL)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary sound buffer");
    return(false);
  }
  pucL = (_vuchar *) malloc(tHeader.un32Size);
  if(pucL == NULL)
  {
    free(pucLoad);
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary sound buffer");
    return(false);
  }
  pucR = NULL;
  if(tHeader.tFormat.un8Channels == 2)
  {
    pucR = (_vuchar *) malloc(tHeader.un32Size);
    if(pucR == NULL)
    {
      free(pucLoad);
      free(pucL);
      fclose(ptFile);
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for temporary sound buffer");
      return(false);
    }
  }
  
  // ..... Read & uncompress left / mono channel
  if(fread(pucLoad, tHeader.un32SizeCompL, 1, ptFile) < 1) goto ERRRead;
  if(tHeader.un8CompL == _VSW_DUMMY_)
  {
    if(!_UncompressSoundBuffer_(pucL,  pucLoad))
    {
      free(pucLoad);
      free(pucL);
      if(tHeader.tFormat.un8Channels == 2) free(pucR);
      VSetErrorCode(VERR_UncompressData);
      VError(_F_, _L_, "Failed uncompressing image data");
      return(false);
    }
  }    
  else
    memcpy(pucL, pucLoad, tHeader.un32SizeCompL);

  // ..... Read & uncompress right channel
  if(tHeader.tFormat.un8Channels == 2)
  {
    if(fread(pucLoad, tHeader.un32SizeCompR, 1, ptFile) < 1) goto ERRRead;
    if(tHeader.un8CompR == _VSW_DUMMY_)
    {
      if(!_UncompressSoundBuffer_(pucR,  pucLoad))
      {
        free(pucLoad);
        free(pucL);
        free(pucR);
        VSetErrorCode(VERR_UncompressData);
        VError(_F_, _L_, "Failed uncompressing image data");
        return(false);
      }
    }    
    else
      memcpy(pucR, pucLoad, tHeader.un32SizeCompR);
  }

  // ..... Close file
  fclose(ptFile);

  // ..... Free memory
  free(pucLoad);
  
  // ..... Free previous sound (if any)
  _FreeBuffer_();
  memset(&mt_tD, 0, sizeof(mt_tD));
  _pfnGD_ = NULL;

  // ..... Copy data
  memcpy(&mt_tF, &tHeader.tFormat, sizeof(mt_tF));

  // ..... Calculate some parameters
  if(mt_tF.un8Channels == 2)
  {
    mt_tD.pucLeft     = pucL;
    mt_tD.pucLeftMax  = mt_tD.pucLeft + tHeader.un32Size - 1;
    mt_tD.pucRight    = pucR;
    mt_tD.pucRightMax = mt_tD.pucRight + tHeader.un32Size - 1;
  }
  else
  {
    mt_tD.pucLeft     = pucL;
    mt_tD.pucLeftMax  = mt_tD.pucLeft + tHeader.un32Size - 1;
    mt_tD.pucRight    = NULL;
    mt_tD.pucRightMax = NULL;
  }
  if(mt_tF.un32Frequency != _pcsAudioEnv_->mt_tD.tFormat.un32Frequency)
  {
    mt_tD.un32PosMax  = tHeader.un32Size << _WSTEP_PREC_;
    mt_tD.un32PosStep = _vuint32(_vfloat(mt_tF.un32Frequency) / _vfloat(_pcsAudioEnv_->mt_tD.tFormat.un32Frequency) * _vfloat(1 << _WSTEP_PREC_));
    if(mt_tF.un8Bits == 16) mt_tD.un32PosStep <<= 1;
  }
  else
  {
    mt_tD.un32PosMax  = 0;
    mt_tD.un32PosStep = 0;
  }
  
  // ..... Assuming that this system is a little-endian (x86),
  //       swap the endian if the sound buffer's is 16-bit and not little-endian
  if((mt_tF.un8Bits == 16) && (mt_tF.un8Endian != 0)) _SwapEndian_();
    
  // ..... The VortexGE sound mixer always use signed numbers,
  //       make sure that the sound buffer is also signed
  if(mt_tF.un8Sign == 0) _InvertSign_();
  
  // ..... If the primary sound buffer is mono, there is no need for us to have
  //       a stereo sound buffer, so convert it to mono
  if((mt_tF.un8Channels == 1) && (_pcsAudioEnv_->mt_tD.tFormat.un8Channels == 0)) _Stereo2Mono_();
    
  // ..... If the primary sound buffer is 8-bit, there is no need for us to have
  //       an 16-bit sound buffer, so convert it to 8-bit
  if((mt_tF.un8Bits == 16) && (_pcsAudioEnv_->mt_tD.tFormat.un8Bits == 8))
  {  
    if(!_1628_())
    {
      free(mt_tD.pucLeft);
      if(mt_tD.pucRight) free(mt_tD.pucRight);
      memset(&mt_tD, 0, sizeof(mt_tD));
      memset(&mt_tF, 0, sizeof(mt_tF));
      VError(_F_, _L_, "Failed to convert 16-bit to 8-bit");
      return(false);
    }
  }

  // ..... Assign real-time conversion function
  _AssignCnvFunc_(mt_tD.un32PosStep != 0);
 
  // ..... OK !
  return(true);
  
  // ..... Error reading file
ERRRead:;
  free(pucLoad);
  free(pucL);
  if(tHeader.tFormat.un8Channels == 2) free(pucR);
  fclose(ptFile);
  VSetErrorCode(VERR_ReadFile);
  VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
  return(false);
}
_vbool _CVortexSoundBuffer::SaveVSW(const _vchar *pcFileName)
{
  // ..... Temporary variables
  _TVortexFILE *ptFile;
  _TVSWHeader  tHeader;
  _vuchar      *pucCompL, *pucCompR;
  _vint32      un32SizeCompL, un32SizeCompR, un32Size;
  _vchar       acFile[256];
  
  // ..... Check
  if(mt_tD.pucLeft == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "No data to save (this _CVortexSoundBuffer is not initialized)");
    return(false);
  }
  
  // ..... Allocate memory for temporary buffers
  un32Size = mt_tD.pucLeftMax - mt_tD.pucLeft + 1;
  pucCompL = (_vuchar *) malloc(un32Size);
  if(pucCompL == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary sound buffer");
    return(false);
  }
  pucCompR = NULL;
  if(mt_tF.un8Channels == 2)
  {
    pucCompR = (_vuchar *) malloc(un32Size);
    if(pucCompR == NULL)
    {
      free(pucCompL);
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for temporary sound buffer");
      return(false);
    }
  }

  // ..... Compress sound buffer
  un32SizeCompL = _CompressSoundBuffer_(pucCompL, mt_tD.pucLeft, un32Size);
  un32SizeCompR = 0;
  if(mt_tF.un8Channels == 2)
    un32SizeCompR = _CompressSoundBuffer_(pucCompR, mt_tD.pucRight, un32Size);
  
  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".vsw") == NULL) strcat(acFile, ".vsw");
  ptFile = fopen(acFile, "wb");
  if(ptFile == NULL)
  {
    free(pucCompL);
    if(mt_tF.un8Channels == 2) free(pucCompR);
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    
  
  // ..... Make header
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]    = _VSW_ID1_;
  tHeader.acID[1]    = _VSW_ID2_;
  tHeader.acID[2]    = _VSW_ID3_;
  tHeader.un8Version = _VSW_VER_;
  memcpy(&tHeader.tFormat, &mt_tF, sizeof(mt_tF));
  tHeader.un32Size   = un32Size;
  if(un32SizeCompL)
  {
    tHeader.un8CompL      = _VSW_DUMMY_;
    tHeader.un32SizeCompL = un32SizeCompL;
  }
  else
  {
    tHeader.un8CompL      = _VSW_NONE_;
    tHeader.un32SizeCompL = un32Size;
  }
  if(mt_tF.un8Channels == 2)
  {
    if(un32SizeCompL)
    {
      tHeader.un8CompR      = _VSW_DUMMY_;
      tHeader.un32SizeCompR = un32SizeCompR;
    }
    else
    {
      tHeader.un8CompR      = _VSW_NONE_;
      tHeader.un32SizeCompR = un32Size;
    }
  }
  else
  {
    tHeader.un8CompR      = _VSW_NONE_;
    tHeader.un32SizeCompR = 0;
  }

  // ..... Write header
  if(fwrite(&tHeader, sizeof(tHeader), 1, ptFile) < 1) goto ERRWrite;

  // ..... Write image data
  if(un32SizeCompL)
  {
    if(fwrite(pucCompL, un32SizeCompL, 1, ptFile) < 1) goto ERRWrite;
  }
  else
    if(fwrite(mt_tD.pucLeft, un32Size, 1, ptFile) < 1) goto ERRWrite;
  if(mt_tF.un8Channels == 2)
  {
    if(un32SizeCompR)
    {
      if(fwrite(pucCompR, un32SizeCompR, 1, ptFile) < 1) goto ERRWrite;
    }
    else
      if(fwrite(mt_tD.pucRight, un32Size, 1, ptFile) < 1) goto ERRWrite;
  }
  
  // ..... OK !
  free(pucCompL);
  if(mt_tF.un8Channels == 2) free(pucCompR);
  fclose(ptFile);
  return(true);
  
  // ..... Error writing file
ERRWrite:;
  free(pucCompL);
  if(mt_tF.un8Channels == 2) free(pucCompR);
  fclose(ptFile);
  VSetErrorCode(VERR_WriteFile);
  VError(_F_, _L_, "Failed writing file \"%s\"", acFile);
  return(false);
}

// ============================================================================================
// Local functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Sound functions
// --------------------------------------------------------------------------------------------
// ***** For I/O functions
_vuint32 _CompressSoundBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff, _vuint32 un32SrcBuffSize)
{
  // ..... Not implemented yet
  pucDstBuff      = pucDstBuff;
  pucSrcBuff      = pucSrcBuff;
  un32SrcBuffSize = un32SrcBuffSize;
  return(0);
}
_vuint32 _UncompressSoundBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff)
{
  // ..... Not implemented yet
  pucDstBuff = pucDstBuff;
  pucSrcBuff = pucSrcBuff;
  return(0);
}

// ============================================================================================
// End of vdspsb.cpp
// ============================================================================================
