// ############################################################################################
//  v3denv.cpp
//  Implementation of _CVortex3DEnv class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  Some of the source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
#include <stdlib.h>
// ***** VortexGE headers
#include <v3denv.h>
#include <vmdisp.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** The default render size
#define _XRENDER_ 320
#define _YRENDER_ 240

// ***** The default scalling factor
#define _XSCALE_ 100
#define _YSCALE_ -100

// ============================================================================================
// Exported variables declaration
// ============================================================================================
// ***** The active 3D environment
_CVortex3DEnv *_pcsAct3DEnv_ = NULL;

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3D environment class
// --------------------------------------------------------------------------------------------
// ***** Constructor
_CVortex3DEnv::_CVortex3DEnv()
{
}

// ***** Initialization functions
_vbool _CVortex3DEnv::Initialize(_CVortexVideoEnv *pcsVVideoEnv)
{
  // ..... Check
  if(!pcsVVideoEnv->mt_bAllInitialized)
  {
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "CVortexVideoEnv must be completely initialized before initializing CVortex3DEnv");
    return(false);
  }

  // ..... Copy variables
  mt_tRB = &(pcsVVideoEnv->mt_tRB);
  
  // ..... Initialize the default ambient light
  mt_tDefALight.un8R = mt_tDefALight.un8G = mt_tDefALight.un8B = MAX_COLOR_VALUE / 2;

  // ..... Initialize the default omni light
  mt_tDefOLight.un8R = mt_tDefOLight.un8G = mt_tDefOLight.un8B = MAX_COLOR_VALUE;
  mt_tDefOLight.fX   = mt_tDefOLight.fY   = 1000;
  mt_tDefOLight.fZ   = -1000; 
  
  // ..... Initialize the viewport
  SetZClipping();
  SetViewport();

  // ..... Initialize Z buffer waiting value
  mt_tRB->n32ZWait = _vint32(pow(2, 30 - _ZSTEP_PREC_));
  
  // ..... Initialize LUTs
  _VInitSinCosTable_();
  _VInitLightingTable_();
  SetFogState(16, 16, 16, 8);
  
  // ..... If this is the first 3D environment then set it as the active one
  if(_pcsAct3DEnv_ == NULL) SetAsTheActiveEnv();
 
  // ..... OK !
  return(true);
}

// ***** Environment functions
_vvoid _CVortex3DEnv::SetFogState(_vuint8 un8FogColorR, _vuint8 un8FogColorG, _vuint8 un8FogColorB, _vuint8 un8FogDepth)
{
  // ..... Temporary variables
  _vuint32 un32CF, un32C1, un32C2, un32OC, un32T, un32R, un32G, un32B;
  
  // ..... Set depth
  if(un8FogDepth <= 1)
    mt_un32FogShift = 15;
  else if(un8FogDepth >= 15)
    mt_un32FogShift = 1;
  else
    mt_un32FogShift = 16 - un8FogDepth;
  
  // ..... Generate the LUTS
  for(un32CF = 0; un32CF < 32; un32CF++)
  {
    for(un32C1 = 0; un32C1 < 32; un32C1++)
    {
      for(un32C2 = 0; un32C2 < 32; un32C2++)
      {
        // Generate the fogged-lighting LUTs
        un32T = ((un32C1 * un32C2) / 31) & 0x1F;
        un32R = ((un32CF * _vuint32(un8FogColorR)) + ((31 - un32CF) * un32T)) / 31;
        un32G = ((un32CF * _vuint32(un8FogColorG)) + ((31 - un32CF) * un32T)) / 31;
        un32B = ((un32CF * _vuint32(un8FogColorB)) + ((31 - un32CF) * un32T)) / 31;
        mt_aun32RFLT[(un32CF << 10) | (un32C1 << 5) | un32C2] = un32R << 19;
        mt_aun32GFLT[(un32CF << 10) | (un32C1 << 5) | un32C2] = un32G << 11;
        mt_aun32BFLT[(un32CF << 10) | (un32C1 << 5) | un32C2] = un32B << 3;
        // Generate the fogged-lighting LUTs with transparency
        for(un32OC = 0; un32OC < 32; un32OC++)
        {
          mt_aun8TRFLT[(un32OC << 15) | (un32CF << 10) | (un32C1 << 5) | un32C2] = _vuint8((un32OC + un32R) / 2) << 3;
          mt_aun8TGFLT[(un32OC << 15) | (un32CF << 10) | (un32C1 << 5) | un32C2] = _vuint8((un32OC + un32G) / 2) << 3;
          mt_aun8TBFLT[(un32OC << 15) | (un32CF << 10) | (un32C1 << 5) | un32C2] = _vuint8((un32OC + un32B) / 2) << 3;
        }
      }
      // Generate the fogged-color LUTs
      un32R = ((un32CF * _vuint32(un8FogColorR)) + ((31 - un32CF) * un32C1)) / 31;
      un32G = ((un32CF * _vuint32(un8FogColorG)) + ((31 - un32CF) * un32C1)) / 31;
      un32B = ((un32CF * _vuint32(un8FogColorB)) + ((31 - un32CF) * un32C1)) / 31;
      mt_aun32RFCT[(un32CF << 5) | un32C1] = un32R << 19;
      mt_aun32GFCT[(un32CF << 5) | un32C1] = un32G << 11;
      mt_aun32BFCT[(un32CF << 5) | un32C1] = un32B << 3;
      // Generate the fogged-color LUTs with transparency
      for(un32OC = 0; un32OC < 32; un32OC++)
      {
        mt_aun8TRFCT[(un32OC << 10) | (un32CF << 5) | un32C1] = _vuint8((un32R + un32OC) / 2) << 3;
        mt_aun8TGFCT[(un32OC << 10) | (un32CF << 5) | un32C1] = _vuint8((un32G + un32OC) / 2) << 3;
        mt_aun8TBFCT[(un32OC << 10) | (un32CF << 5) | un32C1] = _vuint8((un32B + un32OC) / 2) << 3;
      }
    }
  } 
}
_vvoid _CVortex3DEnv::SetZClipping(_vuint32 un32MinZ, _vuint32 un32MaxZ)
{
  if(un32MinZ)
    mt_un32MinZ = un32MinZ;
  else
    mt_un32MinZ = 50;
  if(un32MaxZ)
    mt_un32MaxZ = un32MaxZ;
  else
    mt_un32MaxZ = 5000;
}
_vvoid _CVortex3DEnv::SetViewport(_vint32 n32X1, _vint32 n32Y1, _vint32 n32X2, _vint32 n32Y2)
{
  // ..... Calculate the clipping region
  if(n32X1 >= 0)
    mt_n32MinX = _vuint32(n32X1);
  else  
    mt_n32MinX = 2;
  if(n32Y1 >= 0)
    mt_n32MinY = _vuint32(n32Y1);
  else  
    mt_n32MinY = 2;
  if(n32X2 >= 0)
    mt_n32MaxX = _vuint32(n32X2);
  else  
    mt_n32MaxX = _vint32(mt_tRB->un32Width - 2);
  if(n32Y2 >= 0)
    mt_n32MaxY = _vuint32(n32Y2);
  else  
    mt_n32MaxY = _vint32(mt_tRB->un32Height - 2);
  
  // ..... Validate the clipping region
  if(mt_n32MinX < 0) mt_n32MinX = 0;
  if(mt_n32MinY < 0) mt_n32MinY = 0;
  if(_vuint32(mt_n32MaxX) >= mt_tRB->un32Width)  mt_n32MaxX = _vint32(mt_tRB->un32Width - 1);
  if(_vuint32(mt_n32MaxY) >= mt_tRB->un32Height) mt_n32MaxY = _vint32(mt_tRB->un32Height - 1);

  // ..... Calculate projection shift and scale
  mt_fXShift = _vfloat(mt_n32MinX) + _vfloat(mt_n32MaxX - mt_n32MinX) / _vfloat(2.0);
  mt_fYShift = _vfloat(mt_n32MinY) + _vfloat(mt_n32MaxY - mt_n32MinY) / _vfloat(2.0);
  mt_fXScale = _vfloat(_XSCALE_) * _vfloat(mt_n32MaxX - mt_n32MinX) / _vfloat(_XRENDER_);
  mt_fYScale = _vfloat(_YSCALE_) * _vfloat(mt_n32MaxY - mt_n32MinY) / _vfloat(_YRENDER_);
}

// ============================================================================================
// End of v3denv.cpp
// ============================================================================================
