// ############################################################################################
//  v3denv.h
//  Definition & implementation of _CVortex3DEnv class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  Some of the source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _V3DENV_H_
#define _V3DENV_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
#include <time.h>
// ***** VortexGE headers
#include <vluts.h>
#include <vimage.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Fixed point constants
#define _ZSTEP_PREC_ 24
#define _XSTEP_PREC_ 12
#define _CEIL_FRACT_ ((1 << _XSTEP_PREC_) - 1)
#define _ISTEP_PREC_ 16
#define _LSTEP_PREC_ 16
#define _TSTEP_PREC_ _ZSTEP_PREC_
#define _TMASK_      (0xFFFFFFFF << _LSTEP_PREC_)
#define _SSTEP_PREC_ 12

// ***** Subdivision constants
#define _RUNSHIFT_ 5
#define _RUNLEN_   32

// ***** Color macros for VortexGE render buffer
#define EXCLUDED_COLOR_FIRST_PIXEL 0xFFFFFFFF
#define MAX_COLOR_VALUE            31
#define V_RGB(un8R, un8G, un8B) ((_vuint32(un8R) << 19) | (_vuint32(un8G) << 11) | (_vuint32(un8B) << 3))

// ============================================================================================
// Reference to External variables
// ============================================================================================
// ***** The active 3D environment
extern _CVortex3DEnv *_pcsAct3DEnv_;

// ============================================================================================
// Enumerations
// ============================================================================================
// ***** Pivot position
typedef enum _EVP_  { ETopLeft,    ETopCenter,    ETopRight, 
                      ECenterLeft, ECenterCenter, ECenterRight,
                      EBottomLeft, EBottomCenter, EBottomRight } _EVortexPivot;

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Omni light
struct _TVortexOmniLight
{
  _vfloat fX, fY, fZ;       // The omni light position
#ifdef _ASM_SSE_
  _vfloat _fdummy_to_make_the_XYZ_position_128_bits_long_;
#endif
  _vuint8 un8R, un8G, un8B; // The omni light color
};

// ***** Ambient light
struct _TVortexAmbientLight
{
  _vuint8 un8R, un8G, un8B; // The ambient light color
};

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The 3D environment class
class _CVortex3DEnv
{
  protected:
    // Variables
    _TVortexAmbientLight      mt_tDefALight;             // The default ambient light
    _TVortexOmniLight    _sse mt_tDefOLight;             // The default omni light
    _TVortexRenderBuffer      *mt_tRB;                   // The render buffer
    _vint32                   mt_n32MinX;                // The minimum X clipping
    _vint32                   mt_n32MaxX;                // The maximum X clipping
    _vfloat                   mt_fXShift;                // XMin + (XMax - XMin) / 2
    _vfloat                   mt_fXScale;                // The X scale factor
    _vint32                   mt_n32MinY;                // The minimum Y clipping
    _vint32                   mt_n32MaxY;                // The maximum Y clipping
    _vfloat                   mt_fYShift;                // YMin + (YMax - YMin) / 2
    _vfloat                   mt_fYScale;                // The Y scale factor
    _vuint32                  mt_un32MinZ;               // The minimum Z clipping
    _vuint32                  mt_un32MaxZ;               // The maximum Z clipping
    _vuint32                  mt_un32FogShift;           // The fog shifting factor
    _vuint32                  mt_aun32RFLT[32*32*32];    // The R component of the fogged-lighting LUT
    _vuint32                  mt_aun32GFLT[32*32*32];    // The G component of the fogged-lighting LUT
    _vuint32                  mt_aun32BFLT[32*32*32];    // The B component of the fogged-lighting LUT
    _vuint32                  mt_aun32RFCT[32*32];       // The R component of the fogged-color LUT
    _vuint32                  mt_aun32GFCT[32*32];       // The G component of the fogged-color LUT
    _vuint32                  mt_aun32BFCT[32*32];       // The B component of the fogged-color LUT
    _vuint8                   mt_aun8TRFCT[32*32*32];    // The R component of the fogged-color LUT with transparency
    _vuint8                   mt_aun8TGFCT[32*32*32];    // The G component of the fogged-color LUT with transparency
    _vuint8                   mt_aun8TBFCT[32*32*32];    // The B component of the fogged-color LUT with transparency
    _vuint8                   mt_aun8TRFLT[32*32*32*32]; // The R component of the fogged-lighting LUT with transparency
    _vuint8                   mt_aun8TGFLT[32*32*32*32]; // The G component of the fogged-lighting LUT with transparency
    _vuint8                   mt_aun8TBFLT[32*32*32*32]; // The B component of the fogged-lighting LUT with transparency
    
  public:
    // Constructor
    _CVortex3DEnv();
    // Initialization functions
    _vbool Initialize(_CVortexVideoEnv *pcsVVideoEnv);
    // Profiling functions
    inline _vuint64 RDTSC(_vvoid);
    inline _vuint64 GetTicks(_vvoid);
    inline _vfloat  Ticks2mS(_vuint64 un64Ticks);
    inline _vfloat  CalcIntervalmS(_vuint64 un64StartTicks, _vuint64 un64EndTicks);
    inline _vfloat  CalcFPS(_vvoid);
    inline _vfloat  CalcAvgFPS(_vuint64 un64TotFrames, _vuint64 un64StartTicks, _vuint64 un64EndTicks);
    // Environment functions
    inline _vvoid SetAsTheActiveEnv(_vvoid);
    inline _CVortex3DEnv *GetTheActiveEnv(_vvoid);
    inline _vbool IsTheActiveEnv(_vvoid);
    inline _vvoid SetDefaultAmbientLightColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B);
    inline _TVortexAmbientLight *GetDefaultAmbientLight(_vvoid);
    inline _vvoid SetDefaultOmniLightColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B);
    inline _vvoid SetDefaultOmniLightPosition(_vfloat fX, _vfloat fY, _vfloat fZ);
    inline _TVortexOmniLight *GetDefaultOmniLight(_vvoid);
    _vvoid SetFogState(_vuint8 un8FogColorR, _vuint8 un8FogColorG, _vuint8 un8FogColorB, _vuint8 un8FogDepth);
    _vvoid SetZClipping(_vuint32 un32MinZ = 0, _vuint32 un32MaxZ = 0);
    _vvoid SetViewport(_vint32 n32X1 = -1, _vint32 n32Y1 = -1, _vint32 n32X2 = -1, _vint32 n32Y2 = -1);
    // Rendering functions 
    inline _vvoid ClearRenderBuffer(_vvoid);
    inline _vvoid ClearRenderBuffer(_vuint32 un32ColorVRGB);
    inline _vvoid ClearRenderBuffer(_CVortexImage &csImageBackground);
    inline _vvoid ClearRenderBuffer(_CVortexAnimatedImage &csImageBackground);
    inline _vvoid ClearZBuffer(_vvoid);
    inline _vvoid ForceClearZBuffer(_vvoid);
    // Drawing functions (with static image)
    inline _vvoid PutImage(_CVortexImage &csImagePut, _vuint32 un32X, _vint32 un32Y);
    inline _vvoid PutImage(_CVortexImage &csImagePut, _vuint32 un32X, _vint32 un32Y, _vuint32 un32ExcludedColorVRGB);
    inline _vvoid ClipPutImage(_CVortexImage &csImagePut, _vint32 n32X, _vint32 n32Y);
    inline _vvoid ClipPutImage(_CVortexImage &csImagePut, _vint32 n32X, _vint32 n32Y, _vuint32 un32ExcludedColorVRGB);
    // Drawing functions (with animated image)
    inline _vvoid PutImage(_CVortexAnimatedImage &csImagePut, _vuint32 un32X, _vint32 un32Y);
    inline _vvoid PutImage(_CVortexAnimatedImage &csImagePut, _vuint32 un32X, _vint32 un32Y, _vuint32 un32ExcludedColorVRGB);
    inline _vvoid ClipPutImage(_CVortexAnimatedImage &csImagePut, _vint32 n32X, _vint32 n32Y);
    inline _vvoid ClipPutImage(_CVortexAnimatedImage &csImagePut, _vint32 n32X, _vint32 n32Y, _vuint32 un32ExcludedColorVRGB);
    // Special overloaded operators (to support aligned memory allocation for SSE)
    inline _vvoid *operator new      (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid *operator new[]    (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid  operator delete   (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
    inline _vvoid  operator delete[] (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }

  // Friend classes
  friend class _CVortexPanel3D;
  friend class _CVortexUFCeilLine;
  friend class _CVortexUGCeilLine;
  friend class _CVortexTFCeilLine;
  friend class _CVortexTGCeilLine;
  friend class _CVortexFogUFCeilLine;
  friend class _CVortexFogUGCeilLine;
  friend class _CVortexFogTFCeilLine;
  friend class _CVortexFogTGCeilLine;
  friend class _CVortexImageBuilder;
  friend class _CVortexStaticSprite2D;
    
  // Friend functions
  friend _vvoid _UFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep);
  friend _vvoid _UGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                              _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                              _vint32 &n32B, _vint32 &n32BStep);
  friend _vvoid _TFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                              _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep);
  friend _vvoid _TGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                              _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep,
                              _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                              _vint32 &n32B, _vint32 &n32BStep);
  friend _vvoid _FogUFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                                 _vint32 &n32F, _vint32 &n32FStep);
  friend _vvoid _FogUGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                                 _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                                 _vint32 &n32B, _vint32 &n32BStep, _vint32 &n32F, _vint32 &n32FStep);
  friend _vvoid _FogTFClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                                 _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep,
                                 _vint32 &n32F, _vint32 &n32FStep);
  friend _vvoid _FogTGClipHLine_(_vint32 &n32X1, _vint32 &n32X2, _vint32 &n32Z, _vint32 &n32ZStep,
                                 _vint32 &n32U, _vint32 &n32UStep, _vint32 &n32V, _vint32 &n32VStep,
                                 _vint32 &n32R, _vint32 &n32RStep, _vint32 &n32G, _vint32 &n32GStep,
                                 _vint32 &n32B, _vint32 &n32BStep, _vint32 &n32F, _vint32 &n32FStep);
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3D environment class
// --------------------------------------------------------------------------------------------
// Profiling functions
_vuint64 _CVortex3DEnv::RDTSC(_vvoid)
{
  _vuint64 un64Val;
  _vuint32 edi;
  __asm__ __volatile__
  (
    "rdtsc              \n"
    "mov %%eax,  (%%edi)\n"
    "mov %%edx, 4(%%edi)\n"
    : "=D" (edi) : "0" (&un64Val) : "eax", "edx", "memory"
  );
  return(un64Val);
}
_vuint64 _CVortex3DEnv::GetTicks(_vvoid)
{
  return(_vuint64(clock()));
}
_vfloat _CVortex3DEnv::Ticks2mS(_vuint64 un64Ticks)
{
  return(_vfloat(_vdouble(un64Ticks) / _vdouble(CLOCKS_PER_SEC)));
}
_vfloat _CVortex3DEnv::CalcIntervalmS(_vuint64 un64StartTicks, _vuint64 un64EndTicks)
{
  return(_vfloat(_vdouble(un64EndTicks - un64StartTicks) / _vdouble(CLOCKS_PER_SEC)));
}
_vfloat _CVortex3DEnv::CalcFPS(_vvoid)
{
  // ..... Static variables
  static clock_t tClock1   = 0;
  static clock_t tClock2   = 0;
  static _vint32 n32Frames = 0;
  static _vfloat fFPS      = 50;
  
  // ..... Temporary variables
  _vuint64 un64Dif;
  
  // ..... Calculate FPS
  n32Frames++;
  tClock2 = clock();
  un64Dif = tClock2 - tClock1;
  if(un64Dif >= CLOCKS_PER_SEC)
  {
    fFPS      = _vfloat(n32Frames) / _vfloat(un64Dif) * _vfloat(CLOCKS_PER_SEC);
    tClock1   = tClock2;
    n32Frames = 0; 
  }

  // ..... Return the FPS
  return(fFPS);
}
_vfloat _CVortex3DEnv::CalcAvgFPS(_vuint64 un64TotFrames, _vuint64 un64StartTicks, _vuint64 un64EndTicks)
{
  return(_vfloat(_vdouble(un64TotFrames) / 
                 (_vdouble(un64EndTicks - un64StartTicks) / _vdouble(CLOCKS_PER_SEC))));
}

// Environment functions
_vvoid _CVortex3DEnv::SetAsTheActiveEnv(_vvoid)
{
  _pcsAct3DEnv_         = this;
  _pun32RFogLightTable_ = mt_aun32RFLT;
  _pun32GFogLightTable_ = mt_aun32GFLT;
  _pun32BFogLightTable_ = mt_aun32BFLT;
  _pun32RFogColorTable_ = mt_aun32RFCT;
  _pun32GFogColorTable_ = mt_aun32GFCT;
  _pun32BFogColorTable_ = mt_aun32BFCT;
  _pun8TRFogColorTable_ = mt_aun8TRFCT;
  _pun8TGFogColorTable_ = mt_aun8TGFCT;
  _pun8TBFogColorTable_ = mt_aun8TBFCT;
  _pun8TRFogLightTable_ = mt_aun8TRFLT;
  _pun8TGFogLightTable_ = mt_aun8TGFLT;
  _pun8TBFogLightTable_ = mt_aun8TBFLT;
}
_CVortex3DEnv *_CVortex3DEnv::GetTheActiveEnv(_vvoid)
{
  return(_pcsAct3DEnv_);
}
_vbool _CVortex3DEnv::IsTheActiveEnv(_vvoid)
{
  return(_pcsAct3DEnv_ == this);
}
_vvoid _CVortex3DEnv::SetDefaultAmbientLightColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B)
{
  mt_tDefALight.un8R = un8R;
  mt_tDefALight.un8G = un8G;
  mt_tDefALight.un8B = un8B;
}
_TVortexAmbientLight *_CVortex3DEnv::GetDefaultAmbientLight(_vvoid)
{
  return(&mt_tDefALight);
}
_vvoid _CVortex3DEnv::SetDefaultOmniLightColor(_vuint8 un8R, _vuint8 un8G, _vuint8 un8B)
{
  mt_tDefOLight.un8R = un8R;
  mt_tDefOLight.un8G = un8G;
  mt_tDefOLight.un8B = un8B;
}
_vvoid _CVortex3DEnv::SetDefaultOmniLightPosition(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  mt_tDefOLight.fX = fX;
  mt_tDefOLight.fY = fY;
  mt_tDefOLight.fZ = fZ;
}
_TVortexOmniLight *_CVortex3DEnv::GetDefaultOmniLight(_vvoid)
{
  return(&mt_tDefOLight);
}

// ***** Rendering functions 
_vvoid _CVortex3DEnv::ClearRenderBuffer(_vvoid)
{
  ZeroMem4X8((_vuint32 *) mt_tRB->pun8Buffer, mt_tRB->un32Size48);
}
_vvoid _CVortex3DEnv::ClearRenderBuffer(_vuint32 un32ColorVRGB)
{ 
  FillMem4X8((_vuint32 *) mt_tRB->pun8Buffer, un32ColorVRGB, mt_tRB->un32Size48);
}
_vvoid _CVortex3DEnv::ClearRenderBuffer(_CVortexImage &csImageBackground)
{
  CopyMem4X8((_vuint32 *) mt_tRB->pun8Buffer, (_vuint32 *) csImageBackground.mt_pun8Buffer, mt_tRB->un32Size48);
}
_vvoid _CVortex3DEnv::ClearRenderBuffer(_CVortexAnimatedImage &csImageBackground)
{
  ClearRenderBuffer((_CVortexImage &) csImageBackground);
}
_vvoid _CVortex3DEnv::ClearZBuffer(_vvoid)
{
  // ..... Static variables
  static _vint32 n32FrameCount = 0;
  
  // ..... Clear Z buffer if necessary
  if(n32FrameCount >= mt_tRB->n32ZWait)
  {
    n32FrameCount = 0;
    mt_tRB->n32ZTrans  = 0;
    ZeroMem4X8((_vuint32 *) mt_tRB->pn32ZBuffer, mt_tRB->un32Size48);
  }
  else
  {
    n32FrameCount++; 
    mt_tRB->n32ZTrans += (1 << _ZSTEP_PREC_);
  }
}
_vvoid _CVortex3DEnv::ForceClearZBuffer(_vvoid)
{
  mt_tRB->n32ZTrans  = 0;
  ZeroMem4X8((_vuint32 *) mt_tRB->pn32ZBuffer, mt_tRB->un32Size48);
}

// ***** Drawing functions (with static image)
_vvoid _CVortex3DEnv::PutImage(_CVortexImage &csImagePut, _vuint32 un32X, _vint32 un32Y)
{
  // .....Temporary variables
  _vuint32 un32Xc, un32Yc, un32Add;
  _vuint32 *pun32STmp;
  _vuint32 *pun32DTmp;

  // ..... Preparation
  pun32STmp = (_vuint32 *) csImagePut.mt_pun8Buffer;
  pun32DTmp = (_vuint32 *) &mt_tRB->pun8Buffer[(un32Y * mt_tRB->un32Width + un32X) << 2];
  un32Add   = mt_tRB->un32Width - csImagePut.mt_un32Width;

  // ..... Blit
  for(un32Yc = 0; un32Yc < csImagePut.mt_un32Height; un32Yc++)
  {
    for(un32Xc = 0; un32Xc < csImagePut.mt_un32Width; un32Xc++)
      *pun32DTmp++ = *pun32STmp++;
    pun32DTmp += un32Add;
  }
}
_vvoid _CVortex3DEnv::PutImage(_CVortexImage &csImagePut, _vuint32 un32X, _vint32 un32Y, _vuint32 un32ExcludedColorVRGB)
{
  // .....Temporary variables
  _vuint32 un32Xc, un32Yc, un32Add;
  _vuint32 *pun32STmp;
  _vuint32 *pun32DTmp;

  // ..... Preparation
  pun32STmp = (_vuint32 *) csImagePut.mt_pun8Buffer;
  pun32DTmp = (_vuint32 *) &mt_tRB->pun8Buffer[(un32Y * mt_tRB->un32Width + un32X) << 2];
  un32Add   = mt_tRB->un32Width - csImagePut.mt_un32Width;
  if(un32ExcludedColorVRGB == EXCLUDED_COLOR_FIRST_PIXEL) un32ExcludedColorVRGB = *pun32STmp;

  // ..... Blit
  for(un32Yc = 0; un32Yc < csImagePut.mt_un32Height; un32Yc++)
  {
    for(un32Xc = 0; un32Xc < csImagePut.mt_un32Width; un32Xc++)
    {
      if(*pun32STmp != un32ExcludedColorVRGB)
        *pun32DTmp++ = *pun32STmp++;
      else
      {
        pun32DTmp++;
        pun32STmp++;
      }
    }
    pun32DTmp += un32Add;
  }
}
_vvoid _CVortex3DEnv::ClipPutImage(_CVortexImage &csImagePut, _vint32 n32X, _vint32 n32Y)
{
  // .....Temporary variables
  _vuint32 un32Xc, un32Yc, un32Ws, un32Hs;
  _vuint32 *pun32STmp1, *pun32STmp2;
  _vuint32 *pun32DTmp1, *pun32DTmp2;

  // ..... Preparation
  if(n32Y < 0)
  {
    if(n32X < 0)
    {
      pun32STmp1 = (_vuint32 *) &csImagePut.mt_pun8Buffer[(_vuint32(-n32Y) * csImagePut.mt_un32Width + _vuint32(-n32X)) << 2];
      pun32DTmp1 = (_vuint32 *) mt_tRB->pun8Buffer;
      un32Ws     = _vuint32(_vint32(csImagePut.mt_un32Width) + n32X);
      n32X       = 0;
    }
    else
    {
      pun32STmp1 = (_vuint32 *) &csImagePut.mt_pun8Buffer[_vuint32(-n32Y) * csImagePut.mt_un32SizeSL];
      pun32DTmp1 = (_vuint32 *) &mt_tRB->pun8Buffer[n32X << 2];
      un32Ws     = csImagePut.mt_un32Width;
    }
    un32Hs = _vuint32(_vint32(csImagePut.mt_un32Height) + n32Y);
    n32Y   = 0;
  }
  else
  {
    if(n32X < 0)
    {
      pun32STmp1 = (_vuint32 *) &csImagePut.mt_pun8Buffer[_vuint32(-n32X) << 2];
      pun32DTmp1 = (_vuint32 *) &mt_tRB->pun8Buffer[n32Y * mt_tRB->un32SizeSL];
      un32Ws     = _vuint32(_vint32(csImagePut.mt_un32Width) + n32X);
      n32X       = 0;
    }
    else
    {
      pun32STmp1 = (_vuint32 *) csImagePut.mt_pun8Buffer;
      pun32DTmp1 = (_vuint32 *) &mt_tRB->pun8Buffer[(n32Y * mt_tRB->un32Width + n32X) << 2];
      un32Ws     = csImagePut.mt_un32Width;
    }
    un32Hs = csImagePut.mt_un32Height;
  }
    
  // ..... Blit
  for(un32Yc = 0; un32Yc < un32Hs; un32Yc++)
  {
    if((un32Yc + n32Y) >= mt_tRB->un32Height) break;
    pun32STmp2 = pun32STmp1;
    pun32DTmp2 = pun32DTmp1;
    for(un32Xc = 0; un32Xc < un32Ws; un32Xc++)
    {
      if((un32Xc + n32X) >= mt_tRB->un32Width) break;
      *pun32DTmp2++ = *pun32STmp2++;
    }
    pun32STmp1 += csImagePut.mt_un32Width;
    pun32DTmp1 += mt_tRB->un32Width;
  }
}
_vvoid _CVortex3DEnv::ClipPutImage(_CVortexImage &csImagePut, _vint32 n32X, _vint32 n32Y, _vuint32 un32ExcludedColorVRGB)
{
  // .....Temporary variables
  _vuint32 un32Xc, un32Yc, un32Ws, un32Hs;
  _vuint32 *pun32STmp1, *pun32STmp2;
  _vuint32 *pun32DTmp1, *pun32DTmp2;

  // ..... Preparation
  if(n32Y < 0)
  {
    if(n32X < 0)
    {
      pun32STmp1 = (_vuint32 *) &csImagePut.mt_pun8Buffer[(_vuint32(-n32Y) * csImagePut.mt_un32Width + _vuint32(-n32X)) << 2];
      pun32DTmp1 = (_vuint32 *) mt_tRB->pun8Buffer;
      un32Ws     = _vuint32(_vint32(csImagePut.mt_un32Width) + n32X);
      n32X       = 0;
    }
    else
    {
      pun32STmp1 = (_vuint32 *) &csImagePut.mt_pun8Buffer[_vuint32(-n32Y) * csImagePut.mt_un32SizeSL];
      pun32DTmp1 = (_vuint32 *) &mt_tRB->pun8Buffer[n32X << 2];
      un32Ws     = csImagePut.mt_un32Width;
    }
    un32Hs = _vuint32(_vint32(csImagePut.mt_un32Height) + n32Y);
    n32Y   = 0;
  }
  else
  {
    if(n32X < 0)
    {
      pun32STmp1 = (_vuint32 *) &csImagePut.mt_pun8Buffer[_vuint32(-n32X) << 2];
      pun32DTmp1 = (_vuint32 *) &mt_tRB->pun8Buffer[n32Y * mt_tRB->un32SizeSL];
      un32Ws     = _vuint32(_vint32(csImagePut.mt_un32Width) + n32X);
      n32X       = 0;
    }
    else
    {
      pun32STmp1 = (_vuint32 *) csImagePut.mt_pun8Buffer;
      pun32DTmp1 = (_vuint32 *) &mt_tRB->pun8Buffer[(n32Y * mt_tRB->un32Width + n32X) << 2];
      un32Ws     = csImagePut.mt_un32Width;
    }
    un32Hs = csImagePut.mt_un32Height;
  }
  if(un32ExcludedColorVRGB == EXCLUDED_COLOR_FIRST_PIXEL)
    un32ExcludedColorVRGB = *((_vuint32 *) csImagePut.mt_pun8Buffer);
    
  // ..... Blit
  for(un32Yc = 0; un32Yc < un32Hs; un32Yc++)
  {
    if((un32Yc + n32Y) >= mt_tRB->un32Height) break;
    pun32STmp2 = pun32STmp1;
    pun32DTmp2 = pun32DTmp1;
    for(un32Xc = 0; un32Xc < un32Ws; un32Xc++)
    {
      if((un32Xc + n32X) >= mt_tRB->un32Width) break;
      if(*pun32STmp2 != un32ExcludedColorVRGB)
        *pun32DTmp2++ = *pun32STmp2++;
      else
      {
        pun32DTmp2++;
        pun32STmp2++;
      }
    }
    pun32STmp1 += csImagePut.mt_un32Width;
    pun32DTmp1 += mt_tRB->un32Width;
  }
}

// ***** Drawing functions (with animated image)
_vvoid _CVortex3DEnv::PutImage(_CVortexAnimatedImage &csImagePut, _vuint32 un32X, _vint32 un32Y)
{
  PutImage((_CVortexImage &) csImagePut, un32X, un32Y);
}
_vvoid _CVortex3DEnv::PutImage(_CVortexAnimatedImage &csImagePut, _vuint32 un32X, _vint32 un32Y, _vuint32 un32ExcludedColorVRGB)
{
  PutImage((_CVortexImage &) csImagePut, un32X, un32Y, un32ExcludedColorVRGB);
}
_vvoid _CVortex3DEnv::ClipPutImage(_CVortexAnimatedImage &csImagePut, _vint32 n32X, _vint32 n32Y)
{
  ClipPutImage((_CVortexImage &) csImagePut, n32X, n32Y);
}
_vvoid _CVortex3DEnv::ClipPutImage(_CVortexAnimatedImage &csImagePut, _vint32 n32X, _vint32 n32Y, _vuint32 un32ExcludedColorVRGB)
{
  ClipPutImage((_CVortexImage &) csImagePut, n32X, n32Y, un32ExcludedColorVRGB);
}

// ============================================================================================
// Exported inline functions implementation
// ============================================================================================
// ***** Calculate the angle (cosine value) between the panel's normal and the light's vector
inline _vfloat VCalcLightCosValue(const _vfloat *pfPanelNormalXYZ4, const _vfloat *pfPanelPointXYZ4, const _vfloat *pfLightPosXYZ4)
{
  // Looks like after the new and delete operator are overloaded in some classes, this
  // SSE implementation of this function is no longer cause an application to crash.
  //
  // However, if somehow it still causing a crash then just disable it. Also please let me
  // know about it and how to reproduce the crash (if possible), just contact me via e-mail.
  //
#ifdef _ASM_SSE_
  static _vfloat _sse fTmp;
  _vuint32 ebx;
  __asm__ __volatile__
  (
    // esi -> panel's point    : P[? z y x] 
    // edi -> panel's normal   : N[? z y x] 
    // edx -> light's position : L[? z y x]    
    // ebx -> the cosine value
    "movaps (%%esi), %%xmm0         \n" // xmm0 = P[? z y x]                           (panel's point)
    "movaps (%%edx), %%xmm1         \n" // xmm1 = L[? z y x]                           (light's position)
    "subps   %%xmm0, %%xmm1         \n" // xmm1 = L[? z y x] - P[? z y x] = V[? z y x] (the light vector)
    "movaps  %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x]
    "mulps   %%xmm0, %%xmm1         \n" // xmm1 = V[? z y x] ^ 2          = S[? z y x] (the squared V)
    "movaps  %%xmm1, %%xmm2         \n" // xmm2 = S[? z y x]
    "shufps   $0x55, %%xmm2, %%xmm2 \n" // xmm2 = S[? ? ? y]
    "movaps  %%xmm1, %%xmm3         \n" // xmm3 = S[? z y x]
    "shufps   $0xAA, %%xmm3, %%xmm3 \n" // xmm3 = S[? ? ? z]
    "addss   %%xmm2, %%xmm1         \n" // xmm1 = S[? ? ? x+y]
    "addss   %%xmm3, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z]
    "rsqrtss %%xmm1, %%xmm1         \n" // xmm1 = S[? ? ? x+y+z] ^ 0.5    = M[? ? ? m] (the magnitude of V)
    "shufps   $0x00, %%xmm1, %%xmm1 \n" // xmm1 = M[? m m m]
    "mulps   %%xmm1, %%xmm0         \n" // xmm0 = V[? z y x] / M[? m m m] = R[? z y x] (the normalized V)
    "mulps  (%%edi), %%xmm0         \n" // xmm0 = R[? z y x] * N[? z y x] = C[? z y x] (the cosine component)
    "movaps  %%xmm0, %%xmm1         \n" // xmm1 = C[? z y x]
    "shufps   $0xFD, %%xmm1, %%xmm1 \n" // xmm1 = C[? ? ? y]
    "movaps  %%xmm0, %%xmm2         \n" // xmm2 = C[? z y x]
    "shufps   $0xFE, %%xmm2, %%xmm2 \n" // xmm2 = C[? ? ? z]
    "addss   %%xmm1, %%xmm0         \n" // xmm0 = C[? ? ? x+y]
    "addss   %%xmm2, %%xmm0         \n" // xmm0 = C[? ? ? x+y+z]          = I[? ? ? i] (the cosine value)
    "xorps   %%xmm4, %%xmm4         \n" // zero the xmm4
    "maxss   %%xmm4, %%xmm0         \n" // make sure the cosine value >= 0
    "movss   %%xmm0, (%%ebx)        \n" // save the final result
    : "=b" (ebx)                                                                          \
    :  "0" (&fTmp), "S" (pfPanelPointXYZ4), "D" (pfPanelNormalXYZ4), "d" (pfLightPosXYZ4) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "memory"
  );
  return(fTmp);
#else
  // ..... Temporary variables
  _vfloat afT[4], fMag, fCos;
  
  // ..... Create the normalized light vector from the panel center to the omni light position
  afT[0] = pfLightPosXYZ4[0] - pfPanelPointXYZ4[0];
  afT[1] = pfLightPosXYZ4[1] - pfPanelPointXYZ4[1];
  afT[2] = pfLightPosXYZ4[2] - pfPanelPointXYZ4[2];
  fMag = sqrt(afT[0] * afT[0] + afT[1] * afT[1] + afT[2] * afT[2]);
  afT[0] /= fMag; afT[1] /= fMag; afT[2] /= fMag;
    
  // ..... Return the angle (cosine value) between panel's normal and the light vector
  fCos= pfPanelNormalXYZ4[0] * afT[0] + pfPanelNormalXYZ4[1] * afT[1] + pfPanelNormalXYZ4[2] * afT[2];
  fCos = (fCos < 0) ? 0 : fCos;
  return(fCos);
#endif
}

#endif
// ============================================================================================
// End of v3denv.h
// ============================================================================================
