// ############################################################################################
//  vfastmem.h
//  Implementation of fast-memory-manipulation functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VFASTMEM_H_
#define _VFASTMEM_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vbtypes.h>

// ============================================================================================
// Exported functions prototype
// ============================================================================================
// ***** Aligned memory allocation
extern _vvoid *amalloc(_vuint32 un32Size, _vuint32 un32Alignment = 16);
extern _vvoid *agetrealaddress(_vvoid *pvBuffer);
extern _vvoid afree(_vvoid *pvBuffer);

// ============================================================================================
// Exported inline functions implementation
// ============================================================================================
// ***** Aligned memory allocation for SSE
inline _vvoid *ssemalloc(_vuint32 un32Size)
{
#ifdef _ASM_SSE_
  return(amalloc(un32Size, 16));
#else
  return(malloc(un32Size));
#endif
}
inline _vvoid *ssegetrealaddress(_vvoid *pvBuffer)
{
#ifdef _ASM_SSE_
  return(agetrealaddress(pvBuffer));
#else
  return(pvBuffer);
#endif
}
inline _vvoid ssefree(_vvoid *pvBuffer)
{
#ifdef _ASM_SSE_
  afree(pvBuffer);
#else
  free(pvBuffer);
#endif
}

// ***** Memory operations using inline assembly with SSE
#ifdef _ASM_SSE_
inline _vvoid ZeroMem4X8(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  __asm__ __volatile__
  (
#ifdef _ALIGN_BUFF_MEM_  
    "xorps %%xmm0, %%xmm0        \n"
    "1:                          \n"
    "  movntps %%xmm0,   (%%edi) \n"
    "  movntps %%xmm0, 16(%%edi) \n"
    "  add     $32, %%edi        \n"
    "loop 1b                     \n"
#else
    "pxor %%mm0, %%mm0         \n"
    "1:                        \n"
    "  movntq %%mm0,   (%%edi) \n"
    "  movntq %%mm0,  8(%%edi) \n"
    "  movntq %%mm0, 16(%%edi) \n"
    "  movntq %%mm0, 24(%%edi) \n"
    "  add    $32, %%edi       \n"
    "loop 1b                   \n"
    "emms                      \n"
#endif
    : "=&D" (edi)                                          \
    :   "0" (pun32Buffer), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}
inline _vvoid FillMem4X8(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  _vuint64 un64Value = (_vuint64(un32Value) << 32) | _vuint64(un32Value);
  __asm__ __volatile__
  (
#ifdef _ALIGN_BUFF_MEM_  
    "movhps (%%ebx), %%xmm0      \n"
    "movlps (%%ebx), %%xmm0      \n"
    "1:                          \n"
    "  movntps %%xmm0,   (%%edi) \n"
    "  movntps %%xmm0, 16(%%edi) \n"
    "  add     $32, %%edi        \n"
    "loop 1b                     \n"
#else
    "movq (%%ebx), %%mm0       \n"
    "1:                        \n"
    "  movntq %%mm0,   (%%edi) \n"
    "  movntq %%mm0,  8(%%edi) \n"
    "  movntq %%mm0, 16(%%edi) \n"
    "  movntq %%mm0, 24(%%edi) \n"
    "  add    $32, %%edi       \n"
    "loop 1b                   \n"
    "emms                      \n"
#endif
    : "=&D" (edi)                                                            \
    :   "0" (pun32Buffer), "b" (&un64Value), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}
inline _vvoid CopyMem4X8(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  __asm__ __volatile__
  (
#ifdef _ALIGN_BUFF_MEM_  
    "prefetchnta   (%%esi)       \n"
    "prefetchnta 32(%%esi)       \n"
    "prefetchnta 64(%%esi)       \n"
    "1:                          \n"
    "  prefetchnta 96(%%esi)     \n"
    "  movaps    (%%esi), %%xmm0 \n"
    "  movaps  16(%%esi), %%xmm1 \n"
    "  add     $32, %%esi        \n"
    "  movntps %%xmm0,   (%%edi) \n"
    "  movntps %%xmm1, 16(%%edi) \n"
    "  add     $32, %%edi        \n"
    "loop 1b                     \n"
#else
    "prefetchnta   (%%esi)     \n"
    "prefetchnta 32(%%esi)     \n"
    "prefetchnta 64(%%esi)     \n"
    "1:                        \n"
    "  prefetchnta 96(%%esi)   \n"
    "  movq     (%%esi), %%mm0 \n"
    "  movq    8(%%esi), %%mm1 \n"
    "  movq   16(%%esi), %%mm2 \n"
    "  movq   24(%%esi), %%mm3 \n"
    "  add    $32, %%esi       \n"
    "  movntq %%mm0,   (%%edi) \n"
    "  movntq %%mm1,  8(%%edi) \n"
    "  movntq %%mm2, 16(%%edi) \n"
    "  movntq %%mm3, 24(%%edi) \n"
    "  add    $32, %%edi       \n"
    "loop 1b                   \n"
    "emms                      \n"
#endif
    : "=&D" (edi)                                                                   \
    :   "0" (pun32DstBuffer), "S" (pun32SrcBuffer), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}

// ***** Memory operations using inline assembly with MMX
//       On Pentium 4 this will even slower than the plain x86 inline assembly version
#elif defined(_ASM_MMX_)
inline _vvoid ZeroMem4X8(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  __asm__ __volatile__
  (
    "pxor %%mm0, %%mm0       \n"
    "1:                      \n"
    "  movq %%mm0,   (%%edi) \n"
    "  movq %%mm0,  8(%%edi) \n"
    "  movq %%mm0, 16(%%edi) \n"
    "  movq %%mm0, 24(%%edi) \n"
    "  add  $32, %%edi       \n"
    "loop 1b                 \n"
    "emms                    \n"
    : "=&D" (edi)                                          \
    :   "0" (pun32Buffer), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}
inline _vvoid FillMem4X8(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  _vuint64 un64Value = (_vuint64(un32Value) << 32) | _vuint64(un32Value);
  __asm__ __volatile__
  (
    "movq (%%ebx), %%mm0     \n"
    "1:                      \n"
    "  movq %%mm0,   (%%edi) \n"
    "  movq %%mm0,  8(%%edi) \n"
    "  movq %%mm0, 16(%%edi) \n"
    "  movq %%mm0, 24(%%edi) \n"
    "  add  $32, %%edi       \n"
    "loop 1b                 \n"
    "emms                    \n"
    : "=&D" (edi)                                                            \
    :   "0" (pun32Buffer), "b" (&un64Value), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}
inline _vvoid CopyMem4X8(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  __asm__ __volatile__
  (
    "movq   (%%esi), %%mm0   \n"
    "movq  8(%%esi), %%mm1   \n"
    "movq 16(%%esi), %%mm2   \n"
    "movq 24(%%esi), %%mm3   \n"
    "add  $32, %%esi         \n"
    "1:                      \n"
    "  movq %%mm0,   (%%edi) \n"
    "  movq %%mm1,  8(%%edi) \n"
    "  movq %%mm2, 16(%%edi) \n"
    "  movq %%mm3, 24(%%edi) \n"
    "  add  $32, %%edi       \n"
    "  movq   (%%esi), %%mm0 \n"
    "  movq  8(%%esi), %%mm1 \n"
    "  movq 16(%%esi), %%mm2 \n"
    "  movq 24(%%esi), %%mm3 \n"
    "  add  $32, %%esi       \n"
    "loop 1b                 \n"
    "emms                    \n"
    : "=&D" (edi)                                                                   \
    :   "0" (pun32DstBuffer), "S" (pun32SrcBuffer), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}

// ***** Memory operations using plain x86 inline assembly
#elif defined(_ASM_ASM_)
inline _vvoid ZeroMem4X8(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  __asm__ __volatile__
  (
    "xor %%eax, %%eax \n"
    "shl $3, %%ecx    \n"
    "cld              \n"
    "rep stosl        \n"
    : "=&D" (edi)                                          \
    :   "0" (pun32Buffer), "c" (un32BufferSizeDividedBy32) \
    : "eax", "memory"
  );
}
inline _vvoid FillMem4X8(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  __asm__ __volatile__
  (
    "shl $3, %%ecx    \n"
    "cld              \n"
    "rep stosl        \n"
    : "=&D" (edi)                                                           \
    :   "0" (pun32Buffer), "a" (un32Value), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}
inline _vvoid CopyMem4X8(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 edi;
  __asm__ __volatile__
  (
    "shl $3, %%ecx \n"
    "cld           \n"
    "rep movsl     \n"
    : "=&D" (edi)                                                                   \
    :   "0" (pun32DstBuffer), "S" (pun32SrcBuffer), "c" (un32BufferSizeDividedBy32) \
    : "memory"
  );
}

// ***** Memory operations using plain C with QWORD memory transfer
#elif defined(__USE_QW_MEMOPS__)
inline _vvoid ZeroMem4X8(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint64 *pun64Tmp = (_vuint64*) pun32Buffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun64Tmp++ = 0; *pun64Tmp++ = 0; *pun64Tmp++ = 0; *pun64Tmp++ = 0;
  }
}
inline _vvoid FillMem4X8(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint64 *pun64Tmp = (_vuint64 *) pun32Buffer;
  _vuint64 un64Value = (_vuint64(un32Value) << 32) | _vuint64(un32Value);
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun64Tmp++ = un64Value; *pun64Tmp++ = un64Value; *pun64Tmp++ = un64Value; *pun64Tmp++ = un64Value;
  }
}
inline _vvoid CopyMem4X8(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint64 *pun64Dst = (_vuint64 *) pun32DstBuffer;
  _vuint64 *pun64Src = (_vuint64 *) pun32SrcBuffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun64Dst++ = *pun64Src++; *pun64Dst++ = *pun64Src++; *pun64Dst++ = *pun64Src++; *pun64Dst++ = *pun64Src++;
  }
}

// ***** Memory operations using plain C with DWORD memory transfer
#else
inline _vvoid ZeroMem4X8(_vuint32 *pun32Buffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 *pun32Tmp = pun32Buffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0;
    *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0; *pun32Tmp++ = 0;
  }
}
inline _vvoid FillMem4X8(_vuint32 *pun32Buffer, _vuint32 un32Value, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 *pun32Tmp = pun32Buffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value;
    *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value; *pun32Tmp++ = un32Value;
  }
}
inline _vvoid CopyMem4X8(_vuint32 *pun32DstBuffer, _vuint32 *pun32SrcBuffer, _vuint32 un32BufferSizeDividedBy32)
{
  _vuint32 *pun32Dst = pun32DstBuffer;
  _vuint32 *pun32Src = pun32SrcBuffer;
  for(_vuint32 un32Cnt = 0; un32Cnt < un32BufferSizeDividedBy32; un32Cnt++)
  {
    *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++;
    *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++; *pun32Dst++ = *pun32Src++;
  }
}
#endif

#endif
// ============================================================================================
// End of vfastmem.h
// ============================================================================================
