// ############################################################################################
//  vhnode.h
//  Definition & implementation of _CVortexHNode class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VHNODE_H_
#define _VHNODE_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vmat3d.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Maximum children allowed
#define _MAX_CHILD_ 16

// ############################################################################################
// ##################################  W A R N I N G  ! ! !  ##################################
// ############################################################################################
// ###### - NEVER delete any child's node manually because all children will be deleted  ######
// ######   automatically when their parent is deleted.                                  ######
// ###### - If you need to preserve a child's node then change its parent to NULL before ######
// ######   deleting the parent.                                                         ######
// ############################################################################################
// ##################################  W A R N I N G  ! ! !  ##################################
// ############################################################################################
// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The hierarchical node class
class _CVortexHNode
{
  protected:
    // Variables
    struct _TMatrixData
    {
      _vfloat               afScl[3];   // Dynamic scalling factor XYZ
      _vint32               an32Rot[3]; // Dynamic rotation factor XYZ
      _vfloat               afPos[3];   // Dynamic position factor XYZ
      _CVortexMatrix3D _sse csStc;      // Static transformation matrix 
      _CVortexMatrix3D _sse csTot;      // Total transformation matrix
    } _sse mt_tM;
    _CVortexHNode *mt_pcsParent;              // Pointer to parent node
    _CVortexHNode *mt_apcsChild[_MAX_CHILD_]; // Pointer to child nodes

  public:
    // Constructor & destructor
    _CVortexHNode();
    ~_CVortexHNode();
    // Hierarchy constructions
    _vbool SetParent(_CVortexHNode *pcsNewParent);
    _CVortexHNode *AddChild(_vvoid);
    _vvoid DelChild(_CVortexHNode *pcsChild);
    _vvoid DelAllChild(_vvoid);
    // Dynamic matrix operations
    inline _CVortexMatrix3D &Matrix(_vvoid);
    inline _vvoid UpdateMatrix(_vbool bUpdateChildren = true);
    // Static matrix operations
    inline _vvoid StcIdentity(_vvoid);
    inline _vvoid StcShear(_vfloat fX, _vfloat fY);
    inline _vvoid StcScale(_vfloat fX, _vfloat fY, _vfloat fZ);
    inline _vvoid StcRotate(_vint32 n32X, _vint32 n32Y, _vint32 n32Z);
    inline _vvoid StcTranslate(_vfloat fX, _vfloat fY, _vfloat fZ);
    inline _vvoid StcSet(_vfloat const aafNewMatrix[4][4]);
    inline _vvoid StcSet(_CVortexMatrix3D &csVNewMatrix);
    // Dynamic matrix operations
    inline _vvoid DynSetIdentity(_vvoid);
    inline _vvoid DynSetScalling(_vfloat fX, _vfloat fY, _vfloat fZ);
    inline _vvoid DynSetRotation(_vint32 n32X, _vint32 n32Y, _vint32 n32Z);
    inline _vvoid DynSetPosition(_vfloat fX, _vfloat fY, _vfloat fZ);
    inline _vvoid DynScale(_vfloat fX, _vfloat fY, _vfloat fZ);
    inline _vvoid DynRotate(_vint32 n32X, _vint32 n32Y, _vint32 n32Z);
    inline _vvoid DynRotateX(_vint32 n32X);
    inline _vvoid DynRotateY(_vint32 n32Y);
    inline _vvoid DynRotateZ(_vint32 n32Z);
    inline _vvoid DynTranslate(_vfloat fX, _vfloat fY, _vfloat fZ);
    // Special overloaded operators (to support aligned memory allocation for SSE)
    inline _vvoid *operator new      (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid *operator new[]    (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid  operator delete   (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
    inline _vvoid  operator delete[] (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The hierarchical node class
// --------------------------------------------------------------------------------------------
// ***** Dynamic matrix operations
_CVortexMatrix3D &_CVortexHNode::Matrix(_vvoid)
{
  return(mt_tM.csTot);
}
_vvoid _CVortexHNode::UpdateMatrix(_vbool bUpdateChildren)
{
  // ..... TO DO : Need optimization using cache matrix so that not all parts are updated
  //               (only necessary parts should be updated)
  
  // ..... Update myself
  mt_tM.csTot.SetMatrix(mt_tM.csStc);
  mt_tM.csTot.Scale(mt_tM.afScl[0], mt_tM.afScl[1], mt_tM.afScl[2]);
  mt_tM.csTot.Rotate(mt_tM.an32Rot[0], mt_tM.an32Rot[1], mt_tM.an32Rot[2]);
  mt_tM.csTot.Translate(mt_tM.afPos[0], mt_tM.afPos[1], mt_tM.afPos[2]);
  if(mt_pcsParent) mt_tM.csTot.MultMatrixCCN(mt_pcsParent->mt_tM.csTot);
    
  // ..... Update children
  if(bUpdateChildren)
  {
    for(_vuint8 un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
      if(mt_apcsChild[un8Cnt]) mt_apcsChild[un8Cnt]->UpdateMatrix(true);
  }
}

// ***** Static matrix operations
_vvoid _CVortexHNode::StcIdentity(_vvoid)
{
  mt_tM.csStc.Identity();
}
_vvoid _CVortexHNode::StcShear(_vfloat fX, _vfloat fY)
{
  mt_tM.csStc.Shear(fX, fY);
}
_vvoid _CVortexHNode::StcScale(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  mt_tM.csStc.Scale(fX, fY, fZ);
}
_vvoid _CVortexHNode::StcRotate(_vint32 n32X, _vint32 n32Y, _vint32 n32Z)
{
  mt_tM.csStc.Rotate(n32X, n32Y, n32Z);
}
_vvoid _CVortexHNode::StcTranslate(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  mt_tM.csStc.Translate(fX, fY, fZ);
}
_vvoid _CVortexHNode::StcSet(_vfloat const aafNewMatrix[4][4])
{
  mt_tM.csStc.SetMatrix(aafNewMatrix);
}
_vvoid _CVortexHNode::StcSet(_CVortexMatrix3D &csVNewMatrix)
{
  mt_tM.csStc.SetMatrix(csVNewMatrix);
}

// ***** Dynamic matrix operations
_vvoid _CVortexHNode::DynSetIdentity(_vvoid)
{
  mt_tM.afScl[0]  = 1;  mt_tM.afScl[1]  = 1;  mt_tM.afScl[2]  = 1;
  mt_tM.an32Rot[0] = 0; mt_tM.an32Rot[1] = 0; mt_tM.an32Rot[2] = 0;
  mt_tM.afPos[0]  = 0;  mt_tM.afPos[1]  = 0;  mt_tM.afPos[2]  = 0;
}
_vvoid _CVortexHNode::DynSetScalling(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  mt_tM.afScl[0] = fX; mt_tM.afScl[1] = fY; mt_tM.afScl[2] = fZ;
}
_vvoid _CVortexHNode::DynSetRotation(_vint32 n32X, _vint32 n32Y, _vint32 n32Z)
{
  mt_tM.an32Rot[0] = n32X; mt_tM.an32Rot[1] = n32Y; mt_tM.an32Rot[2] = n32Z;
}
_vvoid _CVortexHNode::DynSetPosition(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  mt_tM.afPos[0] = fX; mt_tM.afPos[1] = fY; mt_tM.afPos[2] = fZ;
}
_vvoid _CVortexHNode::DynScale(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  mt_tM.afScl[0] *= fX; mt_tM.afScl[1] *= fY; mt_tM.afScl[2] *= fZ;
}
_vvoid _CVortexHNode::DynRotate(_vint32 n32X, _vint32 n32Y, _vint32 n32Z)
{
  mt_tM.an32Rot[0] += n32X; mt_tM.an32Rot[1] += n32Y; mt_tM.an32Rot[2] += n32Z;
}
_vvoid _CVortexHNode::DynRotateX(_vint32 n32X)
{
  mt_tM.an32Rot[0] += n32X; 
}
_vvoid _CVortexHNode::DynRotateY(_vint32 n32Y)
{
  mt_tM.an32Rot[1] += n32Y; 
}
_vvoid _CVortexHNode::DynRotateZ(_vint32 n32Z)
{
  mt_tM.an32Rot[2] += n32Z; 
}
_vvoid _CVortexHNode::DynTranslate(_vfloat fX, _vfloat fY, _vfloat fZ)
{
  mt_tM.afPos[0] += fX; mt_tM.afPos[1] += fY; mt_tM.afPos[2] += fZ;
}

#endif
// ============================================================================================
// End of vhnode.h
// ============================================================================================
