// ############################################################################################
//  vimage.cpp
//  Implementation  of _CVortexTexture, _CVortexAnimatedTexture, _CVortexImage,
//  _CVortexAnimatedImage, and _CVortexImageBuilder class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
#include <stdlib.h>
// ***** VortexGE headers
#include <vbmp.h>
#include <v3denv.h>
#include <vimage.h>
#include <vmdisp.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** VIB header
#define _VIB_ID1_ 'V'
#define _VIB_ID2_ 'I'
#define _VIB_ID3_ 'B'
#define _VIB_VER_ 100

// ***** Compression type
#define _VIB_NONE_  0
#define _VIB_DUMMY_ 1

// ============================================================================================
// Local functions prototype
// ============================================================================================
// ***** Image functions
_vuint32 _PackImage_(_vuchar *pucDstBuff, _vuchar*pucSrcBuff, _vuint32 un32SrcSize);
_vbool   _UnpackImage_(_vuchar *pucDstBuff, _vuchar *pucSrcBuff, _vuint32 un32DstSize, _vuint32 un32SrcSize);
_vbool   _GetByte_(_vuchar &pucDst, _vuchar *&pucSrc, _vuchar *&pucSrcMax);
_vvoid   _PackBytes_(_vuchar aucDst[5], const _vuchar aucSrc[8]);
_vvoid   _UnpackBytes_(_vuchar aucDst[8], const _vuchar aucSrc[5]);
_vuint32 _CompressImageBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff, _vuint32 un32SrcBuffSize);
_vuint32 _UncompressImageBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff);

// ============================================================================================
// Structures
// ============================================================================================
// ***** VIB header
struct _TVIBHeader
{ 
  _vchar   acID[3];            // File ID, must be "VIB"
  _vuint8  un8Version;         // File version
  _vuint32 un32ImageWidth;     // Image width in pixels
  _vuint32 un32ImageHeight;    // Image height in pixels
  _vuint8  un8Compression;     // Compression type
  _vuint32 un32SizeCompressed; // Compressed image size in bytes
  _vuint32 un32SizePacked;     // Packed image size in bytes
  _vuint8  un8Reserved1;       // Reserved, must be zero
  _vuint32 un32Reserved2;      // Reserved, must be zero
} __attribute__((packed));

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The image builder class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexImageBuilder::_CVortexImageBuilder()
{
  mt_ptBuffer   = NULL;
  mt_un32Width  = 0;
  mt_un32Height = 0;
}
_CVortexImageBuilder::~_CVortexImageBuilder()
{
  if(mt_ptBuffer) free(mt_ptBuffer);
}

// ***** Image functions
_vbool _CVortexImageBuilder::SmearEffect(_vuint8 un8Radius)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptNew;
  _vuint32       un32CntX, un32CntY, un32PosD, un32PosS;
  _vint32        n32CntX, n32CntY;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Uninitialized _CVortexImageBuilder");
    return(false);
  }
  if(un8Radius == 0)
  { 
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Radius must not zero");
    return(false);
  }
  
  // ..... Allocate memory for new image
  ptNew = (_TVortexUN8RGB *) malloc(sizeof(_TVortexUN8RGB) * mt_un32Width * mt_un32Height);
  if(ptNew == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
    return(false);
  }

  // ..... Smear image
  for(un32CntY = 0; un32CntY < mt_un32Height; un32CntY++)
  {
    un32PosD = un32CntY * mt_un32Width;
    for(un32CntX = 0; un32CntX < mt_un32Width; un32CntX++)
    {
      // Calculate position
      n32CntX = _vint32(un32CntX) + _vint32(_vfloat(random()) / _vfloat(RAND_MAX) * _vfloat(un8Radius * 2) - _vfloat(un8Radius));
      n32CntY = _vint32(un32CntY) + _vint32(_vfloat(random()) / _vfloat(RAND_MAX) * _vfloat(un8Radius * 2) - _vfloat(un8Radius));
      // Copy the color into buffer
      if((n32CntY >= 0) && (n32CntX >= 0) && 
         (n32CntY < _vint32(mt_un32Height)) && (n32CntX < _vint32(mt_un32Width)))
      {
        un32PosS = _vuint32(n32CntY) * mt_un32Width;
        ptNew[un32PosD + un32CntX].un8R = mt_ptBuffer[un32PosS + n32CntX].un8R;
        ptNew[un32PosD + un32CntX].un8G = mt_ptBuffer[un32PosS + n32CntX].un8G;
        ptNew[un32PosD + un32CntX].un8B = mt_ptBuffer[un32PosS + n32CntX].un8B;
      }
      else
      {
        un32PosS = _vuint32(un32CntY) * mt_un32Width;
        ptNew[un32PosD + un32CntX].un8R = mt_ptBuffer[un32PosS + un32CntX].un8R;
        ptNew[un32PosD + un32CntX].un8G = mt_ptBuffer[un32PosS + un32CntX].un8G;
        ptNew[un32PosD + un32CntX].un8B = mt_ptBuffer[un32PosS + un32CntX].un8B;
      }
    }
  }
    
  // ..... Change pointer
  free(mt_ptBuffer);
  mt_ptBuffer = ptNew;

  // ..... OK !
  return(true);
}
_vbool _CVortexImageBuilder::BlurEffect(_vuint8 un8Radius)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptNew;
  _vuint32       un32CntX, un32CntY, un32R, un32G, un32B, un32Tot, un32PosD, un32PosS;
  _vint32        n32CntX, n32CntY;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Uninitialized _CVortexImageBuilder");
    return(false);
  }
  if(un8Radius == 0)
  { 
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Radius must not zero");
    return(false);
  }
  
  // ..... Allocate memory for new image
  ptNew = (_TVortexUN8RGB *) malloc(sizeof(_TVortexUN8RGB) * mt_un32Width * mt_un32Height);
  if(ptNew == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
    return(false);
  }

  // ..... Blur image
  for(un32CntY = 0; un32CntY < mt_un32Height; un32CntY++)
  {
    un32PosD = un32CntY * mt_un32Width;
    for(un32CntX = 0; un32CntX < mt_un32Width; un32CntX++)
    {
      // Grab pixels
      un32R = un32G = un32B = un32Tot = 0;
      for(n32CntY = _vint32(un32CntY) - _vint32(un8Radius); n32CntY <= _vint32(un32CntY + un8Radius); n32CntY++)
      {
        un32PosS = _vuint32(n32CntY) * mt_un32Width;
        for(n32CntX = _vint32(un32CntX) - _vint32(un8Radius); n32CntX <= _vint32(un32CntX + un8Radius); n32CntX++)
        {
          if((n32CntY >= 0) && (n32CntX >= 0) && 
             (n32CntY < _vint32(mt_un32Height)) && (n32CntX < _vint32(mt_un32Width)))
          {
            un32R += mt_ptBuffer[un32PosS + n32CntX].un8R;
            un32G += mt_ptBuffer[un32PosS + n32CntX].un8G;
            un32B += mt_ptBuffer[un32PosS + n32CntX].un8B;
            un32Tot++;
          }
        }
      }
      // Put the mean colors into buffer
      ptNew[un32PosD + un32CntX].un8R = _vuint8(un32R / un32Tot);
      ptNew[un32PosD + un32CntX].un8G = _vuint8(un32G / un32Tot);
      ptNew[un32PosD + un32CntX].un8B = _vuint8(un32B / un32Tot);
    }
  }

  // ..... Change pointer
  free(mt_ptBuffer);
  mt_ptBuffer = ptNew;
    
  // ..... OK !
  return(true);
}
_vbool _CVortexImageBuilder::OilPaintingEffect(_vuint8 un8Factor)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptNew;
  _vuint32       un32CntX, un32CntY, un32PosD, un32PosS;
  _vint32        n32CntX, n32CntY;
  _vuint8        un8R[128], un8G[128], un8B[128], un8C[128], un8Pos, un8Tot, un8Cnt;
  _vbool         bF;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Uninitialized _CVortexImageBuilder");
    return(false);
  }
  if(un8Factor == 0)
  { 
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Factor must not zero");
    return(false);
  }
  if(un8Factor > 5)
  { 
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Maximum factor is 5");
    return(false);
  }
  
  // ..... Allocate memory for new image
  ptNew = (_TVortexUN8RGB *) malloc(sizeof(_TVortexUN8RGB) * mt_un32Width * mt_un32Height);
  if(ptNew == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
    return(false);
  }

  // ..... Perform the effect
  for(un32CntY = 0; un32CntY < mt_un32Height; un32CntY++)
  {
    un32PosD = un32CntY * mt_un32Width;
    for(un32CntX = 0; un32CntX < mt_un32Width; un32CntX++)
    {
      // Grab pixels
      un8Tot = 0;
      for(n32CntY = _vint32(un32CntY) - _vint32(un8Factor); n32CntY <= _vint32(un32CntY + un8Factor); n32CntY++)
      {
        un32PosS = _vuint32(n32CntY) * mt_un32Width;
        for(n32CntX = _vint32(un32CntX) - _vint32(un8Factor); n32CntX <= _vint32(un32CntX + un8Factor); n32CntX++)
        {
          if((n32CntY >= 0) && (n32CntX >= 0) && 
             (n32CntY < _vint32(mt_un32Height)) && (n32CntX < _vint32(mt_un32Width)))
          {
            // Check if the pixel already in the list
            bF = false;
            for(un8Cnt = 0; un8Cnt < un8Tot; un8Cnt++)
            {
              if((un8R[un8Cnt] == mt_ptBuffer[un32PosS + n32CntX].un8R) &&
                 (un8G[un8Cnt] == mt_ptBuffer[un32PosS + n32CntX].un8G) &&
                 (un8B[un8Cnt] == mt_ptBuffer[un32PosS + n32CntX].un8B))
              {
                un8C[un8Cnt]++;
                bF = true;
                break;
              }
            }
            // Add the pixel to the list
            if(!bF)
            {
              un8R[un8Tot] = mt_ptBuffer[un32PosS + n32CntX].un8R;
              un8G[un8Tot] = mt_ptBuffer[un32PosS + n32CntX].un8G;
              un8B[un8Tot] = mt_ptBuffer[un32PosS + n32CntX].un8B;
              un8C[un8Tot] = 1;
              un8Tot++;
            }
          }
        }
      }
      // Find the modulus of the color
      un8Pos = 0;
      for(un8Cnt = 1; un8Cnt < un8Tot; un8Cnt++) if(un8C[un8Cnt] > un8C[un8Pos]) un8Pos = un8Cnt;
      // Put the modulus colors into buffer
      ptNew[un32PosD + un32CntX].un8R = un8R[un8Pos];
      ptNew[un32PosD + un32CntX].un8G = un8G[un8Pos];
      ptNew[un32PosD + un32CntX].un8B = un8B[un8Pos];
    }
  }
  
  // ..... Change pointer
  free(mt_ptBuffer);
  mt_ptBuffer = ptNew;
    
  // ..... OK !
  return(true);
}
_vbool _CVortexImageBuilder::AdjustBrightness(_vfloat fMultFactor, _vint8 un8AddFactor,
                                              _vbool bAffectRed, _vbool bAffectGreen, _vbool bAffectBlue)
{
  // ..... Temporary variables
  _vint32        n32Tmp;
  _vuint32       un32Cnt;
  _TVortexUN8RGB *ptTmp = mt_ptBuffer;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Uninitialized _CVortexImageBuilder");
    return(false);
  }
  
  // ..... Adjust brightness
  for(un32Cnt = 0; un32Cnt < (mt_un32Width * mt_un32Height); un32Cnt++)
  {
    if(bAffectRed)
    {
      n32Tmp = _vint32(_vfloat(ptTmp->un8R) * fMultFactor + _vfloat(un8AddFactor));
      if(n32Tmp < 0) n32Tmp = 0;
      if(n32Tmp > MAX_COLOR_VALUE) n32Tmp = MAX_COLOR_VALUE;
      ptTmp->un8R = _vuint8(n32Tmp);  
    }
    if(bAffectGreen)
    {
      n32Tmp = _vint32(_vfloat(ptTmp->un8G) * fMultFactor + _vfloat(un8AddFactor));
      if(n32Tmp < 0) n32Tmp = 0;
      if(n32Tmp > MAX_COLOR_VALUE) n32Tmp = MAX_COLOR_VALUE;
      ptTmp->un8G = _vuint8(n32Tmp);  
    }
    if(bAffectBlue)
    {
     n32Tmp = _vint32(_vfloat(ptTmp->un8B) * fMultFactor + _vfloat(un8AddFactor));
      if(n32Tmp < 0) n32Tmp = 0;
      if(n32Tmp > MAX_COLOR_VALUE) n32Tmp = MAX_COLOR_VALUE;
      ptTmp->un8B = _vuint8(n32Tmp);  
    }    
    ptTmp++;
  }

  // ..... OK !
  return(true);
}
_vbool _CVortexImageBuilder::Grayscale(_vbool bUseLinearEquation)
{
  // ..... Temporary variables
  _vuint32       un32Cnt, un32R, un32G, un32B;
  _TVortexUN8RGB *ptTmp = mt_ptBuffer;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Uninitialized _CVortexImageBuilder");
    return(false);
  }
  
  // ..... Cconvert to grayscale
  for(un32Cnt = 0; un32Cnt < (mt_un32Width * mt_un32Height); un32Cnt++)
  {
    un32R = ptTmp->un8R;
    un32G = ptTmp->un8G;
    un32B = ptTmp->un8B;
    if(bUseLinearEquation)
      un32R = (un32R + un32G + un32B) / 3;
    else
      un32R = ((un32R * 3) + (un32G * 5) + (un32B * 2)) / 10;
    ptTmp->un8R = un32R;
    ptTmp->un8G = un32R;
    ptTmp->un8B = un32R;
    ptTmp++;
  }

  // ..... OK !
  return(true);
}
_vbool _CVortexImageBuilder::RoughResize(_vuint32 un32NewWidth, _vuint32 un32NewHeight)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptNew, *ptOld = mt_ptBuffer;
  _vuint32       un32X, un32Y, un32S, un32D;
  _vfloat        fRX, fRY, fX, fY;

  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Cannot resize image of an uninitialized _CVortexImageBuilder");
    return(false);
  }
  if((un32NewWidth == mt_un32Width) && (un32NewHeight == mt_un32Height)) return(true);
  
  // ..... Determine size
  if(un32NewWidth)
  {
    if(un32NewHeight)
    {
      if((un32NewWidth < 2) || (un32NewHeight < 2))
      {
        VSetErrorCode(VERR_InvalidParameter);
        VError(_F_, _L_, "New image size to small, minimum width and height are 2 pixels");
        return(false);
      }
    }
    else
      un32NewHeight = _vuint32(_vfloat(mt_un32Height) / _vfloat(mt_un32Width) * _vfloat(un32NewWidth));
  }
  else
  {
    if(un32NewHeight)
      un32NewWidth = _vuint32(_vfloat(mt_un32Width) / _vfloat(mt_un32Height) * _vfloat(un32NewHeight));
    else
    {
      VSetErrorCode(VERR_InvalidParameter);
      VError(_F_, _L_, "Invalid new image size");
      return(false);
    }
  }
 
  // ..... Allocate memory for the new image
  ptNew = (_TVortexUN8RGB *) malloc(sizeof(_TVortexUN8RGB) * un32NewWidth * un32NewHeight);
  if(ptNew == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for image resizing buffer");
    return(false);
  }
  
  // ..... Calculate image ratio
  fRX  = _vfloat(mt_un32Width) / _vfloat(un32NewWidth);
  fRY  = _vfloat(mt_un32Height) / _vfloat(un32NewHeight);

  // ..... Do rough scalling
  fY = 0;
  for(un32Y = 0; un32Y < un32NewHeight; un32Y++)
  {
    un32D = un32Y * un32NewWidth;
    un32S = _vint32(fY) * mt_un32Width;
    fX = 0;
    for(un32X = 0; un32X < un32NewWidth; un32X++)
    {
      ptNew[un32D + un32X].un8R = ptOld[un32S + _vint32(fX)].un8R;
      ptNew[un32D + un32X].un8G = ptOld[un32S + _vint32(fX)].un8G;
      ptNew[un32D + un32X].un8B = ptOld[un32S + _vint32(fX)].un8B;
      fX += fRX;    
    }
    fY += fRY;
  }

  // ..... Destroy old image
  free(mt_ptBuffer);

  // ..... Copy variables
  mt_ptBuffer   = ptNew;
  mt_un32Width  = un32NewWidth;
  mt_un32Height = un32NewHeight;
  
  // ..... OK !
  return(true);
}
_vbool _CVortexImageBuilder::Resize(_vuint32 un32NewWidth, _vuint32 un32NewHeight)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptNew, *ptOld = mt_ptBuffer;
  _vuint32       un32X, un32Y, un32X1, un32Y1, un32X2, un32Y2, un32Pos1, un32Pos2, un32NewPos, un32Tot;
  _vuint32       un32DX, un32CX, un32DY, un32CY, un32R1, un32G1, un32B1, un32R2, un32G2, un32B2;
  _vfloat        fWLX, fWRX, fWTY, fWBY;
  _vfloat        fRX, fRY, fRX2, fRY2, fX, fY;

  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "Cannot resize image of an uninitialized _CVortexImageBuilder");
    return(false);
  }
  if((un32NewWidth == mt_un32Width) && (un32NewHeight == mt_un32Height)) return(true);

  // ..... Determine size
  if(un32NewWidth)
  {
    if(un32NewHeight)
    {
      if((un32NewWidth < 2) || (un32NewHeight < 2))
      {
        VSetErrorCode(VERR_InvalidParameter);
        VError(_F_, _L_, "New image size to small, minimum width and height are 2 pixels");
        return(false);
      }
    }
    else
      un32NewHeight = _vuint32(_vfloat(mt_un32Height) / _vfloat(mt_un32Width) * _vfloat(un32NewWidth));
  }
  else
  {
    if(un32NewHeight)
      un32NewWidth = _vuint32(_vfloat(mt_un32Width) / _vfloat(mt_un32Height) * _vfloat(un32NewHeight));
    else
    {
      VSetErrorCode(VERR_InvalidParameter);
      VError(_F_, _L_, "Invalid new image size");
      return(false);
    }
  }

  // ..... Allocate memory for the new image
  ptNew = (_TVortexUN8RGB *) malloc(sizeof(_TVortexUN8RGB) * un32NewWidth * un32NewHeight);
  if(ptNew == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for image resizing buffer");
    return(false);
  }
  
  // ..... Calculate image ratio
  fRX  = _vfloat(mt_un32Width) / _vfloat(un32NewWidth);
  fRY  = _vfloat(mt_un32Height) / _vfloat(un32NewHeight);
  fRX2 = fRX / _vfloat(2.0);
  fRY2 = fRY / _vfloat(2.0);

  // ..... The new image height is bigger than the source image height
  if(fRY <= 1)
  {
    // The new image width is also bigger than the source image width
    // --> BIGGEN the height
    //     BIGGEN the width         
    if(fRX <= 1)
    {
      // Zero the Y interpolator
      fY = 0; 
      // Loop for the height of the destination image
      for(un32Y = 0; un32Y < un32NewHeight; un32Y++)
      {
        // Calculate source Y coordinates
        un32Y1 = _vuint32(floor(fY));
        un32Y2 = un32Y1 + 1;
        if(un32Y2 >= mt_un32Height) un32Y2 = mt_un32Height - 1;
        // Calculate source Y positions
        un32Pos1 = un32Y1 * mt_un32Width;
        un32Pos2 = un32Y2 * mt_un32Width;
        // Calculate destination Y position
        un32NewPos = un32Y * un32NewWidth;
        // Zero the X interpolator
        fX = 0;
        // Loop for the width of the destination image
        for(un32X = 0; un32X < un32NewWidth; un32X++)
        {
          // Calculate source X coordinates
          un32X1 = _vuint32(floor(fX));
          un32X2 = un32X1 + 1;
          if(un32X2 >= mt_un32Width) un32X2 = mt_un32Width - 1;
          // Calculate weights for wheighted sum
          fWRX = fX - un32X1;
          fWLX = 1 - fWRX;
          fWBY = fY - un32Y1;
          fWTY = 1 - fWBY;
          // Get the top left pixel
          un32R1 = _vuint32(_vfloat(ptOld[un32Pos1 + un32X1].un8R) * fWLX);
          un32G1 = _vuint32(_vfloat(ptOld[un32Pos1 + un32X1].un8G) * fWLX);
          un32B1 = _vuint32(_vfloat(ptOld[un32Pos1 + un32X1].un8B) * fWLX);
          // Get the top right pixel
          un32R1 += _vuint32(_vfloat(ptOld[un32Pos1 + un32X2].un8R) * fWRX);
          un32G1 += _vuint32(_vfloat(ptOld[un32Pos1 + un32X2].un8G) * fWRX);
          un32B1 += _vuint32(_vfloat(ptOld[un32Pos1 + un32X2].un8B) * fWRX);
          // Get the bottom left pixel
          un32R2 = _vuint32(_vfloat(ptOld[un32Pos2 + un32X1].un8R) * fWLX);
          un32G2 = _vuint32(_vfloat(ptOld[un32Pos2 + un32X1].un8G) * fWLX);
          un32B2 = _vuint32(_vfloat(ptOld[un32Pos2 + un32X1].un8B) * fWLX);
          // Get the bottom right pixel
          un32R2 += _vuint32(_vfloat(ptOld[un32Pos2 + un32X2].un8R) * fWRX);
          un32G2 += _vuint32(_vfloat(ptOld[un32Pos2 + un32X2].un8G) * fWRX);
          un32B2 += _vuint32(_vfloat(ptOld[un32Pos2 + un32X2].un8B) * fWRX);
          // Put the final color to the destination buffer
          ptNew[un32NewPos + un32X].un8R = _vuint32(_vfloat(un32R1) * fWTY + _vfloat(un32R2) * fWBY);
          ptNew[un32NewPos + un32X].un8G = _vuint32(_vfloat(un32G1) * fWTY + _vfloat(un32G2) * fWBY);
          ptNew[un32NewPos + un32X].un8B = _vuint32(_vfloat(un32B1) * fWTY + _vfloat(un32B2) * fWBY);
          // Clip the final color if necessary
          if(ptNew[un32NewPos + un32X].un8R > 31) ptNew[un32NewPos + un32X].un8R = 31;
          if(ptNew[un32NewPos + un32X].un8G > 31) ptNew[un32NewPos + un32X].un8G = 31;
          if(ptNew[un32NewPos + un32X].un8B > 31) ptNew[un32NewPos + un32X].un8B = 31;
          // Increment the X interpolator
          fX += fRX;    
        }
        // Increment the Y interpolator
        fY += fRY;
      }
    }
    // CASE 2 : The new image width is smaller than the source image width
    // --> BIGGEN the height
    //     SMALLEN the width         
    else
    {
      // Zero the Y interpolator
      fY = 0; 
      // Loop for the height of the destination image
      for(un32Y = 0; un32Y < un32NewHeight; un32Y++)
      {
        // Calculate source Y coordinates
        un32Y1 = _vuint32(floor(fY));
        un32Y2 = un32Y1 + 1;
        if(un32Y2 >= mt_un32Height) un32Y2 = mt_un32Height - 1;
        // Calculate source Y positions
        un32Pos1 = un32Y1 * mt_un32Width;
        un32Pos2 = un32Y2 * mt_un32Width;
        // Calculate destination Y position
        un32NewPos = un32Y * un32NewWidth;
        // Set the X interpolator
        fX = fRX2;
        // Loop for the width of the destination image
        for(un32X = 0; un32X < un32NewWidth; un32X++)
        {
          // Calculate source X coordinates
          un32X1 = _vuint32(floor(fX - fRX2));
          un32X2 = _vuint32(ceil(fX + fRX2));
          if(un32X2 >= mt_un32Width) un32X2 = mt_un32Width - 1;
          un32DX = un32X2 - un32X1 + 1;
          // Calculate weights for wheighted sum
          fWBY = fY - un32Y1;
          fWTY = 1 - fWBY;
          // Get top line pixels
          un32R1 = un32G1 = un32B1 = 0;
          for(un32CX = un32X1; un32CX <= un32X2; un32CX++)
          {
            un32R1 += ptOld[un32Pos1 + un32CX].un8R;
            un32G1 += ptOld[un32Pos1 + un32CX].un8G;
            un32B1 += ptOld[un32Pos1 + un32CX].un8B;
          }
          un32R1 /= un32DX; un32G1 /= un32DX; un32B1 /= un32DX;
          // Get bottom line pixels
          un32R2 = un32G2 = un32B2 = 0;
          for(un32CX = un32X1; un32CX <= un32X2; un32CX++)
          {
            un32R2 += ptOld[un32Pos2 + un32CX].un8R;
            un32G2 += ptOld[un32Pos2 + un32CX].un8G;
            un32B2 += ptOld[un32Pos2 + un32CX].un8B;
          }
          un32R2 /= un32DX; un32G2 /= un32DX; un32B2 /= un32DX;
          // Put the final color to the destination buffer
          ptNew[un32NewPos + un32X].un8R = _vuint32(_vfloat(un32R1) * fWTY + _vfloat(un32R2) * fWBY);
          ptNew[un32NewPos + un32X].un8G = _vuint32(_vfloat(un32G1) * fWTY + _vfloat(un32G2) * fWBY);
          ptNew[un32NewPos + un32X].un8B = _vuint32(_vfloat(un32B1) * fWTY + _vfloat(un32B2) * fWBY);
          // Clip the final color if necessary
          if(ptNew[un32NewPos + un32X].un8R > 31) ptNew[un32NewPos + un32X].un8R = 31;
          if(ptNew[un32NewPos + un32X].un8G > 31) ptNew[un32NewPos + un32X].un8G = 31;
          if(ptNew[un32NewPos + un32X].un8B > 31) ptNew[un32NewPos + un32X].un8B = 31;
          // Increment the X interpolator
          fX += fRX;    
        }
        // Increment the Y interpolator
        fY += fRY;
      }
    }
  }
  // ..... The new image height is smaller than the source image height
  else
  {
    // CASE 3 : The new image width is bigger than the source image width
    // --> SMALLEN the height
    //     BIGGEN the width         
    if(fRX <= 1)
    {
      // Set the Y interpolator
      fY = fRY2; 
      // Loop for the height of the destination image
      for(un32Y = 0; un32Y < un32NewHeight; un32Y++)
      {
        // Calculate source Y coordinates
        un32Y1 = _vuint32(floor(fY - fRY2));
        un32Y2 = _vuint32(ceil(fY + fRY2));
        if(un32Y2 >= mt_un32Height) un32Y2 = mt_un32Height - 1;
        un32DY = un32Y2 - un32Y1 + 1;
        // Calculate destination Y position
        un32NewPos = un32Y * un32NewWidth;
        // Zero the X interpolator
        fX = 0;
        // Loop for the width of the destination image
        for(un32X = 0; un32X < un32NewWidth; un32X++)
        {
          // Calculate source X coordinates
          un32X1 = _vuint32(floor(fX));
          un32X2 = un32X1 + 1;
          if(un32X2 >= mt_un32Width) un32X2 = mt_un32Width - 1;
          // Calculate weights for wheighted sum
          fWRX = fX - un32X1;
          fWLX = 1 - fWRX;
          // Get the left collumn pixels
          un32R1 = un32G1 = un32B1 = 0;
          for(un32CY = un32Y1; un32CY <= un32Y2; un32CY++)
          {
            un32R1 += ptOld[un32CY * mt_un32Width + un32X1].un8R;
            un32G1 += ptOld[un32CY * mt_un32Width + un32X1].un8G;
            un32B1 += ptOld[un32CY * mt_un32Width + un32X1].un8B;
          }
          un32R1 /= un32DY; un32G1 /= un32DY; un32B1 /= un32DY;
          // Get the rightcollumn pixels
          un32R2 = un32G2 = un32B2 = 0;
          for(un32CY = un32Y1; un32CY <= un32Y2; un32CY++)
          {
            un32R2 += ptOld[un32CY * mt_un32Width + un32X2].un8R;
            un32G2 += ptOld[un32CY * mt_un32Width + un32X2].un8G;
            un32B2 += ptOld[un32CY * mt_un32Width + un32X2].un8B;
          }
          un32R2 /= un32DY; un32G2 /= un32DY; un32B2 /= un32DY;
          // Put the final color to the destination buffer
          ptNew[un32NewPos + un32X].un8R = _vuint32(_vfloat(un32R1) * fWLX + _vfloat(un32R2) * fWRX);
          ptNew[un32NewPos + un32X].un8G = _vuint32(_vfloat(un32G1) * fWLX + _vfloat(un32G2) * fWRX);
          ptNew[un32NewPos + un32X].un8B = _vuint32(_vfloat(un32B1) * fWLX + _vfloat(un32B2) * fWRX);
          // Clip the final color if necessary
          if(ptNew[un32NewPos + un32X].un8R > 31) ptNew[un32NewPos + un32X].un8R = 31;
          if(ptNew[un32NewPos + un32X].un8G > 31) ptNew[un32NewPos + un32X].un8G = 31;
          if(ptNew[un32NewPos + un32X].un8B > 31) ptNew[un32NewPos + un32X].un8B = 31;
          // Increment the X interpolator
          fX += fRX;    
        }
        // Increment the Y interpolator
        fY += fRY;
      }
    }
    // CASE 4 : The new image width is also smaller than the source image width
    // --> SMALLEN the height
    //     SMALLEN the width         
    else
    {
      // Set the Y interpolator
      fY = fRY2; 
      // Loop for the height of the destination image
      for(un32Y = 0; un32Y < un32NewHeight; un32Y++)
      {
        // Calculate source Y coordinates
        un32Y1 = _vuint32(floor(fY - fRY2));
        un32Y2 = _vuint32(ceil(fY + fRY2));
        if(un32Y2 >= mt_un32Height) un32Y2 = mt_un32Height - 1;
        un32DY = un32Y2 - un32Y1 + 1;
        // Calculate destination Y position
        un32NewPos = un32Y * un32NewWidth;
        // Set the X interpolator
        fX = fRX2;
        // Loop for the width of the destination image
        for(un32X = 0; un32X < un32NewWidth; un32X++)
        {
          // Calculate source X coordinates
          un32X1 = _vuint32(floor(fX - fRX2));
          un32X2 = _vuint32(ceil(fX + fRX2));
          if(un32X2 >= mt_un32Width) un32X2 = mt_un32Width - 1;
          un32DX = un32X2 - un32X1 + 1;
          // Grab the source pixels
          un32R1 = un32G1 = un32B1 = un32Tot = 0;
          for(un32CY = un32Y1; un32CY <= un32Y2; un32CY++)
          {
            // Calculate the source Y position
            un32Pos1 = un32CY * mt_un32Width;
            for(un32CX = un32X1; un32CX <= un32X2; un32CX++)
            {
              un32R1 += ptOld[un32Pos1 + un32CX].un8R;
              un32G1 += ptOld[un32Pos1 + un32CX].un8G;
              un32B1 += ptOld[un32Pos1 + un32CX].un8B;
              un32Tot++;
            }
          }
          // Put the final color to the destination buffer
          ptNew[un32NewPos + un32X].un8R = un32R1 / un32Tot;
          ptNew[un32NewPos + un32X].un8G = un32G1 / un32Tot;
          ptNew[un32NewPos + un32X].un8B = un32B1 / un32Tot;
          // Clip the final color if necessary
          if(ptNew[un32NewPos + un32X].un8R > 31) ptNew[un32NewPos + un32X].un8R = 31;
          if(ptNew[un32NewPos + un32X].un8G > 31) ptNew[un32NewPos + un32X].un8G = 31;
          if(ptNew[un32NewPos + un32X].un8B > 31) ptNew[un32NewPos + un32X].un8B = 31;
          // Increment the X interpolator
          fX += fRX;    
        }
        // Increment the Y interpolator
        fY += fRY;
      }
    }
  }

  // ..... Destroy old image
 free(mt_ptBuffer);

  // ..... Copy variables
  mt_ptBuffer   = ptNew;
  mt_un32Width  = un32NewWidth;
  mt_un32Height = un32NewHeight;
  
  // ..... OK !
  return(true);
}
_vbool _CVortexImageBuilder::ResizeToFitBackground(_CVortex3DEnv *pcsV3DEnv)
{
  // ..... Use the supplied 3d environment class
  if(pcsV3DEnv)
    return(Resize(pcsV3DEnv->mt_tRB->un32Width, pcsV3DEnv->mt_tRB->un32Height));
  
  // ..... Use the active 3D environment
  if(_pcsAct3DEnv_)
    return(Resize(_pcsAct3DEnv_->mt_tRB->un32Width, _pcsAct3DEnv_->mt_tRB->un32Height));
  
  // ..... Error
  VSetErrorCode(VERR_NotFound);
  VError(_F_, _L_, "No 3D environment can be used as reference for resizing image");
  return(false);
}
_vbool _CVortexImageBuilder::ResizeToFitTexture(_vuint32 un32TexWidth, _vuint32 un32TexHeight)
{
  // ..... Determine the texture width
  if(un32TexWidth)
  {
    if((un32TexWidth !=   2) && (un32TexWidth !=   2) && (un32TexWidth !=   8) &&
       (un32TexWidth !=  16) && (un32TexWidth !=  32) && (un32TexWidth !=  64) &&
       (un32TexWidth != 128) && (un32TexWidth != 256) && (un32TexWidth != 512))
    {
      VSetErrorCode(VERR_InvalidParameter);
      VError(_F_, _L_, "Texture width must be one of 2/4/8/16/32/64/128/256/512 pixels");
      return(false);
    }
  }
  else
  {
    if(mt_un32Width > 256) 
      un32TexWidth = 512;
    else if(mt_un32Width > 128) 
      un32TexWidth = 256;
    else if(mt_un32Width > 64) 
      un32TexWidth = 128;
    else if(mt_un32Width > 32) 
      un32TexWidth = 64;
    else if(mt_un32Width > 16) 
      un32TexWidth = 32;
    else if(mt_un32Width > 8) 
      un32TexWidth = 16;
    else if(mt_un32Width > 4) 
      un32TexWidth = 8;
    else if(mt_un32Width > 2) 
      un32TexWidth = 4;
    else
      un32TexWidth = 2;
  }
  
  // ..... Determine the texture height
  if(un32TexWidth)
  {
    if((un32TexWidth < 2) || (un32TexWidth > 512))
    {
      VSetErrorCode(VERR_InvalidParameter);
      VError(_F_, _L_, "Texture height must between 2 to 512 pixels");
      return(false);
    }
  }
  else
  {
    un32TexHeight = _vuint32(_vfloat(mt_un32Height) / _vfloat(mt_un32Width) * _vfloat(un32TexWidth));
    if(un32TexHeight < 2)   un32TexHeight = 2;
    if(un32TexHeight > 512) un32TexHeight = 512;
  }

  // ..... Do resize
  return(Resize(un32TexWidth, un32TexHeight));
}
_vbool _CVortexImageBuilder::TransferToImage(_CVortexImage &csVImage, _EVortexAAHint eVAAHint)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptTmp;
  _vuint8        *pun8Buff, *pun8Tmp;
  _vuint32       un32Cnt;
  _vbool         bMemoryAligned = false;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "No data to transfer (this _CVortexImageBuilder is not initialized)");
    return(false);
  }

  // ..... Destroy previous image, if any
  if(csVImage.mt_pun8Buffer)
  {
    if(csVImage.mt_bAligned)
      ssefree(csVImage.mt_pun8Buffer);
    else
      free(csVImage.mt_pun8Buffer);
  }
  csVImage.mt_bAligned   = false;
  csVImage.mt_pun8Buffer = NULL;
  csVImage.mt_un32Width  = 0;
  csVImage.mt_un32Height = 0;
  csVImage.mt_un32SizeSL = 0;

  // ..... Allocate memory
#ifdef _ALIGN_BUFF_MEM_
  if(eVAAHint == AAH_SSE) bMemoryAligned = true;
#else
  eVAAHint = eVAAHint;
#endif  
  if(bMemoryAligned)
    pun8Buff = (_vuint8 *) ssemalloc(mt_un32Width * mt_un32Height * 4);
  else
    pun8Buff = (_vuint8 *) malloc(mt_un32Width * mt_un32Height * 4);
  if(pun8Buff == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for image buffer");
    return(false);
  }
  csVImage.mt_bAligned = bMemoryAligned;
  
  // ..... Convert image
  ptTmp   = mt_ptBuffer;
  pun8Tmp = pun8Buff;
  for(un32Cnt = 0; un32Cnt < (mt_un32Width * mt_un32Height); un32Cnt++)
  {
    *pun8Tmp++ = (ptTmp->un8B << 3);
    *pun8Tmp++ = (ptTmp->un8G << 3);
    *pun8Tmp++ = (ptTmp->un8R << 3);
    *pun8Tmp++ = 0;
    ptTmp++;
  }

  // ..... Reinitialize the image
  csVImage.mt_pun8Buffer = pun8Buff; 
  csVImage.mt_un32Width  = mt_un32Width;
  csVImage.mt_un32Height = mt_un32Height;
  csVImage.mt_un32SizeSL = mt_un32Width * 4;

  // .....OK
  return(true);
}
_vbool _CVortexImageBuilder::TransferToAnimatedImage(_CVortexAnimatedImage &csVAnimatedImage, 
                                                     _vuint8 un8SrcRowCount, _vuint8 un8SrcColCount,
                                                     _EVortexAAHint eVAAHint)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptTmp;
  _vuint8        *pun8Buff, *pun8Tmp, un8R, un8C;
  _vuint32       un32Rs, un32Cs, un32Xs, un32Ys, un32X, un32Y;
  _vbool         bMemoryAligned = false;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "No data to transfer (this _CVortexImageBuilder is not initialized)");
    return(false);
  }
  if((un8SrcRowCount * un8SrcColCount) < 1)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Invalid parameter value");
    return(false);
  }
  if(mt_un32Height % un8SrcRowCount)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Image height must be divisible by un8SrcRowCount");
    return(false);
  }
  if(mt_un32Width % un8SrcColCount)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Image width must be divisible by un8SrcColCount");
    return(false);
  }

  // ..... Destroy previous image, if any
  if(csVAnimatedImage.mt_pun8BufferArray)
  {
    if(csVAnimatedImage.mt_bAligned)
      ssefree(csVAnimatedImage.mt_pun8BufferArray);
    else
      free(csVAnimatedImage.mt_pun8BufferArray);
  }
  csVAnimatedImage.mt_bAligned        = false;
  csVAnimatedImage.mt_pun8BufferArray = NULL;
  csVAnimatedImage.mt_pun8BufferMax   = NULL;
  csVAnimatedImage.mt_pun8Buffer      = NULL;
  csVAnimatedImage.mt_un32Width       = 0;
  csVAnimatedImage.mt_un32Height      = 0;
  csVAnimatedImage.mt_un32TotFrame    = 0;
  csVAnimatedImage.mt_un32CurFrame    = 0;
  csVAnimatedImage.mt_un32FrameSize   = 0;
  csVAnimatedImage.mt_un32StepDelay   = 0;
  csVAnimatedImage.mt_n8Step          = 0;
  csVAnimatedImage.mt_un32SizeSL      = 0;
  
  // ..... Allocate memory
#ifdef _ALIGN_BUFF_MEM_
  if(eVAAHint == AAH_SSE) bMemoryAligned = true;
#else
  eVAAHint = eVAAHint;
#endif  
  if(bMemoryAligned)
    pun8Buff = (_vuint8 *) ssemalloc(mt_un32Width * mt_un32Height * 4);
  else
    pun8Buff = (_vuint8 *) malloc(mt_un32Width * mt_un32Height * 4);
  if(pun8Buff == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for image buffer");
    return(false);
  }
  csVAnimatedImage.mt_bAligned = bMemoryAligned;
  
  // ..... Convert image
  pun8Tmp = pun8Buff;
  un32Rs  = mt_un32Height / un8SrcRowCount;
  un32Cs  = mt_un32Width / un8SrcColCount;
  for(un8R = 0; un8R < un8SrcRowCount; un8R++)
  {
    un32Ys = un8R * un32Rs;
    for(un8C = 0; un8C < un8SrcColCount; un8C++)
    {
      un32Xs = un8C * un32Cs;
      for(un32Y = un32Ys; un32Y < (un32Ys + un32Rs); un32Y++)
      {
        ptTmp  = &mt_ptBuffer[un32Y * mt_un32Width + un32Xs];
        for(un32X = un32Xs; un32X < (un32Xs + un32Cs); un32X++)
        {
          *pun8Tmp++ = (ptTmp->un8B << 3);
          *pun8Tmp++ = (ptTmp->un8G << 3);
          *pun8Tmp++ = (ptTmp->un8R << 3);
          *pun8Tmp++ = 0;
          ptTmp++;
        }
      }
    }
  }

  // ..... Reinitialize the image
  csVAnimatedImage.mt_pun8BufferArray = pun8Buff; 
  csVAnimatedImage.mt_pun8Buffer      = csVAnimatedImage.mt_pun8BufferArray; 
  csVAnimatedImage.mt_un32Width       = un32Cs;
  csVAnimatedImage.mt_un32Height      = un32Rs;
  csVAnimatedImage.mt_un32TotFrame    = un8SrcRowCount * un8SrcColCount;
  csVAnimatedImage.mt_un32FrameSize   = un32Cs * un32Rs * 4;
  csVAnimatedImage.mt_pun8BufferMax   = csVAnimatedImage.mt_pun8BufferArray + 
                                        (csVAnimatedImage.mt_un32FrameSize *
                                         (csVAnimatedImage.mt_un32TotFrame - 1));
  csVAnimatedImage.mt_n8Step          = 1;
  csVAnimatedImage.mt_un32SizeSL      = un32Cs * 4;

  // .....OK
  return(true);
}
_vbool _CVortexImageBuilder::TransferToTexture(_CVortexTexture &csVTexture, _EVortexAAHint eVAAHint)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptTmp;
  _vuint8        *pun8Buff, *pun8Tmp;
  _vuint32       un32Cnt;
  _vbool         bMemoryAligned = false;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "No data to transfer (this _CVortexImageBuilder is not initialized)");
    return(false);
  }
  if((mt_un32Width != 2)  && (mt_un32Width != 4)  && (mt_un32Width != 8)   && (mt_un32Width != 16)  &&
     (mt_un32Width != 32) && (mt_un32Width != 64) && (mt_un32Width != 128) && (mt_un32Width != 256) &&
     (mt_un32Width != 512))
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Texture width must be one of 2/4/8/16/32/64/128/256/512 pixels");
    return(false);
  }

  // ..... Destroy previous image, if any
  if(csVTexture.mt_pun8Buffer)
  {
    if(csVTexture.mt_bAligned)
      ssefree(csVTexture.mt_pun8Buffer);
    else
      free(csVTexture.mt_pun8Buffer);
  }
  csVTexture.mt_bAligned      = false;
  csVTexture.mt_un32Width     = 0;
  csVTexture.mt_un32Height    = 0;
  csVTexture.mt_un32Mask      = 0;
  csVTexture.mt_un32MultShift = 0;
  
  // ..... Allocate memory
#ifdef _ALIGN_BUFF_MEM_
  if(eVAAHint == AAH_SSE) bMemoryAligned = true;
#else
  eVAAHint = eVAAHint;
#endif  
  if(bMemoryAligned)
    pun8Buff = (_vuint8 *) ssemalloc(mt_un32Width * mt_un32Height * 4);
  else
    pun8Buff = (_vuint8 *) malloc(mt_un32Width * mt_un32Height * 4);
  if(pun8Buff == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for texture buffer");
    return(false);
  }
  csVTexture.mt_bAligned = bMemoryAligned;
  
  // ..... Convert image
  ptTmp   = mt_ptBuffer;
  pun8Tmp = pun8Buff;
  for(un32Cnt = 0; un32Cnt < (mt_un32Width * mt_un32Height); un32Cnt++)
  {
    *pun8Tmp++ = ptTmp->un8B;
    *pun8Tmp++ = ptTmp->un8G;
    *pun8Tmp++ = ptTmp->un8R;
    *pun8Tmp++ = 0;
    ptTmp++;
  }

  // ..... Reinitialize the image
  csVTexture.mt_pun8Buffer = pun8Buff; 
  csVTexture.mt_un32Width  = mt_un32Width;
  csVTexture.mt_un32Height = mt_un32Height;
  csVTexture.mt_un32Mask  = _TMASK_ * mt_un32Width;
  if(mt_un32Width == 2)
    csVTexture.mt_un32MultShift = 1;
  else if(mt_un32Width == 4)
    csVTexture.mt_un32MultShift = 2;
  else if(mt_un32Width == 8)
    csVTexture.mt_un32MultShift = 3;
  else if(mt_un32Width == 16)
    csVTexture.mt_un32MultShift = 4;
  else if(mt_un32Width == 32)
    csVTexture.mt_un32MultShift = 5;
  else if(mt_un32Width == 64)
    csVTexture.mt_un32MultShift = 6;
  else if(mt_un32Width == 128)
    csVTexture.mt_un32MultShift = 7;
  else if(mt_un32Width == 256)
    csVTexture.mt_un32MultShift = 8;
  else
    csVTexture.mt_un32MultShift = 9;
  
  // .....OK
  return(true);
}
_vbool _CVortexImageBuilder::TransferToAnimatedTexture(_CVortexAnimatedTexture &csVAnimatedTexture, 
                                                       _vuint8 un8SrcRowCount, _vuint8 un8SrcColCount,
                                                       _EVortexAAHint eVAAHint)
{
  // ..... Temporary variables
  _TVortexUN8RGB *ptTmp;
  _vuint8        *pun8Buff, *pun8Tmp, un8R, un8C;
  _vuint32       un32Rs, un32Cs, un32Xs, un32Ys, un32X, un32Y;
  _vbool         bMemoryAligned = false;
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "No data to transfer (this _CVortexImageBuilder is not initialized)");
    return(false);
  }
  if((un8SrcRowCount * un8SrcColCount) < 1)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Invalid parameter value");
    return(false);
  }
  if(mt_un32Height % un8SrcRowCount)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Texture height must be divisible by un8SrcRowCount");
    return(false);
  }
  if(mt_un32Width % un8SrcColCount)
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Texture width must be divisible by un8SrcColCount");
    return(false);
  }
  
  // ..... Calculate texture size
  un32Rs  = mt_un32Height / un8SrcRowCount;
  un32Cs  = mt_un32Width / un8SrcColCount;
  if((un32Cs != 2)  && (un32Cs != 4)   && (un32Cs != 8)   && (un32Cs != 16) && (un32Cs != 32) &&
     (un32Cs != 64) && (un32Cs != 128) && (un32Cs != 256) && (un32Cs != 512))
  {
    VSetErrorCode(VERR_InvalidParameter);
    VError(_F_, _L_, "Texture (width / un8SrcColCount) must be one of 2/4/8/16/32/64/128/256/512 pixels");
    return(false);
  }

  // ..... Destroy previous image, if any
  if(csVAnimatedTexture.mt_pun8BufferArray)
  {
    if(csVAnimatedTexture.mt_bAligned)
      ssefree(csVAnimatedTexture.mt_pun8BufferArray);
    else
      free(csVAnimatedTexture.mt_pun8BufferArray);
  }
  csVAnimatedTexture.mt_bAligned        = false;
  csVAnimatedTexture.mt_pun8BufferArray = NULL;
  csVAnimatedTexture.mt_pun8BufferMax   = NULL;
  csVAnimatedTexture.mt_pun8Buffer      = NULL;
  csVAnimatedTexture.mt_un32Width       = 0;
  csVAnimatedTexture.mt_un32Height      = 0;
  csVAnimatedTexture.mt_un32TotFrame    = 0;
  csVAnimatedTexture.mt_un32CurFrame    = 0;
  csVAnimatedTexture.mt_un32FrameSize   = 0;
  csVAnimatedTexture.mt_un32StepDelay   = 0;
  csVAnimatedTexture.mt_n8Step          = 0;
  csVAnimatedTexture.mt_un32Mask        = 0;
  csVAnimatedTexture.mt_un32MultShift   = 0;
  
  // ..... Allocate memory
#ifdef _ALIGN_BUFF_MEM_
  if(eVAAHint == AAH_SSE) bMemoryAligned = true;
#else
  eVAAHint = eVAAHint;
#endif  
  if(bMemoryAligned)
    pun8Buff = (_vuint8 *) ssemalloc(mt_un32Width * mt_un32Height * 4);
  else
    pun8Buff = (_vuint8 *) malloc(mt_un32Width * mt_un32Height * 4);
  if(pun8Buff == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for image buffer");
    return(false);
  }
  csVAnimatedTexture.mt_bAligned = bMemoryAligned;
  
  // ..... Convert image
  pun8Tmp = pun8Buff;
  for(un8R = 0; un8R < un8SrcRowCount; un8R++)
  {
    un32Ys = un8R * un32Rs;
    for(un8C = 0; un8C < un8SrcColCount; un8C++)
    {
      un32Xs = un8C * un32Cs;
      for(un32Y = un32Ys; un32Y < (un32Ys + un32Rs); un32Y++)
      {
        ptTmp  = &mt_ptBuffer[un32Y * mt_un32Width + un32Xs];
        for(un32X = un32Xs; un32X < (un32Xs + un32Cs); un32X++)
        {
          *pun8Tmp++ = ptTmp->un8B;
          *pun8Tmp++ = ptTmp->un8G;
          *pun8Tmp++ = ptTmp->un8R;
          *pun8Tmp++ = 0;
          ptTmp++;
        }
      }
    }
  }

  // ..... Reinitialize the image
  csVAnimatedTexture.mt_pun8BufferArray = pun8Buff; 
  csVAnimatedTexture.mt_pun8Buffer      = csVAnimatedTexture.mt_pun8BufferArray; 
  csVAnimatedTexture.mt_un32Width       = un32Cs;
  csVAnimatedTexture.mt_un32Height      = un32Rs;
  csVAnimatedTexture.mt_un32TotFrame    = un8SrcRowCount * un8SrcColCount;
  csVAnimatedTexture.mt_un32FrameSize   = un32Cs * un32Rs * 4;
  csVAnimatedTexture.mt_pun8BufferMax   = csVAnimatedTexture.mt_pun8BufferArray + 
                                          (csVAnimatedTexture.mt_un32FrameSize * 
                                           (csVAnimatedTexture.mt_un32TotFrame - 1));
  csVAnimatedTexture.mt_n8Step          = 1;
  csVAnimatedTexture.mt_un32Mask        = _TMASK_ * un32Cs;
  if(un32Cs == 2)
    csVAnimatedTexture.mt_un32MultShift = 1;
  else if(un32Cs == 4)
    csVAnimatedTexture.mt_un32MultShift = 2;
  else if(un32Cs == 8)
    csVAnimatedTexture.mt_un32MultShift = 3;
  else if(un32Cs == 16)
    csVAnimatedTexture.mt_un32MultShift = 4;
  else if(un32Cs == 32)
    csVAnimatedTexture.mt_un32MultShift = 5;
  else if(un32Cs == 64)
    csVAnimatedTexture.mt_un32MultShift = 6;
  else if(un32Cs == 128)
    csVAnimatedTexture.mt_un32MultShift = 7;
  else if(un32Cs == 256)
    csVAnimatedTexture.mt_un32MultShift = 8;
  else
    csVAnimatedTexture.mt_un32MultShift = 9;

  // .....OK
  return(true);
}

// ***** File I/O functions
_vbool _CVortexImageBuilder::LoadBMP(const _vchar *pcFileName, _vbool bUseDithering)
{
  // ..... Temporary variables
  _TVortexFILE       *ptFile;
  _TBITMAPFILEHEADER tFileHeader;
  _TBITMAPINFOHEADER tInfoHeader;
  _TRGBQUAD          *ptRGBQ;
  _TVortexUN8RGB     *ptTmp;
  _vuchar            *pucData, *pucTmp;
  _vint32            n32ImageSize, n32ScanWidth, n32ScanPadding, n32YPos, n32Y;
  _vint32            n32ColorCount, n32BitCount, n32Width, n32Height, n32LoadSize;
  _vuint32           un32Skip, un32CntX, un32CntY, un32AGet, un32XGet;
  _vuint8            un8R, un8G, un8B, un8D;
  _vchar             acFile[256];
  
  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".bmp") == NULL) strcat(acFile, ".bmp");
  ptFile = fopen(acFile, "rb");
  if(ptFile == NULL)
  {
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    
  
  // ..... Load the file header
  if(fread(&tFileHeader, sizeof(tFileHeader), 1, ptFile) < 1) goto ERRRead;

  // ..... Check the header
  if((tFileHeader.aucBM[0] != 'B') || (tFileHeader.aucBM[1] != 'M'))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" is an invalid BMP file", acFile);
    return(false);
  }    
  
  // ..... Load the information header data
  if(fread(&tInfoHeader, sizeof(tInfoHeader), 1, ptFile) < 1) goto ERRRead;

  // ..... Check
  if((tInfoHeader.un16BitCount != 1) && (tInfoHeader.un16BitCount != 8) && 
     (tInfoHeader.un16BitCount != 24))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "LoadBMP can only load 1-bit, 8-bit, and 24-bit bitmaps");
    return(false);
  }
  if(tInfoHeader.un32Compression != BI_RGB)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "LoadBMP can only load uncompressed bitmaps");
    return(false);
  }
  
  // ..... Calculate some information
  if(tInfoHeader.un32ClrUsed)
    n32ColorCount  = tInfoHeader.un32ClrUsed;
  else
    n32ColorCount  = _vint32(pow(2, tInfoHeader.un16BitCount));
  n32BitCount    = tInfoHeader.un16BitCount;
  n32Width       = tInfoHeader.n32Width;
  n32Height      = tInfoHeader.n32Height;
  n32ImageSize   = n32Width * n32Height;
  if(n32BitCount == 1)
  {  
    if(n32Width % 8)
    {
      n32ScanWidth   = ((n32Width / 8 + 1) + 3) & (~3);
      n32ScanPadding = n32ScanWidth - (n32Width / 8 + 1);
      n32LoadSize    = n32Width / 8 + 1;
    }
    else
    {
      n32ScanWidth   = ((n32Width / 8) + 3) & (~3);
      n32ScanPadding = n32ScanWidth - (n32Width / 8);
      n32LoadSize    = n32Width / 8;
    }
  }
  else
  {
    n32ScanWidth   = ((n32Width * (n32BitCount / 8)) + 3) & (~3);
    n32ScanPadding = n32ScanWidth - (n32Width * (n32BitCount / 8));
    n32LoadSize    = n32Width * (n32BitCount / 8);
  }  
  
  // ..... Skip some bytes if necessary
  un32Skip = tInfoHeader.un32Size - sizeof(tInfoHeader);
  if(un32Skip > 0)
    if(fseek(ptFile, un32Skip, SEEK_CUR) < 0) goto ERRRead;
  
  // ..... Load the pallete
  ptRGBQ = NULL;
  if(n32BitCount <= 8)
  {
    // Allocate memory
    ptRGBQ = (_TRGBQUAD *) malloc(sizeof(_TRGBQUAD) * n32ColorCount);
    if(ptRGBQ == NULL)
    {
      fclose(ptFile);
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for palette");
      return(false);
    }
    
    // Load it
    if(fread(ptRGBQ, sizeof(_TRGBQUAD), n32ColorCount, ptFile) < _vuint32(n32ColorCount))
    {
      free(ptRGBQ);
      goto ERRRead;
    }
  }
  
  // ..... Allocate memory for image
  if(n32BitCount == 1)
    pucData = (_vuchar *) malloc(n32ImageSize);
  else
    pucData = (_vuchar *) malloc(n32ImageSize * (n32BitCount / 8));
  if(pucData == NULL)
  {
    free(ptRGBQ);
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for loading bitmap");
    return(false);
  }
    
  // ..... Load the bitmap
  if(n32Height < 0)
  {
    for(n32Y = 0; n32Y < (-n32Height); n32Y++)
    {
      n32YPos = n32Y * n32LoadSize;
      if(fread(&pucData[n32YPos], n32LoadSize, 1, ptFile) < 1)
      {
        free(ptRGBQ);
        free(pucData);
        goto ERRRead;
      }
      if(fseek(ptFile, n32ScanPadding, SEEK_CUR) < 0) goto ERRRead;
    }
  }
  else
  {
    for(n32Y = 0; n32Y < n32Height; n32Y++)
    {
      n32YPos = ((n32Height - n32Y) * n32LoadSize) - n32LoadSize;
      if(fread(&pucData[n32YPos], n32LoadSize, 1, ptFile) < 1)
      {
        free(ptRGBQ);
        free(pucData);
        goto ERRRead;
      }
      if(fseek(ptFile, n32ScanPadding, SEEK_CUR) < 0) goto ERRRead;
    }
  }

  // ..... Close file
  fclose(ptFile);
  
  // ..... Destroy old image, if any
  FreeImage();
  
  // ..... Allocate memory for image buffer
  mt_un32Width  = _vuint32(n32Width);
  mt_un32Height = _vuint32((n32Height < 0) ? -n32Height : n32Height);
  mt_ptBuffer   = (_TVortexUN8RGB *) malloc(sizeof(_TVortexUN8RGB) * mt_un32Width * mt_un32Height);
  if(mt_ptBuffer == NULL)
  {
    free(ptRGBQ);
    free(pucData);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for image buffer");
    return(false);
  }

  // ..... Convert image
  ptTmp  = mt_ptBuffer;
  pucTmp = pucData;
  if(n32BitCount == 1)
  {
    un32AGet = 0;
    un32XGet = 0;
    for(un32CntY = 0; un32CntY < (_vuint32(n32LoadSize) * mt_un32Height); un32CntY++)
    {
      un8R = *pucTmp++;
      for(un32CntX = 0; un32CntX < 8; un32CntX++)
      {
        if(un8R & (0x80 >> un32CntX))
        {
          ptTmp->un8R = ptRGBQ[1].un8R >> 3;
          ptTmp->un8G = ptRGBQ[1].un8G >> 3;
          ptTmp->un8B = ptRGBQ[1].un8B >> 3;
        }
        else
        {
          ptTmp->un8R = ptRGBQ[0].un8R >> 3;
          ptTmp->un8G = ptRGBQ[0].un8G >> 3;
          ptTmp->un8B = ptRGBQ[0].un8B >> 3;
        }
        ptTmp++;
        un32AGet++;
        if(un32AGet >= (mt_un32Width * mt_un32Height)) break;
        un32XGet++;
        if(un32XGet >= mt_un32Width)
        {
          un32XGet = 0;
          break;
        }
      }
      if(un32AGet >= (mt_un32Width * mt_un32Height)) break;
    }
  }
  else
  {
    for(un32CntY = 0; un32CntY < mt_un32Height; un32CntY++)
    {
      for(un32CntX = 0; un32CntX < mt_un32Width; un32CntX++)
      {
        // Grab the pixel
        if(n32BitCount == 8)
        {
          un8R = ptRGBQ[*pucTmp].un8R;
          un8G = ptRGBQ[*pucTmp].un8G;
          un8B = ptRGBQ[*pucTmp].un8B;
          pucTmp++;
        }
        else
        {
          un8R = *pucTmp++;
          un8G = *pucTmp++;
          un8B = *pucTmp++;
        }  
        // Calculate destination pixelcolor
        if(((un32CntX & 0x0001) == (un32CntY & 0x0001)) && bUseDithering)
        {
          un8D = un8R - (un8R & 0xF8);
          un8R >>= 3;
          if((un8D >= 3) && (un8D <= 5))
          {
            un8R++;
            if(un8R > 31) un8R = 31;
          }
          un8D = un8G - (un8G & 0xF8);
          un8G >>= 3;
          if((un8D >= 3) && (un8D <= 5))
          {
            un8G++;
            if(un8G > 31) un8G = 31;
          }
          un8D = un8B - (un8B & 0xF8);
          un8B >>= 3;
          if((un8D >= 3) && (un8D <= 5))
          {
            un8B++;
            if(un8B > 31) un8B = 31;
          }
        }
        else
        {
          un8R >>= 3;
          un8G >>= 3;
          un8B >>= 3;
        }
        // Put the destination pixel
        ptTmp->un8R = un8R;
        ptTmp->un8G = un8G;
        ptTmp->un8B = un8B;
        ptTmp++;
      } 
    }
  }
  
  // ..... Free memory
  free(ptRGBQ);
  free(pucData);
  
  // ..... OK !
  return(true);

  // ..... Error reading file
ERRRead:;
  fclose(ptFile);
  VSetErrorCode(VERR_ReadFile);
  VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
  return(false);  
}
_vbool _CVortexImageBuilder::LoadVIB(const _vchar *pcFileName)
{
  // ..... Temporary variables
  _TVortexFILE *ptFile;
  _TVIBHeader  tHeader;
  _vuint32     un32Size;
  _vuchar      *pucLoad, *pucPacked, *pucUnpacked;
  _vchar       acFile[256];

  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".vib") == NULL) strcat(acFile, ".vib");
  ptFile = fopen(acFile, "rb");
  if(ptFile == NULL)
  {
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    

  // ..... Read header
  if(fread(&tHeader, sizeof(tHeader), 1, ptFile) < 1) goto ERRRead;
    
  // ..... Check header
  if((tHeader.acID[0] != _VIB_ID1_) || (tHeader.acID[1] != _VIB_ID2_) || (tHeader.acID[2] != _VIB_ID3_))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFile);
    VError(_F_, _L_, "\"%s\" is an invalid VIB file", acFile);
    return(false);
  }
  if(tHeader.un8Version != _VIB_VER_)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_IncorrectVersion);
    VError(_F_, _L_, "\"%s\" has an incorrect version", acFile);
    return(false);
  }
  if((tHeader.un8Compression != _VIB_NONE_) && (tHeader.un8Compression != _VIB_DUMMY_))
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFileHeader);
    VError(_F_, _L_, "Invalid compression type found in file \"%s\"", acFile);
    return(false);
  }
  if(tHeader.un8Reserved1 || tHeader.un32Reserved2)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_InvalidFileHeader);
    VError(_F_, _L_, "Invalid header data found in file \"%s\"", acFile);
    return(false);
  }

  // ..... Set variables
  pucPacked = NULL;
  un32Size = tHeader.un32ImageWidth * tHeader.un32ImageHeight * 3;

  // ..... Allocate memory for temporary buffer
  pucLoad = (_vuchar *) malloc(tHeader.un32SizeCompressed);
  if(pucLoad == NULL)
  {
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
    return(false);
  }
  if(tHeader.un8Compression != _VIB_NONE_)
  {
    pucPacked = (_vuchar *) malloc(tHeader.un32SizePacked);
    if(pucPacked == NULL)
    {
      free(pucLoad);
      fclose(ptFile);
      VSetErrorCode(VERR_NotEnoughMemory);
      VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
      return(false);
    }
  }
  pucUnpacked = (_vuchar *) malloc(un32Size);
  if(pucUnpacked == NULL)
  {
    if(pucPacked) free(pucPacked);
    free(pucLoad);
    fclose(ptFile);
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
    return(false);
  }

  // ..... Read image data
  if(fread(pucLoad, tHeader.un32SizeCompressed, 1, ptFile) < 1)
  { 
    if(pucPacked) free(pucPacked);
    free(pucUnpacked);
    free(pucLoad);
    goto ERRRead;
  }

  // ..... Close file
  fclose(ptFile);

  // ..... Uncompress image
  if(tHeader.un8Compression == _VIB_DUMMY_)
  {
    if(_UncompressImageBuffer_(pucPacked, pucLoad) != tHeader.un32SizePacked)
    {
      if(pucPacked) free(pucPacked);
      free(pucUnpacked);
      free(pucLoad);
      VSetErrorCode(VERR_UncompressData);
      VError(_F_, _L_, "Failed uncompressing image data");
      return(false);
    }
  }

  // ..... Unpack
  if(tHeader.un8Compression == _VIB_NONE_) pucPacked = pucLoad;
  if(!_UnpackImage_(pucUnpacked, pucPacked, un32Size, tHeader.un32SizePacked))
  {
    if(pucPacked) free(pucPacked);
    free(pucUnpacked);
    free(pucLoad);
    VSetErrorCode(VERR_UnpackData);
    VError(_F_, _L_, "Failed unpacking image data");
    return(false);    
  }

  // ..... Free memory
  if(pucPacked) free(pucPacked);
  if(pucLoad != pucPacked) free(pucLoad);
  
  // ..... Destroy previous image, if any
  FreeImage();
  
  // ..... Copy variables
  mt_un32Width  = tHeader.un32ImageWidth;
  mt_un32Height = tHeader.un32ImageHeight;
  mt_ptBuffer   = (_TVortexUN8RGB *) pucUnpacked;
  
  // ..... OK !
  return(true);
  
  // ..... Error reading file
ERRRead:;
  fclose(ptFile);
  VSetErrorCode(VERR_ReadFile);
  VError(_F_, _L_, "Failed reading file \"%s\"", acFile);
  return(false);
}
_vbool _CVortexImageBuilder::SaveVIB(const _vchar *pcFileName)
{
  // ..... Temporary variables
  _TVortexFILE *ptFile;
  _TVIBHeader  tHeader;
  _vuchar      *pucPacked, *pucCompressed;
  _vint32      un32PackedSize, un32CompressedSize;
  _vchar       acFile[256];
  
  // ..... Check
  if(mt_ptBuffer == NULL)
  { 
    VSetErrorCode(VERR_NotInitialized);
    VError(_F_, _L_, "No data to save (this _CVortexImageBuilder is not initialized)");
    return(false);
  }
  
  // ..... Allocate memory for temporary buffers
  pucPacked = (_vuchar *) malloc(mt_un32Width * mt_un32Height * 3);
  if(pucPacked == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
    return(false);
  }
  pucCompressed = (_vuchar *) malloc(mt_un32Width * mt_un32Height * 3);
  if(pucCompressed == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for temporary image buffer");
    return(false);
  }
  
  // ..... Open file
  strcpy(acFile, pcFileName);
  if(strstr(acFile, ".vib") == NULL) strcat(acFile, ".vib");
  ptFile = fopen(acFile, "wb");
  if(ptFile == NULL)
  {
    free(pucPacked);
    free(pucCompressed);
    VSetErrorCode(VERR_OpenFile);
    VError(_F_, _L_, "Failed opening file \"%s\"", acFile);
    return(false);
  }    
  
  // ..... Pack image
  un32PackedSize = _PackImage_(pucPacked, (_vuchar *) mt_ptBuffer, mt_un32Width * mt_un32Height * 3);
  
  // ..... Compress image
  un32CompressedSize = _CompressImageBuffer_(pucCompressed, pucPacked, un32PackedSize);

  // ..... Make header
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]         = _VIB_ID1_;
  tHeader.acID[1]         = _VIB_ID2_;
  tHeader.acID[2]         = _VIB_ID3_;
  tHeader.un8Version      = _VIB_VER_;
  tHeader.un32ImageWidth  = mt_un32Width;
  tHeader.un32ImageHeight = mt_un32Height;
  tHeader.un32SizePacked  = un32PackedSize;
  if(un32CompressedSize)
  {
    tHeader.un8Compression     = _VIB_DUMMY_;
    tHeader.un32SizeCompressed = un32CompressedSize;
  }
  else
  {
    tHeader.un8Compression     = _VIB_NONE_;
    tHeader.un32SizeCompressed = un32PackedSize;
  }

  // ..... Write header
  if(fwrite(&tHeader, sizeof(tHeader), 1, ptFile) < 1) goto ERRWrite;

  // ..... Write image data
  if(un32CompressedSize)
  {
    if(fwrite(pucCompressed, un32CompressedSize, 1, ptFile) < 1) goto ERRWrite;
  }
  else
    if(fwrite(pucPacked, un32PackedSize, 1, ptFile) < 1) goto ERRWrite;
  
  // ..... OK !
  free(pucCompressed);
  free(pucPacked);
  fclose(ptFile);
  return(true);
  
  // ..... Error writing file
ERRWrite:;
  free(pucCompressed);
  free(pucPacked);
  fclose(ptFile);
  VSetErrorCode(VERR_WriteFile);
  VError(_F_, _L_, "Failed writing file \"%s\"", acFile);
  return(false);
}

// ============================================================================================
// Local functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Image functions
// --------------------------------------------------------------------------------------------
// ***** For I/O functions
_vuint32 _PackImage_(_vuchar *pucDstBuff, _vuchar*pucSrcBuff, _vuint32 un32SrcSize)
{
  // ..... Temporary variables
  _vuint32 un32Size;
  _vuchar  aucP[5], aucU[8];
  _vuchar  *pucDstTmp = pucDstBuff;
  _vuchar  *pucSrcTmp = pucSrcBuff;
  _vuchar  *pucSrcMax = pucSrcBuff + un32SrcSize - 1;
  _vbool   bRun;

  // ..... Pack image
  un32Size = 0; bRun = true;
  while(bRun)
  {
    bRun = _GetByte_(aucU[0], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucU[1], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucU[2], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucU[3], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucU[4], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucU[5], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucU[6], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucU[7], pucSrcTmp, pucSrcMax);
    _PackBytes_(aucP, aucU);
    *pucDstTmp++ = aucP[0];
    *pucDstTmp++ = aucP[1];
    *pucDstTmp++ = aucP[2];
    *pucDstTmp++ = aucP[3];
    *pucDstTmp++ = aucP[4];
    un32Size += 5;
  }
  return(un32Size);
}
_vbool _UnpackImage_(_vuchar *pucDstBuff, _vuchar *pucSrcBuff, _vuint32 un32DstSize, _vuint32 un32SrcSize)
{
  // ..... Temporary variables
  _vuint32 un32Size, un32Cnt;
  _vuchar  aucP[5], aucU[8];
  _vuchar  *pucDstTmp = pucDstBuff;
  _vuchar  *pucSrcTmp = pucSrcBuff;
  _vuchar  *pucSrcMax = pucSrcBuff + un32SrcSize - 1;
  _vbool   bRun;

  // ..... Unpack image
  un32Size = 0; bRun = true;
  while(bRun)
  {
    bRun = _GetByte_(aucP[0], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucP[1], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucP[2], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucP[3], pucSrcTmp, pucSrcMax);
    bRun = _GetByte_(aucP[4], pucSrcTmp, pucSrcMax);
    _UnpackBytes_(aucU, aucP);
    for(un32Cnt = 0; un32Cnt < 8; un32Cnt++)
    {
      *pucDstTmp++ = aucU[un32Cnt];
      un32Size++;
      if(un32Size >= un32DstSize) return(true);
    }
  }
  return(false);
}
_vbool _GetByte_(_vuchar &pucDst, _vuchar *&pucSrc, _vuchar *&pucSrcMax)
{
  if(pucSrc > pucSrcMax)
  {
    pucDst = 0;
    return(false);
  }
  pucDst = *pucSrc++;
  return(true);
}
_vvoid _PackBytes_(_vuchar aucDst[5], const _vuchar aucSrc[8])
{
  aucDst[0] = (aucSrc[0] << 3) | (aucSrc[1] >> 2);
  aucDst[1] = (aucSrc[1] << 6) | (aucSrc[2] << 1) | (aucSrc[3] >> 4);
  aucDst[2] = (aucSrc[3] << 4) | (aucSrc[4] >> 1);
  aucDst[3] = (aucSrc[4] << 7) | (aucSrc[5] << 2) | (aucSrc[6] >> 3);
  aucDst[4] = (aucSrc[6] << 5) |  aucSrc[7];
}
_vvoid _UnpackBytes_(_vuchar aucDst[8], const _vuchar aucSrc[5])
{
  aucDst[0] =   aucSrc[0] >> 3;
  aucDst[1] = ((aucSrc[0] << 2) & 0x1F) | (aucSrc[1] >> 6);
  aucDst[2] =  (aucSrc[1] >> 1) & 0x1F;
  aucDst[3] = ((aucSrc[1] << 4) & 0x1F) | (aucSrc[2] >> 4);
  aucDst[4] = ((aucSrc[2] << 1) & 0x1F) | (aucSrc[3] >> 7);
  aucDst[5] =  (aucSrc[3] >> 2) & 0x1F;
  aucDst[6] = ((aucSrc[3] << 3) & 0x1F) | (aucSrc[4] >> 5);
  aucDst[7] =   aucSrc[4]       & 0x1F;
}
_vuint32 _CompressImageBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff, _vuint32 un32SrcBuffSize)
{
  // ..... Not implemented yet
  pucDstBuff      = pucDstBuff;
  pucSrcBuff      = pucSrcBuff;
  un32SrcBuffSize = un32SrcBuffSize;
  return(0);
}
_vuint32 _UncompressImageBuffer_(_vuchar *pucDstBuff, const _vuchar *pucSrcBuff)
{
  // ..... Not implemented yet
  pucDstBuff = pucDstBuff;
  pucSrcBuff = pucSrcBuff;
  return(0);
}

// ============================================================================================
// End of vimage.cpp
// ============================================================================================
