// ############################################################################################
//  vluts.cpp
//  Implementation of various LUTs
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is partialy adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
// ***** VortexGE headers
#include <vluts.h>

// ============================================================================================
// Exported variables declaration
// ============================================================================================
// ***** Math LUTs
_vfloat _afSinTable_[_DEGREECOUNT_] = {0};
_vfloat _afCosTable_[_DEGREECOUNT_] = {0};

// ***** Lighting LUTs
_vuint32 _aun32RLightTable_[32 * 32] = {0};
_vuint32 _aun32GLightTable_[32 * 32] = {0};
_vuint32 _aun32BLightTable_[32 * 32] = {0};

// ***** Fogged-lighting LUTs
_vuint32 *_pun32RFogLightTable_ = NULL;
_vuint32 *_pun32GFogLightTable_ = NULL;
_vuint32 *_pun32BFogLightTable_ = NULL;

// ***** Fogged-color LUTs
_vuint32 *_pun32RFogColorTable_ = NULL;
_vuint32 *_pun32GFogColorTable_ = NULL;
_vuint32 *_pun32BFogColorTable_ = NULL;

// ***** addition LUT
_vuint8 _aun8CAddTable_[32 * 32] = {0};

// ***** Transparency LUTs
_vuint8 _aun8CMeanTable_[32 * 32]       = {0};
_vuint8 _aun8TLightTable_[32 * 32 * 32] = {0};
_vuint8 *_pun8TRFogColorTable_          = NULL;
_vuint8 *_pun8TGFogColorTable_          = NULL;
_vuint8 *_pun8TBFogColorTable_          = NULL;
_vuint8 *_pun8TRFogLightTable_          = NULL;
_vuint8 *_pun8TGFogLightTable_          = NULL;
_vuint8 *_pun8TBFogLightTable_          = NULL;

// ============================================================================================
// Exported functions implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// Table generation
// --------------------------------------------------------------------------------------------
// ***** LUT initialization functions
_vvoid _VInitSinCosTable_(_vvoid)
{
  // ..... Static variables
  static _vbool bInitialized = false;
  
  // ..... Check
  if(bInitialized) return;
    
  // ..... Generate the table
  _vfloat fUnit = _vfloat(3.141592654) * _vfloat(2.0) / _vfloat(_DEGREECOUNT_);
  for(_vuint16 un16Cnt = 0; un16Cnt < _DEGREECOUNT_; un16Cnt++)
  {
    _afSinTable_[un16Cnt] = (_vfloat) sin(fUnit * _vfloat(un16Cnt));
    _afCosTable_[un16Cnt] = (_vfloat) cos(fUnit * _vfloat(un16Cnt));
  }
  bInitialized = true;
}
_vvoid _VInitLightingTable_(_vvoid)
{
  // ..... Static variables
  static _vbool bInitialized = false;

  // ..... Temporary variables
  _vuint32 un32C1, un32C2, un32OC, un32T;
  
  // ..... Check
  if(bInitialized) return;
    
  // ..... Generate the lighting LUTs
  for(un32C1 = 0; un32C1 < 32; un32C1++)
    for(un32C2 = 0; un32C2 < 32; un32C2++)
    {
      un32T = (un32C1 * un32C2) / 31;
      _aun32RLightTable_[(un32C1 << 5) | un32C2] = un32T << 19;
      _aun32GLightTable_[(un32C1 << 5) | un32C2] = un32T << 11;
      _aun32BLightTable_[(un32C1 << 5) | un32C2] = un32T << 3;
    }
  
  // ..... Generate the mean and addition LUTs
  for(_vuint32 un32C1 = 0; un32C1 < 32; un32C1++)
    for(_vuint32 un32C2 = 0; un32C2 < 32; un32C2++)
    {
      // Mean LUT
      _aun8CMeanTable_[(un32C1 << 5) | un32C2] = _vuint8((un32C1 + un32C2) / 2) << 3;
      // Addition LUT
      _aun8CAddTable_[(un32C1 << 5) | un32C2] = _vuint8(un32C1 + un32C2);
      if(_aun8CAddTable_[(un32C1 << 5) | un32C2] > 31) _aun8CAddTable_[(un32C1 << 5) | un32C2] = 31;
      _aun8CAddTable_[(un32C1 << 5) | un32C2] <<= 3;
    }

  // ..... Generate the lighting LUTs with transparency
  for(un32C1 = 0; un32C1 < 32; un32C1++)
    for(un32C2 = 0; un32C2 < 32; un32C2++)
    {
      un32T = (un32C1 * un32C2) / 31;
      for(un32OC = 0; un32OC < 32; un32OC++)
        _aun8TLightTable_[(un32OC << 10) | (un32C1 << 5) | un32C2] = _vuint8((un32T + un32OC) / 2) << 3;
    }

  // ..... OK !
  bInitialized = true;
}

// ============================================================================================
// End of vluts.cpp
// ============================================================================================
