// ############################################################################################
//  vluts.h
//  Declaration of various LUTs
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is partialy adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VLUTS_H_
#define _VLUTS_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** VortexGE headers
#include <vbtypes.h>

// ============================================================================================
// Defines
// ============================================================================================
// ***** Table size, must be power of two
#define _DEGREECOUNT_ 4096

// ***** Access to sin and cos table
#define SIN(n32Val) (_afSinTable_[n32Val & (_DEGREECOUNT_ - 1)]);
#define COS(n32Val) (_afCosTable_[n32Val & (_DEGREECOUNT_ - 1)]);

// ============================================================================================
// Exported functions prototype
// ============================================================================================
// ***** LUT initialization functions
extern _vvoid _VInitSinCosTable_(_vvoid);
extern _vvoid _VInitLightingTable_(_vvoid);

// ============================================================================================
// Reference to External variables
// ============================================================================================
// ***** Math LUTs
// These LUTs contains the sine and cosine values from 0.1 to 360.0 degree
extern _vfloat _afSinTable_[_DEGREECOUNT_];
extern _vfloat _afCosTable_[_DEGREECOUNT_];

// ***** Lighting LUTs
//  Each LUT represent R, G, and B components of this following lighting equation :
//    LUT[(C1 << 5) | C2] = (C1 * C2) / 31
//  C1 and C2 are swapable.
//  If C1 is assumed as the panel's color then C2 is the light's color.
//  If C2 is assumed as the panel's color then C1 is the light's color.
extern _vuint32 _aun32RLightTable_[32 * 32];
extern _vuint32 _aun32GLightTable_[32 * 32];
extern _vuint32 _aun32BLightTable_[32 * 32];

// ***** Fogged-lighting LUTs
//  Each LUT represent R, G, and B components of this following fogged-lighting equation :
//    LUT[(FI << 10) | (C1 << 5) | C2] = ((FI * FC) + ((31 - FI) * CT)) / 31;
//  with :
//    CT = ((C1 * C2) / 31) & 0x1F
//  FI is the fog's intensity and FC is the fog's color for each R, G, B components.
//  C1 and C2 are the same as in the lighting LUTs.
extern _vuint32 *_pun32RFogLightTable_;
extern _vuint32 *_pun32GFogLightTable_;
extern _vuint32 *_pun32BFogLightTable_;

// ***** Fogged-color LUTs
//  Each LUT represent R, G, and B components of this following fogged-color equation :
//    LUT[(FI << 5) | CL] = ((FI * FC) + ((31 - FI) * CL)) / 31
//  FI is the fog intensity and FC is the fog color for each R, G, B components.
//  CL can be the panel's color or the light's color.
extern _vuint32 *_pun32RFogColorTable_;
extern _vuint32 *_pun32GFogColorTable_;
extern _vuint32 *_pun32BFogColorTable_;

// ***** Color addition LUT
//  This LUT represent the result of this following addition equation :
//    LUT[(C1 << 5) | C2] = C1 + C2
//  with restriction :
//    0<= LUT[*] <= 31
//  C1 and C2 are swapable.
extern _vuint8 _aun8CAddTable_[32 * 32];

// ***** Transparency LUTs
//  This LUT below represent the result of this following mean equation :
//    LUT[(C1 << 5) | C2] = ((C1 + C2) / 2) << 3
//  C1 and C2 are swapable.
extern _vuint8 _aun8CMeanTable_[32 * 32];
//
//  This LUT bellow represent any color component of this following lighting equation :
//   LUT[(OC << 10) | (C1 << 5) | C2] = ((OC + NC) / 2) << 3
//  with :
//    NC = (C1 * C2) / 31
//  OC and NC are the colors to be combined.
//  C1 and C2 are the same as in the lighting LUTs.
extern _vuint8 _aun8TLightTable_[32 * 32 * 32];
//
//  Each LUT bellow represent any color component of this following lighting equation :
//   LUT[(OC << 10) | (FI << 5) | PC] = ((OC + NC) / 2) << 3
//  with :
//    NC = ((FI * FC) + ((31 - FI) * PC)) / 31
//  OC and NC are the colors to be combined.
//  FI is the fog's intensity and FC is the fog's color for any component.
//  PC is the panel's color
extern _vuint8 *_pun8TRFogColorTable_;
extern _vuint8 *_pun8TGFogColorTable_;
extern _vuint8 *_pun8TBFogColorTable_;
//
//  Each LUT bellow represent any color component of this following lighting equation :
//   LUT[(OC << 15) | (FI << 10) | (C1 << 5) | C2] = ((OC + NC) / 2) << 3
//  with :
//    NC = ((FI * FC) + ((31 - FI) * CL)) / 31
//    CL = (C1 * C2) / 31 
//  OC and NC are the colors to be combined.
//  FI is the fog's intensity and FC is the fog's color for any component.
//  C1 and C2 are the same as in the lighting LUTs.
extern _vuint8 *_pun8TRFogLightTable_;
extern _vuint8 *_pun8TGFogLightTable_;
extern _vuint8 *_pun8TBFogLightTable_;

#endif
// ============================================================================================
// End of vluts.h
// ============================================================================================
