// ############################################################################################
//  vpoint3d.h
//  Definition & implementation of _CVortexPoint3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VPOINT3D_H_
#define _VPOINT3D_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
// ***** VortexGE headers
#include <vfastmem.h>

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Triplet float XYZ
struct _TVortexFXYZ
{
  _vfloat fX, fY, fZ;
#ifdef _ASM_SSE_
  _vfloat _fdummy_to_make_this_structure_128_bits_long_;
#endif
};

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The 3D point class
class _CVortexPoint3D
{
  public:
    // Variables
    _TVortexFXYZ _sse m_tLocal; // The local (untransformed) XYZ
    _TVortexFXYZ _sse m_tWorld; // The world (transformed) XYZ
    // Constructors
    inline _CVortexPoint3D();
    inline _CVortexPoint3D(_vfloat fLocalXYZ);
    inline _CVortexPoint3D(_vfloat fLocalX, _vfloat fLocalY, _vfloat fLocalZ);
    inline _CVortexPoint3D(const _TVortexFXYZ &tLocal);
    inline _CVortexPoint3D(const _CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D(const _CVortexPoint3D *pcsVPoint3D);
    // Math functions
    inline _vfloat Mag(void);
    inline _vfloat DotUnit(_CVortexPoint3D &csVPoint3D);
    inline _vfloat DotNonUnit(_CVortexPoint3D &csVPoint3D);
    // Overloaded operators
    inline _vbool           operator == (_CVortexPoint3D &csVPoint3D);
    inline _vbool           operator != (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D  operator +  (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D  operator -  (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D  operator *  (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D  operator /  (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D &operator += (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D &operator -= (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D &operator *= (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D &operator /= (_CVortexPoint3D &csVPoint3D);
    inline _CVortexPoint3D  operator +  (_vfloat fValue);
    inline _CVortexPoint3D  operator -  (_vfloat fValue);
    inline _CVortexPoint3D  operator *  (_vfloat fValue);
    inline _CVortexPoint3D  operator /  (_vfloat fValue);
    inline _CVortexPoint3D &operator += (_vfloat fValue);
    inline _CVortexPoint3D &operator -= (_vfloat fValue);
    inline _CVortexPoint3D &operator *= (_vfloat fValue);
    inline _CVortexPoint3D &operator /= (_vfloat fValue);
    // Special overloaded operators (to support aligned memory allocation for SSE)
    inline _vvoid *operator new      (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid *operator new[]    (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
    inline _vvoid  operator delete   (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
    inline _vvoid  operator delete[] (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
};

// ============================================================================================
// Exported functions prototype
// ============================================================================================
// ***** Vertex helper functions
extern _vbool   UniqueVert(_CVortexPoint3D &csVPoint3D, _CVortexPoint3D *pcsList, _vuint32 un32Range);
extern _vuint32 GetVertIndex(_CVortexPoint3D &csVPoint3D, _CVortexPoint3D *pcsList, _vuint32 un32Range);

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 3D point class
// --------------------------------------------------------------------------------------------
// ***** Constructors
_CVortexPoint3D::_CVortexPoint3D()
{
  m_tLocal.fX = m_tLocal.fY = m_tLocal.fZ = 0;
}
_CVortexPoint3D::_CVortexPoint3D(_vfloat fLocalXYZ)
{
  m_tLocal.fX = m_tLocal.fY = m_tLocal.fZ = fLocalXYZ;
}
_CVortexPoint3D::_CVortexPoint3D(_vfloat fLocalX, _vfloat fLocalY, _vfloat fLocalZ)
{
  m_tLocal.fX = fLocalX;
  m_tLocal.fY = fLocalY;
  m_tLocal.fZ = fLocalZ;
}
_CVortexPoint3D::_CVortexPoint3D(const _TVortexFXYZ &tLocal)
{
  m_tLocal.fX = tLocal.fX;
  m_tLocal.fY = tLocal.fY;
  m_tLocal.fZ = tLocal.fZ;
}
_CVortexPoint3D::_CVortexPoint3D(const _CVortexPoint3D &csVPoint3D)
{
  m_tLocal.fX = csVPoint3D.m_tLocal.fX;
  m_tLocal.fY = csVPoint3D.m_tLocal.fY;
  m_tLocal.fZ = csVPoint3D.m_tLocal.fZ;
}
_CVortexPoint3D::_CVortexPoint3D(const _CVortexPoint3D *pcsVPoint3D)
{
  m_tLocal.fX = pcsVPoint3D->m_tLocal.fX;
  m_tLocal.fY = pcsVPoint3D->m_tLocal.fY;
  m_tLocal.fZ = pcsVPoint3D->m_tLocal.fZ;
}

// ***** Math functions
_vfloat _CVortexPoint3D::Mag(void)
{
  return(sqrt(m_tLocal.fX * m_tLocal.fX +
              m_tLocal.fY * m_tLocal.fY +
              m_tLocal.fZ * m_tLocal.fZ));
}
_vfloat _CVortexPoint3D::DotUnit(_CVortexPoint3D &csVPoint3D)
{
  return(m_tLocal.fX * csVPoint3D.m_tLocal.fX + 
         m_tLocal.fY * csVPoint3D.m_tLocal.fY +
         m_tLocal.fZ * csVPoint3D.m_tLocal.fZ);
}
_vfloat _CVortexPoint3D::DotNonUnit(_CVortexPoint3D &csVPoint3D)
{
  return((m_tLocal.fX * csVPoint3D.m_tLocal.fX + 
          m_tLocal.fY * csVPoint3D.m_tLocal.fY +
          m_tLocal.fZ * csVPoint3D.m_tLocal.fZ) /
          (Mag() * csVPoint3D.Mag()));
}

// ***** Overloaded operators
_vbool _CVortexPoint3D::operator == (_CVortexPoint3D &csVPoint3D)
{
  return((m_tLocal.fX == csVPoint3D.m_tLocal.fX) &&
         (m_tLocal.fY == csVPoint3D.m_tLocal.fY) &&
         (m_tLocal.fZ == csVPoint3D.m_tLocal.fZ));
}
_vbool _CVortexPoint3D::operator != (_CVortexPoint3D &csVPoint3D)
{
  return((m_tLocal.fX != csVPoint3D.m_tLocal.fX) ||
         (m_tLocal.fY != csVPoint3D.m_tLocal.fY) ||
         (m_tLocal.fZ != csVPoint3D.m_tLocal.fZ));
}
_CVortexPoint3D _CVortexPoint3D::operator + (_CVortexPoint3D &csVPoint3D)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX + csVPoint3D.m_tLocal.fX;
  clsTmp.m_tLocal.fY = m_tLocal.fY + csVPoint3D.m_tLocal.fY;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ + csVPoint3D.m_tLocal.fZ;
  return(clsTmp);
}
_CVortexPoint3D _CVortexPoint3D::operator - (_CVortexPoint3D &csVPoint3D)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX - csVPoint3D.m_tLocal.fX;
  clsTmp.m_tLocal.fY = m_tLocal.fY - csVPoint3D.m_tLocal.fY;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ - csVPoint3D.m_tLocal.fZ;
  return(clsTmp);
}
_CVortexPoint3D _CVortexPoint3D::operator * (_CVortexPoint3D &csVPoint3D)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX * csVPoint3D.m_tLocal.fX;
  clsTmp.m_tLocal.fY = m_tLocal.fY * csVPoint3D.m_tLocal.fY;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ * csVPoint3D.m_tLocal.fZ;
  return(clsTmp);
}
_CVortexPoint3D _CVortexPoint3D::operator / (_CVortexPoint3D &csVPoint3D)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX / csVPoint3D.m_tLocal.fX;
  clsTmp.m_tLocal.fY = m_tLocal.fY / csVPoint3D.m_tLocal.fY;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ / csVPoint3D.m_tLocal.fZ;
  return(clsTmp);

}
_CVortexPoint3D &_CVortexPoint3D::operator += (_CVortexPoint3D &csVPoint3D)
{
  m_tLocal.fX += csVPoint3D.m_tLocal.fX;
  m_tLocal.fY += csVPoint3D.m_tLocal.fY;
  m_tLocal.fZ += csVPoint3D.m_tLocal.fZ;
  return(*this);
}
_CVortexPoint3D &_CVortexPoint3D::operator -= (_CVortexPoint3D &csVPoint3D)
{
  m_tLocal.fX -= csVPoint3D.m_tLocal.fX;
  m_tLocal.fY -= csVPoint3D.m_tLocal.fY;
  m_tLocal.fZ -= csVPoint3D.m_tLocal.fZ;
  return(*this);
}
_CVortexPoint3D &_CVortexPoint3D::operator *= (_CVortexPoint3D &csVPoint3D)
{
  m_tLocal.fX *= csVPoint3D.m_tLocal.fX;
  m_tLocal.fY *= csVPoint3D.m_tLocal.fY;
  m_tLocal.fZ *= csVPoint3D.m_tLocal.fZ;
  return(*this);
}
_CVortexPoint3D &_CVortexPoint3D::operator /= (_CVortexPoint3D &csVPoint3D)
{
  m_tLocal.fX /= csVPoint3D.m_tLocal.fX;
  m_tLocal.fY /= csVPoint3D.m_tLocal.fY;
  m_tLocal.fZ /= csVPoint3D.m_tLocal.fZ;
  return(*this);
}
_CVortexPoint3D _CVortexPoint3D::operator + (_vfloat fValue)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX + fValue;
  clsTmp.m_tLocal.fY = m_tLocal.fY + fValue;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ + fValue;
  return(clsTmp);
}
_CVortexPoint3D _CVortexPoint3D::operator - (_vfloat fValue)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX - fValue;
  clsTmp.m_tLocal.fY = m_tLocal.fY - fValue;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ - fValue;
  return(clsTmp);
}
_CVortexPoint3D _CVortexPoint3D::operator * (_vfloat fValue)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX * fValue;
  clsTmp.m_tLocal.fY = m_tLocal.fY * fValue;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ * fValue;
  return(clsTmp);
}
_CVortexPoint3D _CVortexPoint3D::operator / (_vfloat fValue)
{
  _CVortexPoint3D clsTmp;
  clsTmp.m_tLocal.fX = m_tLocal.fX / fValue;
  clsTmp.m_tLocal.fY = m_tLocal.fY / fValue;
  clsTmp.m_tLocal.fZ = m_tLocal.fZ / fValue;
  return(clsTmp);
}
_CVortexPoint3D &_CVortexPoint3D::operator += (_vfloat fValue)
{
  m_tLocal.fX += fValue;
  m_tLocal.fY += fValue;
  m_tLocal.fZ += fValue;
  return(*this);
}
_CVortexPoint3D &_CVortexPoint3D::operator -= (_vfloat fValue)
{
  m_tLocal.fX -= fValue;
  m_tLocal.fY -= fValue;
  m_tLocal.fZ -= fValue;
  return(*this);
}
_CVortexPoint3D &_CVortexPoint3D::operator *= (_vfloat fValue)
{
  m_tLocal.fX *= fValue;
  m_tLocal.fY *= fValue;
  m_tLocal.fZ *= fValue;
  return(*this);
}
_CVortexPoint3D &_CVortexPoint3D::operator /= (_vfloat fValue)
{
  m_tLocal.fX /= fValue;
  m_tLocal.fY /= fValue;
  m_tLocal.fZ /= fValue;
  return(*this);
}

#endif
// ============================================================================================
// End of vpoint3d.h
// ============================================================================================
