// ############################################################################################
//  vspr2d.cpp
//  Implementation of _CVortexStaticSprite2D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
// ***** VortexGE headers
#include <v3denv.h>
#include <vimage.h>
#include <vmdisp.h>
#include <vspr2d.h>

// ============================================================================================
// Exported classes implementation
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 2D static sprite class
// --------------------------------------------------------------------------------------------
// ***** Constructor & destructor
_CVortexStaticSprite2D::_CVortexStaticSprite2D()
{
  mt_ptData         = NULL;
  mt_pcsImageSprite = NULL;
  mt_ptOmni         = NULL;
  mt_un8TotData     = 0;
}
_CVortexStaticSprite2D::~_CVortexStaticSprite2D()
{
  if(mt_ptData) delete [] mt_ptData;
}

// ***** Initialization functions
_vbool _CVortexStaticSprite2D::Initialize(_vfloat *pfPosXYZs, _vuint8 un8PosCount, _CVortexImage *pcsImageSprite, 
                                          _EVortexPivot ePivot, _vuint32 un32ExcludedColorVRGB,
                                          _TVortexAmbientLight *ptVAmbientLight, 
                                          _TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat *pfTmp = pfPosXYZs;
  
  // ..... Allocate memory for data buffer
  mt_ptData = new _TSSpriteData[un8PosCount];
  if(mt_ptData == NULL)
  {
    VSetErrorCode(VERR_NotEnoughMemory);
    VError(_F_, _L_, "Failed to allocate memory for sprite's data buffer");
    return(false);
  }

  // ..... Copy variables
  mt_pcsImageSprite = pcsImageSprite;
  if(un32ExcludedColorVRGB == EXCLUDED_COLOR_FIRST_PIXEL)
    mt_un32ExclColor = *((_vuint32 *) mt_pcsImageSprite->mt_pun8Buffer);
  else
    mt_un32ExclColor = un32ExcludedColorVRGB;
  mt_un8TotData = un8PosCount;
  mt_ptOmni = ptVOmniLight;
  
  // ..... Calculate position offset for pivoting
  mt_fXo = mt_fYo = _vfloat(0.0);
  if(ePivot == ETopLeft)
  {
    mt_fXo = _vfloat(mt_pcsImageSprite->mt_un32Width >> 1);
    mt_fYo = -_vfloat(mt_pcsImageSprite->mt_un32Height >> 1);
  }
  else if(ePivot == ETopCenter)
    mt_fYo = -_vfloat(mt_pcsImageSprite->mt_un32Height >> 1);
  else if(ePivot == ETopRight)
  {
    mt_fXo = -_vfloat(mt_pcsImageSprite->mt_un32Width >> 1);
    mt_fYo = -_vfloat(mt_pcsImageSprite->mt_un32Height >> 1);
  }
  else if(ePivot == ECenterLeft)
    mt_fXo = _vfloat(mt_pcsImageSprite->mt_un32Width >> 1);
  else if(ePivot == ECenterRight)
    mt_fXo = -_vfloat(mt_pcsImageSprite->mt_un32Width >> 1);
  else if(ePivot == EBottomLeft)
  {
    mt_fXo = _vfloat(mt_pcsImageSprite->mt_un32Width >> 1);
    mt_fYo = _vfloat(mt_pcsImageSprite->mt_un32Height >> 1);
  }
  else if(ePivot == EBottomCenter)
    mt_fYo = _vfloat(mt_pcsImageSprite->mt_un32Height >> 1);
  else if(ePivot == EBottomRight)
  {
    mt_fXo = -_vfloat(mt_pcsImageSprite->mt_un32Width >> 1);
    mt_fYo = _vfloat(mt_pcsImageSprite->mt_un32Height >> 1);
  }
  
  // ..... Set initial positions and normals
  for(_vuint8 un8Cnt = 0; un8Cnt < un8PosCount; un8Cnt++)
  {
    // Set normal
    mt_ptData[un8Cnt].fX = _vfloat(0.0);
    mt_ptData[un8Cnt].fY = _vfloat(0.316227766);
    mt_ptData[un8Cnt].fZ = _vfloat(0.948683298);
    
    // Set position
    mt_ptData[un8Cnt].csPos.m_tLocal.fX = (*pfTmp++) + mt_fXo;
    mt_ptData[un8Cnt].csPos.m_tLocal.fY = (*pfTmp++) + mt_fYo;
    mt_ptData[un8Cnt].csPos.m_tLocal.fZ = *pfTmp++;
  }
  
  // ..... Calc light intensities
  _CalcIntenAmbient_(ptVAmbientLight);
  _CalcIntenOmni_(ptVOmniLight);
  _CalcInten_();

  // ..... OK !
  return(true);
}
_vvoid _CVortexStaticSprite2D::_CalcIntenOmni_(_TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat fXm, fYm, fZm, fMag, fCos;

  // ..... Calculate omni light intensity
  for(_vuint8 un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // ..... Use the supplied omni light
    if(ptVOmniLight)
    {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = ptVOmniLight->fX - mt_ptData[un8Cnt].csPos.m_tLocal.fX;
      fYm = ptVOmniLight->fY - mt_ptData[un8Cnt].csPos.m_tLocal.fY;
      fZm = ptVOmniLight->fZ - mt_ptData[un8Cnt].csPos.m_tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;
    
      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = mt_ptData[un8Cnt].fX * fXm + mt_ptData[un8Cnt].fY * fYm - mt_ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;
    
      // Calculate the color value
      mt_ptData[un8Cnt].tOmni.un8R = _vuint8(ptVOmniLight->un8R * fCos);
      mt_ptData[un8Cnt].tOmni.un8G = _vuint8(ptVOmniLight->un8G * fCos);
      mt_ptData[un8Cnt].tOmni.un8B = _vuint8(ptVOmniLight->un8B * fCos);
    }
    // ..... Use the default omni light from the active 3D environment
    else
    {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = _pcsAct3DEnv_->mt_tDefOLight.fX - mt_ptData[un8Cnt].csPos.m_tLocal.fX;
      fYm = _pcsAct3DEnv_->mt_tDefOLight.fY - mt_ptData[un8Cnt].csPos.m_tLocal.fY;
      fZm = _pcsAct3DEnv_->mt_tDefOLight.fZ - mt_ptData[un8Cnt].csPos.m_tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;

      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = mt_ptData[un8Cnt].fX * fXm + mt_ptData[un8Cnt].fY * fYm - mt_ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;

      // Calculate the color value
      mt_ptData[un8Cnt].tOmni.un8R = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8R * fCos);
      mt_ptData[un8Cnt].tOmni.un8G = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8G * fCos);
      mt_ptData[un8Cnt].tOmni.un8B = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8B * fCos);
    }
  }
}
_vvoid _CVortexStaticSprite2D::_CalcIntenOmni_(_vuint8 un8Index, _TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat fXm, fYm, fZm, fMag, fCos;

  // ..... Use the supplied omni light
  if(ptVOmniLight)
  {
    // Create the normalized light vector from the panel center to the omni light position
    fXm = ptVOmniLight->fX - mt_ptData[un8Index].csPos.m_tLocal.fX;
    fYm = ptVOmniLight->fY - mt_ptData[un8Index].csPos.m_tLocal.fY;
    fZm = ptVOmniLight->fZ - mt_ptData[un8Index].csPos.m_tLocal.fZ;
    fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
    fYm /= fMag;
    fZm /= fMag;
    
    // Calculate the angle between sprite's normal (fixed) and the light vector
    fCos = mt_ptData[un8Index].fX * fXm + mt_ptData[un8Index].fY * fYm - mt_ptData[un8Index].fZ * fZm;
    fCos = (fCos < 0) ? 0 : fCos;
    
    // Calculate the color value
    mt_ptData[un8Index].tOmni.un8R = _vuint8(ptVOmniLight->un8R * fCos);
    mt_ptData[un8Index].tOmni.un8G = _vuint8(ptVOmniLight->un8G * fCos);
    mt_ptData[un8Index].tOmni.un8B = _vuint8(ptVOmniLight->un8B * fCos);
  }
  // ..... Use the default omni light from the active 3D environment
  else
  {
    // Create the normalized light vector from the panel center to the omni light position
    fXm = _pcsAct3DEnv_->mt_tDefOLight.fX - mt_ptData[un8Index].csPos.m_tLocal.fX;
    fYm = _pcsAct3DEnv_->mt_tDefOLight.fY - mt_ptData[un8Index].csPos.m_tLocal.fY;
    fZm = _pcsAct3DEnv_->mt_tDefOLight.fZ - mt_ptData[un8Index].csPos.m_tLocal.fZ;
    fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
    fYm /= fMag;
    fZm /= fMag;

    // Calculate the angle between sprite's normal (fixed) and the light vector
    fCos = mt_ptData[un8Index].fX * fXm + mt_ptData[un8Index].fY * fYm - mt_ptData[un8Index].fZ * fZm;
    fCos = (fCos < 0) ? 0 : fCos;

    // Calculate the color value
    mt_ptData[un8Index].tOmni.un8R = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8R * fCos);
    mt_ptData[un8Index].tOmni.un8G = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8G * fCos);
    mt_ptData[un8Index].tOmni.un8B = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8B * fCos);
  }
}
_vvoid _CVortexStaticSprite2D::_CalcIntenOmni_(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _TVortexOmniLight *ptVOmniLight)
{
  // ..... Temporary variables
  _vfloat fXm, fYm, fZm, fMag, fCos;

  // ..... Calculate omni light intensity
  for(_vuint8 un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++)
  {
    // ..... Use the supplied omni light
    if(ptVOmniLight)
    {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = ptVOmniLight->fX - mt_ptData[un8Cnt].csPos.m_tLocal.fX;
      fYm = ptVOmniLight->fY - mt_ptData[un8Cnt].csPos.m_tLocal.fY;
      fZm = ptVOmniLight->fZ - mt_ptData[un8Cnt].csPos.m_tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;
    
      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = mt_ptData[un8Cnt].fX * fXm + mt_ptData[un8Cnt].fY * fYm - mt_ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;
    
      // Calculate the color value
      mt_ptData[un8Cnt].tOmni.un8R = _vuint8(ptVOmniLight->un8R * fCos);
      mt_ptData[un8Cnt].tOmni.un8G = _vuint8(ptVOmniLight->un8G * fCos);
      mt_ptData[un8Cnt].tOmni.un8B = _vuint8(ptVOmniLight->un8B * fCos);
    }
    // ..... Use the default omni light from the active 3D environment
    else
    {
      // Create the normalized light vector from the panel center to the omni light position
      fXm = _pcsAct3DEnv_->mt_tDefOLight.fX - mt_ptData[un8Cnt].csPos.m_tLocal.fX;
      fYm = _pcsAct3DEnv_->mt_tDefOLight.fY - mt_ptData[un8Cnt].csPos.m_tLocal.fY;
      fZm = _pcsAct3DEnv_->mt_tDefOLight.fZ - mt_ptData[un8Cnt].csPos.m_tLocal.fZ;
      fMag = sqrt(fXm * fXm + fYm * fYm + fZm * fZm);
      fYm /= fMag;
      fZm /= fMag;

      // Calculate the angle between sprite's normal (fixed) and the light vector
      fCos = mt_ptData[un8Cnt].fX * fXm + mt_ptData[un8Cnt].fY * fYm - mt_ptData[un8Cnt].fZ * fZm;
      fCos = (fCos < 0) ? 0 : fCos;

      // Calculate the color value
      mt_ptData[un8Cnt].tOmni.un8R = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8R * fCos);
      mt_ptData[un8Cnt].tOmni.un8G = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8G * fCos);
      mt_ptData[un8Cnt].tOmni.un8B = _vuint8(_pcsAct3DEnv_->mt_tDefOLight.un8B * fCos);
    }
  }
}

// ***** Normal rendering functions
_vvoid _CVortexStaticSprite2D::SNRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur;
  _vuint32 un32YIndex, *pun32Buff, *pun32Tex;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun32Tex = (_vuint32 *) &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_;
          if(*(pun32Tex + n32Ur) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *pun32Buff = *(pun32Tex + n32Ur);
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}
_vvoid _CVortexStaticSprite2D::SFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur;
  _vuint32 un32YIndex, *pun32Buff;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex   = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *pun32Buff = _aun32RLightTable_[(_vuint32(pun8Tex[n32Ur + 2]) << 2) | mt_ptData[un8Cnt].tTotal.un8R] |
                         _aun32GLightTable_[(_vuint32(pun8Tex[n32Ur + 1]) << 2) | mt_ptData[un8Cnt].tTotal.un8G] |
                         _aun32BLightTable_[(_vuint32(pun8Tex[n32Ur + 0]) << 2) | mt_ptData[un8Cnt].tTotal.un8B];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}
_vvoid _CVortexStaticSprite2D::TNRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex, *pun8Buff;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur;
  _vuint32 un32YIndex;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun8Buff = &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex  = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *(pun8Buff + 2) = _aun8CMeanTable_[(_vuint32(pun8Tex[n32Ur + 2]) << 2) | (*(pun8Buff + 2) >> 3)];
            *(pun8Buff + 1) = _aun8CMeanTable_[(_vuint32(pun8Tex[n32Ur + 1]) << 2) | (*(pun8Buff + 1) >> 3)];
            *(pun8Buff + 0) = _aun8CMeanTable_[(_vuint32(pun8Tex[n32Ur + 0]) << 2) | (*(pun8Buff + 0) >> 3)];
          }
        }
          
        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}
_vvoid _CVortexStaticSprite2D::TFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex, *pun8Buff;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur;
  _vuint32 un32YIndex;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun8Buff = &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex  = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *(pun8Buff + 2) = _aun8TLightTable_[(*(pun8Buff + 2) << 7) | (_vuint32(pun8Tex[n32Ur + 2]) << 2) | mt_ptData[un8Cnt].tTotal.un8R];
            *(pun8Buff + 1) = _aun8TLightTable_[(*(pun8Buff + 1) << 7) | (_vuint32(pun8Tex[n32Ur + 1]) << 2) | mt_ptData[un8Cnt].tTotal.un8G];
            *(pun8Buff + 0) = _aun8TLightTable_[(*(pun8Buff + 0) << 7) | (_vuint32(pun8Tex[n32Ur + 0]) << 2) | mt_ptData[un8Cnt].tTotal.un8B];
          }
        }
          
        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// **** Fogged rendering functions
_vvoid _CVortexStaticSprite2D::FogSNRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur, n32Fog;
  _vuint32 un32YIndex, *pun32Buff;
  _vfloat  fX, fY, fW, fH;

  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Calculate fog intensity
    n32Fog = _vint32(mt_ptData[un8Cnt].csPos.m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    n32Fog <<= 5;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex   = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *pun32Buff = _pun32RFogColorTable_[n32Fog | (_vuint32(pun8Tex[n32Ur + 2]) >> 3)] |
                         _pun32GFogColorTable_[n32Fog | (_vuint32(pun8Tex[n32Ur + 1]) >> 3)] |
                         _pun32BFogColorTable_[n32Fog | (_vuint32(pun8Tex[n32Ur + 0]) >> 3)];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}
_vvoid _CVortexStaticSprite2D::FogSFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur, n32Fog;
  _vuint32 un32YIndex, *pun32Buff;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Calculate fog intensity
    n32Fog = _vint32(mt_ptData[un8Cnt].csPos.m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    n32Fog <<= 10;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U      = n32TexLeft << _SSTEP_PREC_;
      pn32Z     = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun32Buff = (_vuint32 *) &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex   = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *pun32Buff = _pun32RFogLightTable_[n32Fog | (_vuint32(pun8Tex[n32Ur + 2]) << 2) | mt_ptData[un8Cnt].tTotal.un8R] |
                         _pun32GFogLightTable_[n32Fog | (_vuint32(pun8Tex[n32Ur + 1]) << 2) | mt_ptData[un8Cnt].tTotal.un8G] |
                         _pun32BFogLightTable_[n32Fog | (_vuint32(pun8Tex[n32Ur + 0]) << 2) | mt_ptData[un8Cnt].tTotal.un8B];
          }
        }
        n32U += n32UStep;
        pn32Z++;
        pun32Buff++;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}
_vvoid _CVortexStaticSprite2D::FogTNRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex, *pun8Buff;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur, n32Fog;
  _vuint32 un32YIndex;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Calculate fog intensity
    n32Fog = _vint32(mt_ptData[un8Cnt].csPos.m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    n32Fog <<= 5;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun8Buff = &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex  = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];

      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *(pun8Buff + 2) = _pun8TRFogColorTable_[(*(pun8Buff + 2) << 7) | n32Fog | (_vuint32(pun8Tex[n32Ur + 2]) >> 3)];
            *(pun8Buff + 1) = _pun8TGFogColorTable_[(*(pun8Buff + 1) << 7) | n32Fog | (_vuint32(pun8Tex[n32Ur + 1]) >> 3)];
            *(pun8Buff + 0) = _pun8TBFogColorTable_[(*(pun8Buff + 0) << 7) | n32Fog | (_vuint32(pun8Tex[n32Ur + 0]) >> 3)];
          }
        }
          
        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}
_vvoid _CVortexStaticSprite2D::FogTFRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex, *pun8Buff;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur, n32Fog;
  _vuint32 un32YIndex;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Calculate fog intensity
    n32Fog = _vint32(mt_ptData[un8Cnt].csPos.m_tWorld.fZ - _pcsAct3DEnv_-> mt_un32MinZ) >> _pcsAct3DEnv_->mt_un32FogShift;
    n32Fog <<= 10;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun8Buff = &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex  = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *(pun8Buff + 2) = _pun8TRFogLightTable_[(*(pun8Buff + 2) << 12) | n32Fog | (_vuint32(pun8Tex[n32Ur + 2]) << 2) | mt_ptData[un8Cnt].tTotal.un8R];
            *(pun8Buff + 1) = _pun8TGFogLightTable_[(*(pun8Buff + 1) << 12) | n32Fog | (_vuint32(pun8Tex[n32Ur + 1]) << 2) | mt_ptData[un8Cnt].tTotal.un8G];
            *(pun8Buff + 0) = _pun8TBFogLightTable_[(*(pun8Buff + 0) << 12) | n32Fog | (_vuint32(pun8Tex[n32Ur + 0]) << 2) | mt_ptData[un8Cnt].tTotal.un8B];
          }
        }
          
        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ***** Special rendering functions
_vvoid _CVortexStaticSprite2D::LightRender(_CVortexMatrix3D &csTMatrix)
{
  // ..... Temporary variables
  _vuint8  un8Cnt, *pun8Tex, *pun8Buff;
  _vint32  n32Z, n32L, n32R, n32T, n32B, n32TexLeft = 0, n32TexTop = 0;
  _vint32  n32U, n32UStep, n32V, n32VStep, n32TY, n32TX, *pn32Z;
  _vint32  n32SLSize, n32Ur;
  _vuint32 un32YIndex;
  _vfloat  fX, fY, fW, fH;
  
  // ..... Render all sprites
  n32SLSize  = _vint32(mt_pcsImageSprite->mt_un32Width << 2);
  for(un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    // Transform position
    csTMatrix.Transform(mt_ptData[un8Cnt].csPos);
    // Perform Z clipping
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ > _pcsAct3DEnv_->mt_un32MaxZ) continue;
    if(mt_ptData[un8Cnt].csPos.m_tWorld.fZ < _pcsAct3DEnv_->mt_un32MinZ) continue;
    // Calculate Z value
    n32Z = _vint32(_vfloat(1.0) / mt_ptData[un8Cnt].csPos.m_tWorld.fZ * _vfloat(1 << _ZSTEP_PREC_)) + 
           _pcsAct3DEnv_->mt_tRB->n32ZTrans;
    // Project sprite
    fX = mt_ptData[un8Cnt].csPos.m_tWorld.fX * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fXShift;
    fY = mt_ptData[un8Cnt].csPos.m_tWorld.fY * _pcsAct3DEnv_->mt_fYScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ +
         _pcsAct3DEnv_->mt_fYShift;
    fW = _vfloat(mt_pcsImageSprite->mt_un32Width) * _pcsAct3DEnv_->mt_fXScale / mt_ptData[un8Cnt].csPos.m_tWorld.fZ;
    fH = _vfloat(mt_pcsImageSprite->mt_un32Height) * _pcsAct3DEnv_->mt_fYScale / (-mt_ptData[un8Cnt].csPos.m_tWorld.fZ);
    n32L = _vint32(fX - (fW / _vfloat(2.0)));
    n32R = _vint32(fX + (fW / _vfloat(2.0)));
    n32T = _vint32(fY - (fH / _vfloat(2.0)));
    n32B = _vint32(fY + (fH / _vfloat(2.0)));
    // Perform 2D clipping
    if(n32R < _pcsAct3DEnv_->mt_n32MinX) continue;
    if(n32B < _pcsAct3DEnv_->mt_n32MinY) continue;
    if(n32L > _pcsAct3DEnv_->mt_n32MaxX) continue;
    if(n32T > _pcsAct3DEnv_->mt_n32MaxY) continue;
    if(n32R > _pcsAct3DEnv_->mt_n32MaxX) n32R = _pcsAct3DEnv_->mt_n32MaxX - 1;
    if(n32B > _pcsAct3DEnv_->mt_n32MaxY) n32B = _pcsAct3DEnv_->mt_n32MaxY - 1;
    if(n32L < _pcsAct3DEnv_->mt_n32MinX)
    {
      n32TexLeft = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinX - n32L) * 
                           _vfloat(mt_pcsImageSprite->mt_un32Width) / fW);
      n32L = _vint32(_pcsAct3DEnv_->mt_n32MinX);
    }      
    if(n32T < _pcsAct3DEnv_->mt_n32MinY)
    {
      n32TexTop = _vint32(_vfloat(_pcsAct3DEnv_->mt_n32MinY - n32T) * 
                          _vfloat(mt_pcsImageSprite->mt_un32Height) / fH);
      n32T = _vint32(_pcsAct3DEnv_->mt_n32MinY);
    }

    // Render
    n32V       = n32TexTop << _SSTEP_PREC_;
    n32UStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Width << _SSTEP_PREC_) / fW);
    n32VStep   = _vint32(_vfloat(mt_pcsImageSprite->mt_un32Height << _SSTEP_PREC_) / fH);
    un32YIndex = n32T * _pcsAct3DEnv_->mt_tRB->un32Width;   
    for(n32TY = n32T; n32TY < n32B; n32TY++)
    {
      n32U     = n32TexLeft << _SSTEP_PREC_;
      pn32Z    = &_pcsAct3DEnv_->mt_tRB->pn32ZBuffer[un32YIndex + _vuint32(n32L)];
      pun8Buff = &_pcsAct3DEnv_->mt_tRB->pun8Buffer[(un32YIndex + _vuint32(n32L)) << 2];

      pun8Tex  = &mt_pcsImageSprite->mt_pun8Buffer[(n32V >> _SSTEP_PREC_) * n32SLSize];
      
      for(n32TX = n32L; n32TX < n32R; n32TX++)
      {
        if(*pn32Z < n32Z)
        {
          n32Ur = n32U >> _SSTEP_PREC_ << 2;
          if(*((_vuint32 *)(pun8Tex + n32Ur)) != mt_un32ExclColor)
          {
            *pn32Z     = n32Z;
            *(pun8Buff + 2) = _aun8CAddTable_[(_vuint32(pun8Tex[n32Ur + 2]) << 2) | (*(pun8Buff + 2) >> 3)];
            *(pun8Buff + 1) = _aun8CAddTable_[(_vuint32(pun8Tex[n32Ur + 1]) << 2) | (*(pun8Buff + 1) >> 3)];
            *(pun8Buff + 0) = _aun8CAddTable_[(_vuint32(pun8Tex[n32Ur + 0]) << 2) | (*(pun8Buff + 0) >> 3)];
          }
        }
          
        n32U += n32UStep;
        pn32Z++;
        pun8Buff += 4;
      }
      n32V += n32VStep;
      un32YIndex += _pcsAct3DEnv_->mt_tRB->un32Width;
    }
  }
}

// ============================================================================================
// End of vspr2d.cpp
// ============================================================================================
