// ############################################################################################
//  vspr2d.h
//  Definition & implementation of _CVortexSprite2D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004 Aloysius Indrayanto
//  This file is part of the VortexGE
//
//  The source codes in this file is adopted from the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This program is free software; you can redistribute it and/or modify it under the terms of 
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this 
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, 
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VSPR2D_H_
#define _VSPR2D_H_

// ============================================================================================
// Header files
// ============================================================================================
// ***** C++ headers
#include <math.h>
// ***** VortexGE headers
#include <vmat3d.h>
#include <vimage.h>
#include <vpoint3d.h>

// ============================================================================================
// Exported structures definition
// ============================================================================================
// ***** Static sprite data
struct _TSSpriteData
{
  _CVortexPoint3D csPos;      // The sprite's position in 3D world
  _vfloat         fX, fY, fZ; // The sprite's normal
  _TVortexUN8RGB  tOmni;      // The sprite's color (from omni light)
  _TVortexUN8RGB  tTotal;     // The sprite's color (total)

  // Special overloaded operators (to support aligned memory allocation for SSE)
  inline _vvoid *operator new      (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
  inline _vvoid *operator new[]    (_vuint32 un32Size) { return(ssemalloc(un32Size)); }
  inline _vvoid  operator delete   (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
  inline _vvoid  operator delete[] (_vvoid *pvBuffer)  { return(ssefree(pvBuffer)); }
};

// ============================================================================================
// Exported classes definition
// ============================================================================================
// ***** The 2D static sprite class
class _CVortexStaticSprite2D
{
  protected:
    // Variables
  _TSSpriteData     *mt_ptData;         // The indivudial sprite datas
  _TVortexUN8RGB    mt_tAmbient;        // The sprites' color (from ambient light)
  _CVortexImage     *mt_pcsImageSprite; // The sprites' image
  _vuint32          mt_un32ExclColor;   // The color that will not be drawn
  _vuint8           mt_un8TotData;      // Total sprite datas
  _vfloat           mt_fXo;             // Position offset for pivoting
  _vfloat           mt_fYo;             // Position offset for pivoting
  _TVortexOmniLight *mt_ptOmni;         // Used omni light
  // Initialization functions
  _vvoid _CalcIntenOmni_(_TVortexOmniLight *ptVOmniLight);
  _vvoid _CalcIntenOmni_(_vuint8 un8Index, _TVortexOmniLight *ptVOmniLight);
  _vvoid _CalcIntenOmni_(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _TVortexOmniLight *ptVOmniLight);
  inline _vvoid _CalcIntenAmbient_(_TVortexAmbientLight *ptVAmbientLight);
  inline _vvoid _CalcInten_(_vvoid);
  inline _vvoid _CalcInten_(_vuint8 un8Index);
  inline _vvoid _CalcInten_(_vuint8 un8StartIndex, _vuint8 un8EndIndex);
  
  public:
    // Constructors & destructor
    _CVortexStaticSprite2D();
    ~_CVortexStaticSprite2D();
    // Initialization functions
    _vbool Initialize(_vfloat *pfPosXYZs, _vuint8 un8PosCount, _CVortexImage *pcsImageSprite, 
                      _EVortexPivot ePivot = EBottomCenter,
                      _vuint32 un32ExcludedColorVRGB = EXCLUDED_COLOR_FIRST_PIXEL,
                      _TVortexAmbientLight *ptVAmbientLight = NULL, 
                      _TVortexOmniLight *ptVOmniLight = NULL);
    inline _vbool Initialize(_vfloat *pfPosXYZs, _vuint8 un8PosCount, _CVortexAnimatedImage *pcsImageSprite, 
                             _EVortexPivot ePivot = EBottomCenter,
                             _vuint32 un32ExcludedColorVRGB = EXCLUDED_COLOR_FIRST_PIXEL,
                             _TVortexAmbientLight *ptVAmbientLight = NULL, 
                             _TVortexOmniLight *ptVOmniLight = NULL);
    // Lighting functions
    inline _vvoid ResetAmbientLight(_TVortexAmbientLight *ptVAmbientLight = NULL);
    inline _vvoid ResetOmniLight(_TVortexOmniLight *ptVOmniLight = NULL);
    inline _vvoid ResetLights(_TVortexAmbientLight *ptVAmbientLight = NULL, _TVortexOmniLight *ptVOmniLight = NULL);
    // Dynamic control functions
    inline _vvoid SetPosition(_vfloat fWorldX, _vfloat fWorldY, _vfloat fWorldZ);
    inline _vvoid SetPosition(_vuint8 un8Index, _vfloat fWorldX, _vfloat fWorldY, _vfloat fWorldZ);
    inline _vvoid SetPosition(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _vfloat fWorldX, _vfloat fWorldY, _vfloat fWorldZ);
    inline _vvoid Translate(_vfloat fAmountX, _vfloat fAmountY, _vfloat fAmountZ);
    inline _vvoid Translate(_vuint8 un8Index, _vfloat fAmountX, _vfloat fAmountY, _vfloat fAmountZ);
    inline _vvoid Translate(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _vfloat fAmountX, _vfloat fAmountY, _vfloat fAmountZ);
    inline _vvoid SetDirection(_vfloat fDirX, _vfloat fDirY, _vfloat fDirZ);
    inline _vvoid SetDirection(_vuint8 un8Index, _vfloat fDirX, _vfloat fDirY, _vfloat fDirZ);
    inline _vvoid SetDirection(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _vfloat fDirX, _vfloat fDirY, _vfloat fDirZ);
    inline _vvoid RecalculateLighting(_vvoid);
    inline _vvoid RecalculateLighting(_vuint8 un8Index);
    inline _vvoid RecalculateLighting(_vuint8 un8StartIndex, _vuint8 un8EndIndex);
    // Normal rendering functions
    _vvoid SNRender(_CVortexMatrix3D &csTMatrix); // Solid sprite, not shaded
    _vvoid SFRender(_CVortexMatrix3D &csTMatrix); // Solid sprite, flat shaded
    _vvoid TNRender(_CVortexMatrix3D &csTMatrix); // Solid sprite, not shaded
    _vvoid TFRender(_CVortexMatrix3D &csTMatrix); // Solid sprite, flat shaded
    // Fogged rendering functions
    _vvoid FogSNRender(_CVortexMatrix3D &csTMatrix); // Fogged, solid sprite, not shaded
    _vvoid FogSFRender(_CVortexMatrix3D &csTMatrix); // Fogged, solid sprite, flat shaded
    _vvoid FogTNRender(_CVortexMatrix3D &csTMatrix); // Fogged, solid sprite, not shaded
    _vvoid FogTFRender(_CVortexMatrix3D &csTMatrix); // Fogged, solid sprite, flat shaded
    // Special rendering functions
    _vvoid LightRender(_CVortexMatrix3D &csTMatrix); // Simulate lighting using color addition rendering
};

// ############################################################################################
// Inline functions implementation
// ############################################################################################
// ============================================================================================
// From exported classes
// ============================================================================================
// --------------------------------------------------------------------------------------------
// The 2D static sprite class
// --------------------------------------------------------------------------------------------
// ***** Initialization functions
_vbool _CVortexStaticSprite2D::Initialize(_vfloat *pfPosXYZs, _vuint8 un8PosCount, _CVortexAnimatedImage *pcsImageSprite, 
                                          _EVortexPivot ePivot, _vuint32 un32ExcludedColorVRGB,
                                          _TVortexAmbientLight *ptVAmbientLight,  
                                          _TVortexOmniLight *ptVOmniLight)
{
  return(Initialize(pfPosXYZs, un8PosCount, (_CVortexImage *) pcsImageSprite, ePivot, 
                    un32ExcludedColorVRGB, ptVAmbientLight, ptVOmniLight));
}
_vvoid _CVortexStaticSprite2D::_CalcIntenAmbient_(_TVortexAmbientLight *ptVAmbientLight)
{
  // ..... Use the supplied ambient light
  if(ptVAmbientLight)
  {
    mt_tAmbient.un8R  = ptVAmbientLight->un8R;
    mt_tAmbient.un8G  = ptVAmbientLight->un8G;
    mt_tAmbient.un8B  = ptVAmbientLight->un8B;
  }
  else
  // ..... Use the default ambient light from the active 3D environment
  {
    mt_tAmbient.un8R  = _pcsAct3DEnv_->mt_tDefALight.un8R;
    mt_tAmbient.un8G  = _pcsAct3DEnv_->mt_tDefALight.un8G;
    mt_tAmbient.un8B  = _pcsAct3DEnv_->mt_tDefALight.un8B;
  }
}
_vvoid _CVortexStaticSprite2D::_CalcInten_(_vvoid)
{
  for(_vuint8 un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    mt_ptData[un8Cnt].tTotal.un8R = (mt_ptData[un8Cnt].tOmni.un8R + mt_tAmbient.un8R) >> 1;
    mt_ptData[un8Cnt].tTotal.un8G = (mt_ptData[un8Cnt].tOmni.un8G + mt_tAmbient.un8G) >> 1;
    mt_ptData[un8Cnt].tTotal.un8B = (mt_ptData[un8Cnt].tOmni.un8B + mt_tAmbient.un8B) >> 1;
  }
}
_vvoid _CVortexStaticSprite2D::_CalcInten_(_vuint8 un8Index)
{
  mt_ptData[un8Index].tTotal.un8R = (mt_ptData[un8Index].tOmni.un8R + mt_tAmbient.un8R) >> 1;
  mt_ptData[un8Index].tTotal.un8G = (mt_ptData[un8Index].tOmni.un8G + mt_tAmbient.un8G) >> 1;
  mt_ptData[un8Index].tTotal.un8B = (mt_ptData[un8Index].tOmni.un8B + mt_tAmbient.un8B) >> 1;
}
_vvoid _CVortexStaticSprite2D::_CalcInten_(_vuint8 un8StartIndex, _vuint8 un8EndIndex)
{
  for(_vuint8 un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++)
  {
    mt_ptData[un8Cnt].tTotal.un8R = (mt_ptData[un8Cnt].tOmni.un8R + mt_tAmbient.un8R) >> 1;
    mt_ptData[un8Cnt].tTotal.un8G = (mt_ptData[un8Cnt].tOmni.un8G + mt_tAmbient.un8G) >> 1;
    mt_ptData[un8Cnt].tTotal.un8B = (mt_ptData[un8Cnt].tOmni.un8B + mt_tAmbient.un8B) >> 1;
  }
}

// ***** Lighting functions
_vvoid _CVortexStaticSprite2D::ResetAmbientLight(_TVortexAmbientLight *ptVAmbientLight)
{
  _CalcIntenAmbient_(ptVAmbientLight);
  _CalcInten_();
}
_vvoid _CVortexStaticSprite2D::ResetOmniLight(_TVortexOmniLight *ptVOmniLight)
{
  _CalcIntenOmni_(ptVOmniLight);
  _CalcInten_();
}
_vvoid _CVortexStaticSprite2D::ResetLights(_TVortexAmbientLight *ptVAmbientLight, _TVortexOmniLight *ptVOmniLight)
{
  _CalcIntenAmbient_(ptVAmbientLight);
  _CalcIntenOmni_(ptVOmniLight);
  _CalcInten_();
}

// ***** Dynamic control functions
_vvoid _CVortexStaticSprite2D::SetPosition(_vfloat fWorldX, _vfloat fWorldY, _vfloat fWorldZ)
{
  for(_vuint8 un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    mt_ptData[un8Cnt].csPos.m_tLocal.fX = fWorldX + mt_fXo;
    mt_ptData[un8Cnt].csPos.m_tLocal.fY = fWorldY + mt_fYo;
    mt_ptData[un8Cnt].csPos.m_tLocal.fZ = fWorldZ;
  }
}
_vvoid _CVortexStaticSprite2D::SetPosition(_vuint8 un8Index, _vfloat fWorldX, _vfloat fWorldY, _vfloat fWorldZ)
{
  mt_ptData[un8Index].csPos.m_tLocal.fX = fWorldX + mt_fXo;
  mt_ptData[un8Index].csPos.m_tLocal.fY = fWorldY + mt_fYo;
  mt_ptData[un8Index].csPos.m_tLocal.fZ = fWorldZ;
}
_vvoid _CVortexStaticSprite2D::SetPosition(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _vfloat fWorldX, _vfloat fWorldY, _vfloat fWorldZ)
{
  for(_vuint8 un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++)
  {
    mt_ptData[un8Cnt].csPos.m_tLocal.fX = fWorldX + mt_fXo;
    mt_ptData[un8Cnt].csPos.m_tLocal.fY = fWorldY + mt_fYo;
    mt_ptData[un8Cnt].csPos.m_tLocal.fZ = fWorldZ;
  }
}
_vvoid _CVortexStaticSprite2D::Translate(_vfloat fAmountX, _vfloat fAmountY, _vfloat fAmountZ)
{
  for(_vuint8 un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    mt_ptData[un8Cnt].csPos.m_tLocal.fX += fAmountX;
    mt_ptData[un8Cnt].csPos.m_tLocal.fY += fAmountY;
    mt_ptData[un8Cnt].csPos.m_tLocal.fZ += fAmountZ;
  }
}
_vvoid _CVortexStaticSprite2D::Translate(_vuint8 un8Index, _vfloat fAmountX, _vfloat fAmountY, _vfloat fAmountZ)
{
  mt_ptData[un8Index].csPos.m_tLocal.fX += fAmountX;
  mt_ptData[un8Index].csPos.m_tLocal.fY += fAmountY;
  mt_ptData[un8Index].csPos.m_tLocal.fZ += fAmountZ;
}
_vvoid _CVortexStaticSprite2D::Translate(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _vfloat fAmountX, _vfloat fAmountY, _vfloat fAmountZ)
{
  for(_vuint8 un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++)
  {
    mt_ptData[un8Cnt].csPos.m_tLocal.fX += fAmountX;
    mt_ptData[un8Cnt].csPos.m_tLocal.fY += fAmountY;
    mt_ptData[un8Cnt].csPos.m_tLocal.fZ += fAmountZ;
  }
}
_vvoid _CVortexStaticSprite2D::SetDirection(_vfloat fDirX, _vfloat fDirY, _vfloat fDirZ)
{
  _vfloat fMag = sqrt(fDirX * fDirX + fDirY * fDirY + fDirZ * fDirZ);
  fDirX /= fMag; fDirY /= fMag; fDirZ /= fMag;
  for(_vuint8 un8Cnt = 0; un8Cnt < mt_un8TotData; un8Cnt++)
  {
    mt_ptData[un8Cnt].fX = fDirX;
    mt_ptData[un8Cnt].fY = fDirY;
    mt_ptData[un8Cnt].fZ = fDirZ;
  }
}
_vvoid _CVortexStaticSprite2D::SetDirection(_vuint8 un8Index, _vfloat fDirX, _vfloat fDirY, _vfloat fDirZ)
{
  _vfloat fMag = sqrt(fDirX * fDirX + fDirY * fDirY + fDirZ * fDirZ);
  mt_ptData[un8Index].fX = fDirX / fMag;
  mt_ptData[un8Index].fY = fDirY / fMag;
  mt_ptData[un8Index].fZ = fDirZ / fMag;
}
_vvoid _CVortexStaticSprite2D::SetDirection(_vuint8 un8StartIndex, _vuint8 un8EndIndex, _vfloat fDirX, _vfloat fDirY, _vfloat fDirZ)
{
  _vfloat fMag = sqrt(fDirX * fDirX + fDirY * fDirY + fDirZ * fDirZ);
  fDirX /= fMag; fDirY /= fMag; fDirZ /= fMag;
  for(_vuint8 un8Cnt = un8StartIndex; un8Cnt <= un8EndIndex; un8Cnt++)
  {
    mt_ptData[un8Cnt].fX = fDirX;
    mt_ptData[un8Cnt].fY = fDirY;
    mt_ptData[un8Cnt].fZ = fDirZ;
  }
}
_vvoid _CVortexStaticSprite2D::RecalculateLighting(_vvoid)
{
  _CalcIntenOmni_(mt_ptOmni);
  _CalcInten_();
}
_vvoid _CVortexStaticSprite2D::RecalculateLighting(_vuint8 un8Index)
{
  _CalcIntenOmni_(un8Index, mt_ptOmni);
  _CalcInten_(un8Index);
}
_vvoid _CVortexStaticSprite2D::RecalculateLighting(_vuint8 un8StartIndex, _vuint8 un8EndIndex)
{
  _CalcIntenOmni_(un8StartIndex, un8EndIndex, mt_ptOmni);
  _CalcInten_(un8StartIndex, un8EndIndex);
}

#endif
// ============================================================================================
// End of vspr2d.h
// ============================================================================================
